@@ -5,9 +5,10 @@ test suite on top of PyPy
 
 """
 import py
+import pytest
 import sys
-import pypy
 import re
+import pypy
 from pypy.interpreter.gateway import ApplevelClass
 from pypy.interpreter.error import OperationError
 from pypy.interpreter.module import Module as PyPyModule
@@ -17,7 +18,7 @@ from pypy.interpreter.main import run_string, run_file
 from pypy.conftest import option as pypy_option
 
 from pypy.tool.pytest import appsupport
-from pypy.tool.pytest.confpath import pypydir, rpythondir, testdir, testresultdir
+from pypy.tool.pytest.confpath import pypydir, testdir, testresultdir
 from rpython.config.parse import parse_info
 
 pytest_plugins = "resultlog",
@@ -59,7 +60,11 @@ class RegrTest:
     def __init__(self, basename, core=False, compiler=None, usemodules='',
                  skip=None):
         self.basename = basename
-        self._usemodules = usemodules.split() + ['signal', 'time', 'itertools', '_socket']
+        self._usemodules = usemodules.split() + [
+            '_socket', 'binascii', 'time',
+            'select', 'signal', ]
+        if not sys.platform == 'win32':
+            self._usemodules.extend(['_posixsubprocess', 'fcntl'])
         self._compiler = compiler
         self.core = core
         self.skip = skip
@@ -95,25 +100,22 @@ testmap = [
     RegrTest('test___all__.py', core=True),
     RegrTest('test___future__.py', core=True),
     RegrTest('test__locale.py', usemodules='_locale'),
+    RegrTest('test__opcode.py'),
     RegrTest('test__osx_support.py'),
     RegrTest('test_abc.py'),
     RegrTest('test_abstract_numbers.py'),
-    RegrTest('test_aepack.py'),
     RegrTest('test_aifc.py'),
-    RegrTest('test_al.py'),
-    RegrTest('test_anydbm.py', usemodules='struct'),
-    RegrTest('test_applesingle.py'),
     RegrTest('test_argparse.py', usemodules='binascii'),
     RegrTest('test_array.py', core=True, usemodules='struct array binascii'),
-    RegrTest('test_ascii_formatd.py'),
+    RegrTest('test_asdl_parser.py'),
     RegrTest('test_ast.py', core=True, usemodules='struct'),
     RegrTest('test_asynchat.py', usemodules='select fcntl'),
+    RegrTest('test_asyncio'),
     RegrTest('test_asyncore.py', usemodules='select fcntl'),
     RegrTest('test_atexit.py', core=True),
     RegrTest('test_audioop.py'),
     RegrTest('test_augassign.py', core=True),
     RegrTest('test_base64.py', usemodules='struct'),
-    RegrTest('test_bastion.py'),
     RegrTest('test_bigaddrspace.py'),
     RegrTest('test_bigmem.py'),
     RegrTest('test_binascii.py', usemodules='binascii'),
@@ -121,10 +123,7 @@ testmap = [
     RegrTest('test_binop.py', core=True),
     RegrTest('test_bisect.py', core=True, usemodules='_bisect'),
     RegrTest('test_bool.py', core=True),
-    RegrTest('test_bsddb.py'),
-    RegrTest('test_bsddb185.py'),
-    RegrTest('test_bsddb3.py'),
-    RegrTest('test_buffer.py'),
+    RegrTest('test_buffer.py', core=True),
     RegrTest('test_bufio.py', core=True),
     RegrTest('test_builtin.py', core=True, usemodules='binascii'),
     RegrTest('test_bytes.py', usemodules='struct binascii'),
@@ -132,17 +131,16 @@ testmap = [
     RegrTest('test_calendar.py'),
     RegrTest('test_call.py', core=True),
     RegrTest('test_capi.py', usemodules='cpyext'),
-    RegrTest('test_cd.py'),
-    RegrTest('test_cfgparser.py'),
     RegrTest('test_cgi.py'),
+    RegrTest('test_cgitb.py'),
     RegrTest('test_charmapcodec.py', core=True),
-    RegrTest('test_cl.py'),
     RegrTest('test_class.py', core=True),
     RegrTest('test_cmath.py', core=True),
     RegrTest('test_cmd.py'),
     RegrTest('test_cmd_line.py'),
     RegrTest('test_cmd_line_script.py'),
     RegrTest('test_code.py', core=True),
+    RegrTest('test_code_module.py'),
     RegrTest('test_codeccallbacks.py', core=True),
     RegrTest('test_codecencodings_cn.py', usemodules='_multibytecodec'),
     RegrTest('test_codecencodings_hk.py', usemodules='_multibytecodec'),
@@ -155,57 +153,56 @@ testmap = [
     RegrTest('test_codecmaps_jp.py', usemodules='_multibytecodec'),
     RegrTest('test_codecmaps_kr.py', usemodules='_multibytecodec'),
     RegrTest('test_codecmaps_tw.py', usemodules='_multibytecodec'),
-    RegrTest('test_codecs.py', core=True, usemodules='_multibytecodec'),
+    RegrTest('test_codecs.py', core=True, usemodules='_multibytecodec struct unicodedata array'),
     RegrTest('test_codeop.py', core=True),
-    RegrTest('test_coding.py', core=True),
-    RegrTest('test_coercion.py', core=True),
     RegrTest('test_collections.py', usemodules='binascii struct'),
     RegrTest('test_colorsys.py'),
-    RegrTest('test_commands.py'),
     RegrTest('test_compare.py', core=True),
     RegrTest('test_compile.py', core=True),
     RegrTest('test_compileall.py'),
-    RegrTest('test_compiler.py', core=False, skip="slowly deprecating compiler"),
     RegrTest('test_complex.py', core=True),
-    RegrTest('test_complex_args.py'),
+    RegrTest('test_concurrent_futures.py', skip="XXX: deadlocks" if sys.platform == 'win32' else False),
+    RegrTest('test_configparser.py'),
     RegrTest('test_contains.py', core=True),
     RegrTest('test_contextlib.py', usemodules="thread"),
-    RegrTest('test_cookie.py'),
-    RegrTest('test_cookielib.py'),
     RegrTest('test_copy.py', core=True),
-    RegrTest('test_copy_reg.py', core=True),
-    RegrTest('test_cpickle.py', core=True),
+    RegrTest('test_copyreg.py', core=True),
+    RegrTest('test_coroutines.py'),
     RegrTest('test_cprofile.py'),
+    RegrTest('test_crashers.py'),
     RegrTest('test_crypt.py', usemodules='crypt'),
     RegrTest('test_csv.py', usemodules='_csv'),
     RegrTest('test_ctypes.py', usemodules="_rawffi thread cpyext"),
     RegrTest('test_curses.py'),
     RegrTest('test_datetime.py', usemodules='binascii struct'),
     RegrTest('test_dbm.py'),
+    RegrTest('test_dbm_dumb.py'),
+    RegrTest('test_dbm_gnu.py'),
+    RegrTest('test_dbm_ndbm.py'),
     RegrTest('test_decimal.py'),
     RegrTest('test_decorators.py', core=True),
     RegrTest('test_defaultdict.py', usemodules='_collections'),
-    RegrTest('test_deque.py', core=True, usemodules='_collections'),
+    RegrTest('test_deque.py', core=True, usemodules='_collections struct'),
     RegrTest('test_descr.py', core=True, usemodules='_weakref'),
     RegrTest('test_descrtut.py', core=True),
+    RegrTest('test_devpoll.py'),
     RegrTest('test_dict.py', core=True),
     RegrTest('test_dictcomps.py', core=True),
     RegrTest('test_dictviews.py', core=True),
     RegrTest('test_difflib.py'),
-    RegrTest('test_dircache.py', core=True),
     RegrTest('test_dis.py'),
     RegrTest('test_distutils.py', skip=True),
-    RegrTest('test_dl.py'),
     RegrTest('test_doctest.py', usemodules="thread"),
     RegrTest('test_doctest2.py'),
     RegrTest('test_docxmlrpc.py'),
-    RegrTest('test_dumbdbm.py'),
     RegrTest('test_dummy_thread.py', core=True),
     RegrTest('test_dummy_threading.py', core=True),
-    RegrTest('test_email.py'),
-    RegrTest('test_email_codecs.py'),
-    RegrTest('test_email_renamed.py'),
+    RegrTest('test_dynamic.py'),
+    RegrTest('test_dynamicclassattribute.py'),
+    RegrTest('test_eintr.py'),
+    RegrTest('test_email'),
     RegrTest('test_ensurepip.py'),
+    RegrTest('test_enum.py'),
     RegrTest('test_enumerate.py', core=True),
     RegrTest('test_eof.py', core=True),
     RegrTest('test_epoll.py'),
@@ -213,41 +210,38 @@ testmap = [
     RegrTest('test_exception_variations.py'),
     RegrTest('test_exceptions.py', core=True),
     RegrTest('test_extcall.py', core=True),
+    RegrTest('test_faulthandler.py'),
     RegrTest('test_fcntl.py', usemodules='fcntl'),
     RegrTest('test_file.py', usemodules="posix", core=True),
-    RegrTest('test_file2k.py', usemodules="posix", core=True),
     RegrTest('test_file_eintr.py'),
     RegrTest('test_filecmp.py', core=True),
     RegrTest('test_fileinput.py', core=True),
     RegrTest('test_fileio.py'),
+    RegrTest('test_finalization.py'),
     RegrTest('test_float.py', core=True),
+    RegrTest('test_flufl.py'),
     RegrTest('test_fnmatch.py', core=True),
     RegrTest('test_fork1.py', usemodules="thread"),
     RegrTest('test_format.py', core=True),
-    RegrTest('test_fpformat.py', core=True),
     RegrTest('test_fractions.py'),
-    RegrTest('test_frozen.py', skip="unsupported extension module"),
+    RegrTest('test_frame.py'),
+    RegrTest('test_fstring.py'),
     RegrTest('test_ftplib.py'),
     RegrTest('test_funcattrs.py', core=True),
     RegrTest('test_functools.py'),
     RegrTest('test_future.py', core=True),
-    RegrTest('test_future1.py', core=True),
-    RegrTest('test_future2.py', core=True),
     RegrTest('test_future3.py', core=True),
     RegrTest('test_future4.py', core=True),
     RegrTest('test_future5.py', core=True),
-    RegrTest('test_future_builtins.py'),
     RegrTest('test_gc.py', usemodules='_weakref', skip="implementation detail"),
     RegrTest('test_gdb.py', skip="not applicable"),
-    RegrTest('test_gdbm.py'),
     RegrTest('test_generators.py', core=True, usemodules='thread _weakref'),
     RegrTest('test_genericpath.py'),
     RegrTest('test_genexps.py', core=True, usemodules='_weakref'),
-    RegrTest('test_getargs.py'),
     RegrTest('test_getargs2.py', usemodules='binascii', skip=True),
     RegrTest('test_getopt.py', core=True),
+    RegrTest('test_getpass.py'),
     RegrTest('test_gettext.py'),
-    RegrTest('test_gl.py'),
     RegrTest('test_glob.py', core=True),
     RegrTest('test_global.py', core=True),
     RegrTest('test_grammar.py', core=True),
@@ -257,87 +251,89 @@ testmap = [
     RegrTest('test_hashlib.py', core=True),
     RegrTest('test_heapq.py', core=True),
     RegrTest('test_hmac.py'),
-    RegrTest('test_hotshot.py', skip="unsupported extension module"),
-    RegrTest('test_htmllib.py'),
+    RegrTest('test_html.py'),
     RegrTest('test_htmlparser.py'),
+    RegrTest('test_http_cookiejar.py'),
+    RegrTest('test_http_cookies.py'),
     RegrTest('test_httplib.py'),
     RegrTest('test_httpservers.py'),
     RegrTest('test_idle.py'),
-    RegrTest('test_imageop.py'),
     RegrTest('test_imaplib.py'),
-    RegrTest('test_imgfile.py'),
     RegrTest('test_imghdr.py'),
+    RegrTest('test_import'),
+    RegrTest('test_importlib'),
     RegrTest('test_imp.py', core=True, usemodules='thread'),
-    RegrTest('test_import.py', core=True),
-    RegrTest('test_importhooks.py', core=True),
-    RegrTest('test_importlib.py'),
     RegrTest('test_index.py'),
-    RegrTest('test_inspect.py'),
+    RegrTest('test_inspect.py', usemodules="struct unicodedata"),
     RegrTest('test_int.py', core=True),
     RegrTest('test_int_literal.py', core=True),
-    RegrTest('test_io.py', usemodules='array binascii'),
+    RegrTest('test_io.py', core=True, usemodules='array binascii'),
     RegrTest('test_ioctl.py'),
+    RegrTest('test_ipaddress.py'),
     RegrTest('test_isinstance.py', core=True),
     RegrTest('test_iter.py', core=True),
     RegrTest('test_iterlen.py', core=True, usemodules="_collections itertools"),
     RegrTest('test_itertools.py', core=True, usemodules="itertools struct"),
-    RegrTest('test_json.py'),
+    RegrTest('test_json'),
+    RegrTest('test_keyword.py'),
+    RegrTest('test_keywordonlyarg.py'),
     RegrTest('test_kqueue.py'),
     RegrTest('test_largefile.py'),
     RegrTest('test_lib2to3.py'),
     RegrTest('test_linecache.py'),
-    RegrTest('test_linuxaudiodev.py', skip="unsupported extension module"),
     RegrTest('test_list.py', core=True),
+    RegrTest('test_listcomps.py', core=True),
     RegrTest('test_locale.py', usemodules="_locale"),
     RegrTest('test_logging.py', usemodules='thread'),
     RegrTest('test_long.py', core=True),
-    RegrTest('test_long_future.py', core=True),
     RegrTest('test_longexp.py', core=True),
-    RegrTest('test_macos.py'),
-    RegrTest('test_macostools.py'),
+    RegrTest('test_lzma.py'),
     RegrTest('test_macpath.py'),
     RegrTest('test_macurl2path.py'),
     RegrTest('test_mailbox.py'),
+    RegrTest('test_mailcap.py'),
     RegrTest('test_marshal.py', core=True),
     RegrTest('test_math.py', core=True, usemodules='math'),
     RegrTest('test_memoryio.py'),
     RegrTest('test_memoryview.py'),
-    RegrTest('test_md5.py'),
-    RegrTest('test_mhlib.py'),
-    RegrTest('test_mimetools.py'),
+    RegrTest('test_metaclass.py', core=True),
     RegrTest('test_mimetypes.py'),
-    RegrTest('test_MimeWriter.py', core=False, usemodules='binascii'),
     RegrTest('test_minidom.py'),
     RegrTest('test_mmap.py', usemodules="mmap"),
     RegrTest('test_module.py', core=True),
     RegrTest('test_modulefinder.py'),
     RegrTest('test_msilib.py'),
     RegrTest('test_multibytecodec.py', usemodules='_multibytecodec'),
-    RegrTest('test_multifile.py'),
-    RegrTest('test_multiprocessing.py'),
-    RegrTest('test_mutants.py', core="possibly"),
-    RegrTest('test_mutex.py'),
+    RegrTest('test_multiprocessing_fork.py'),
+    RegrTest('test_multiprocessing_forkserver.py'),
+    RegrTest('test_multiprocessing_main_handling.py'),
+    RegrTest('test_multiprocessing_spawn.py'),
     RegrTest('test_netrc.py'),
-    RegrTest('test_new.py', core=True),
     RegrTest('test_nis.py'),
-    RegrTest('test_normalization.py'),
     RegrTest('test_nntplib.py'),
+    RegrTest('test_normalization.py'),
     RegrTest('test_ntpath.py'),
-    RegrTest('test_old_mailbox.py'),
+    RegrTest('test_numeric_tower.py'),
     RegrTest('test_opcodes.py', core=True),
     RegrTest('test_openpty.py'),
     RegrTest('test_operator.py', core=True),
     RegrTest('test_optparse.py'),
+    RegrTest('test_ordered_dict.py'),
     RegrTest('test_os.py', core=True),
     RegrTest('test_ossaudiodev.py'),
+    RegrTest('test_osx_env.py'),
     RegrTest('test_parser.py', skip="slowly deprecating compiler"),
+    RegrTest('test_pathlib.py'),
     RegrTest('test_pdb.py'),
     RegrTest('test_peepholer.py'),
     RegrTest('test_pep247.py'),
-    RegrTest('test_pep263.py'),
     RegrTest('test_pep277.py'),
-    RegrTest('test_pep292.py'),
+    RegrTest('test_pep3120.py'),
+    RegrTest('test_pep3131.py'),
+    RegrTest('test_pep3151.py'),
     RegrTest('test_pep352.py'),
+    RegrTest('test_pep380.py'),
+    RegrTest('test_pep479.py'),
     RegrTest('test_pickle.py', core=True),
     RegrTest('test_pickletools.py', core=False),
     RegrTest('test_pipes.py'),
@@ -348,7 +344,6 @@ testmap = [
     RegrTest('test_plistlib.py'),
     RegrTest('test_poll.py'),
     RegrTest('test_popen.py'),
-    RegrTest('test_popen2.py'),
     RegrTest('test_poplib.py'),
     RegrTest('test_posix.py', usemodules="_rawffi"),
     RegrTest('test_posixpath.py'),
@@ -359,57 +354,58 @@ testmap = [
     RegrTest('test_property.py', core=True),
     RegrTest('test_pstats.py'),
     RegrTest('test_pty.py', usemodules='fcntl termios select'),
+    RegrTest('test_pulldom.py'),
     RegrTest('test_pwd.py', usemodules="pwd"),
-    RegrTest('test_py3kwarn.py'),
     RegrTest('test_py_compile.py'),
     RegrTest('test_pyclbr.py'),
     RegrTest('test_pydoc.py'),
     RegrTest('test_pyexpat.py'),
     RegrTest('test_queue.py', usemodules='thread'),
     RegrTest('test_quopri.py'),
+    RegrTest('test_raise.py', core=True),
     RegrTest('test_random.py'),
+    RegrTest('test_range.py', core=True),
     RegrTest('test_re.py', core=True),
     RegrTest('test_readline.py'),
-    RegrTest('test_repr.py', core=True),
+    RegrTest('test_regrtest.py'),
+    RegrTest('test_reprlib.py', core=True),
     RegrTest('test_resource.py'),
-    RegrTest('test_rfc822.py'),
     RegrTest('test_richcmp.py', core=True),
     RegrTest('test_rlcompleter.py'),
     RegrTest('test_robotparser.py'),
     RegrTest('test_runpy.py'),
     RegrTest('test_sax.py'),
+    RegrTest('test_sched.py'),
     RegrTest('test_scope.py', core=True),
-    RegrTest('test_scriptpackages.py'),
+    RegrTest('test_script_helper.py'),
     RegrTest('test_select.py'),
+    RegrTest('test_selectors.py'),
     RegrTest('test_set.py', core=True),
-    RegrTest('test_sets.py'),
     RegrTest('test_setcomps.py', core=True),
-    RegrTest('test_sgmllib.py'),
-    RegrTest('test_sha.py'),
     RegrTest('test_shelve.py'),
     RegrTest('test_shlex.py'),
     RegrTest('test_shutil.py'),
     RegrTest('test_signal.py'),
-    RegrTest('test_SimpleHTTPServer.py', usemodules='binascii'),
     RegrTest('test_site.py', core=False),
     RegrTest('test_slice.py', core=True),
+    RegrTest('test_smtpd.py'),
     RegrTest('test_smtplib.py'),
     RegrTest('test_smtpnet.py'),
+    RegrTest('test_sndhdr.py'),
     RegrTest('test_socket.py', usemodules='thread _weakref'),
     RegrTest('test_socketserver.py', usemodules='thread'),
-    RegrTest('test_softspace.py', core=True),
     RegrTest('test_sort.py', core=True),
+    RegrTest('test_source_encoding.py'),
     RegrTest('test_spwd.py'),
     RegrTest('test_sqlite.py', usemodules="thread _rawffi zlib"),
-    RegrTest('test_ssl.py', usemodules='_ssl _socket select'),
+    RegrTest('test_ssl.py', usemodules='_socket select'),
     RegrTest('test_startfile.py'),
     RegrTest('test_stat.py'),
-    RegrTest('test_str.py', core=True),
+    RegrTest('test_statistics.py'),
     RegrTest('test_strftime.py'),
     RegrTest('test_string.py', core=True),
-    RegrTest('test_StringIO.py', core=True, usemodules='cStringIO array'),
     RegrTest('test_stringprep.py'),
-    RegrTest('test_strop.py', skip="deprecated"),
+    RegrTest('test_strlit.py', core=True),
     RegrTest('test_strptime.py'),
     RegrTest('test_strtod.py'),
     RegrTest('test_struct.py', usemodules='struct'),
@@ -417,14 +413,17 @@ testmap = [
     RegrTest('test_structseq.py'),
     RegrTest('test_subprocess.py', usemodules='signal'),
     RegrTest('test_sunau.py'),
-    RegrTest('test_sunaudiodev.py'),
     RegrTest('test_sundry.py'),
+    RegrTest('test_super.py', core=True),
+    RegrTest('test_support.py'),
     RegrTest('test_symtable.py', skip="implementation detail"),
     RegrTest('test_syntax.py', core=True),
     RegrTest('test_sys.py', core=True, usemodules='struct'),
     RegrTest('test_sys_setprofile.py', core=True),
     RegrTest('test_sys_settrace.py', core=True),
     RegrTest('test_sysconfig.py'),
+    RegrTest('test_sysconfig_pypy.py'),
+    RegrTest('test_syslog.py'),
     RegrTest('test_tarfile.py'),
     RegrTest('test_tcl.py'),
     RegrTest('test_telnetlib.py'),
@@ -432,38 +431,41 @@ testmap = [
     RegrTest('test_textwrap.py'),
     RegrTest('test_thread.py', usemodules="thread", core=True),
     RegrTest('test_threaded_import.py', usemodules="thread", core=True),
-    RegrTest('test_threadedtempfile.py',
-             usemodules="thread", core=False),
+    RegrTest('test_threadedtempfile.py', usemodules="thread", core=False),
     RegrTest('test_threading.py', usemodules="thread", core=True),
     RegrTest('test_threading_local.py', usemodules="thread", core=True),
     RegrTest('test_threadsignals.py', usemodules="thread"),
-    RegrTest('test_time.py', core=True),
+    RegrTest('test_time.py', core=True, usemodules="struct thread _rawffi"),
+    RegrTest('test_timeit.py'),
     RegrTest('test_timeout.py'),
+    RegrTest('test_tix.py'),
     RegrTest('test_tk.py'),
     RegrTest('test_tokenize.py'),
-    RegrTest('test_tools.py'),
-    RegrTest('test_trace.py'),
+    RegrTest('test_tools', skip="CPython internal details"),
     RegrTest('test_traceback.py', core=True),
-    RegrTest('test_transformer.py', core=True),
+    RegrTest('test_tracemalloc.py'),
+    RegrTest('test_trace.py'),
     RegrTest('test_ttk_guionly.py'),
     RegrTest('test_ttk_textonly.py'),
     RegrTest('test_tuple.py', core=True),
+    RegrTest('test_turtle.py'),
     RegrTest('test_typechecks.py'),
     RegrTest('test_types.py', core=True),
+    RegrTest('test_typing.py'),
     RegrTest('test_ucn.py'),
     RegrTest('test_unary.py', core=True),
-    RegrTest('test_undocumented_details.py'),
     RegrTest('test_unicode.py', core=True),
     RegrTest('test_unicode_file.py'),
     RegrTest('test_unicodedata.py'),
     RegrTest('test_unittest.py', core=True),
     RegrTest('test_univnewlines.py'),
-    RegrTest('test_univnewlines2k.py', core=True),
     RegrTest('test_unpack.py', core=True),
+    RegrTest('test_unpack_ex.py', core=True),
     RegrTest('test_urllib.py'),
     RegrTest('test_urllib2.py'),
     RegrTest('test_urllib2_localnet.py', usemodules="thread"),
     RegrTest('test_urllib2net.py'),
+    RegrTest('test_urllib_response.py'),
     RegrTest('test_urllibnet.py'),
     RegrTest('test_urlparse.py'),
     RegrTest('test_userdict.py', core=True),
@@ -471,24 +473,25 @@ testmap = [
     RegrTest('test_userstring.py', core=True),
     RegrTest('test_uu.py'),
     RegrTest('test_uuid.py'),
+    RegrTest('test_venv.py', usemodules="struct"),
     RegrTest('test_wait3.py', usemodules="thread"),
     RegrTest('test_wait4.py', usemodules="thread"),
-    RegrTest('test_warnings.py', core=True),
+    RegrTest('test_warnings'),
     RegrTest('test_wave.py'),
     RegrTest('test_weakref.py', core=True, usemodules='_weakref'),
     RegrTest('test_weakset.py'),
-    RegrTest('test_whichdb.py'),
+    RegrTest('test_webbrowser.py'),
     RegrTest('test_winreg.py'),
     RegrTest('test_winsound.py'),
     RegrTest('test_with.py'),
     RegrTest('test_wsgiref.py'),
     RegrTest('test_xdrlib.py'),
+    RegrTest('test_xml_dom_minicompat.py'),
     RegrTest('test_xml_etree.py'),
     RegrTest('test_xml_etree_c.py'),
-    RegrTest('test_xmllib.py'),
     RegrTest('test_xmlrpc.py'),
-    RegrTest('test_xpickle.py'),
-    RegrTest('test_xrange.py', core=True),
+    RegrTest('test_xmlrpc_net.py'),
+    RegrTest('test_zipapp.py'),
     RegrTest('test_zipfile.py'),
     RegrTest('test_zipfile64.py'),
     RegrTest('test_zipimport.py', usemodules='zlib zipimport'),
@@ -516,16 +519,28 @@ def pytest_configure(config):
     for x in testmap:
         cache[x.basename] = x
 
-def pytest_collect_file(path, parent, __multicall__):
-    # don't collect files except through this hook
-    # implemented by clearing the list of to-be-called
-    # remaining hook methods
-    __multicall__.methods[:] = []
-    regrtest = parent.config._basename2spec.get(path.basename, None)
-    if regrtest is None:
-        return
-    if path.dirpath() != testdir:
-        return
+def pytest_ignore_collect(path, config):
+    if path.basename == '__init__.py':
+        return False
+    if path.isfile():
+        regrtest = config._basename2spec.get(path.basename, None)
+        if regrtest is None or path.dirpath() != testdir:
+            return True
+
+def pytest_collect_file(path, parent):
+    if path.basename == '__init__.py':
+        # handle the RegrTest for the whole subpackage here
+        pkg_path = path.dirpath()
+        regrtest = parent.config._basename2spec.get(pkg_path.basename, None)
+        if pkg_path.dirpath() == testdir and regrtest:
+            return RunFileExternal(
+                pkg_path.basename, parent=parent, regrtest=regrtest)
+
+
+@pytest.hookimpl(tryfirst=True)
+def pytest_pycollect_makemodule(path, parent):
+    config = parent.config
+    regrtest = config._basename2spec[path.basename]
     return RunFileExternal(path.basename, parent=parent, regrtest=regrtest)
 
 class RunFileExternal(py.test.collect.File):
@@ -560,7 +575,7 @@ class ReallyRunFileExternal(py.test.collect.Item):
             watchdog_name = 'watchdog_nt.py'
         else:
             watchdog_name = 'watchdog.py'
-        watchdog_script = rpythondir.join('tool', watchdog_name)
+        watchdog_script = pypydir.join('tool', watchdog_name)
 
         regr_script = pypydir.join('tool', 'pytest',
                                    'run-script', 'regrverbose.py')
@@ -583,7 +598,7 @@ class ReallyRunFileExternal(py.test.collect.Item):
                     py.test.skip("%s module not included in %s" % (mod,
                                                                    execpath))
 
-            cmd = "%s %s %s" % (execpath, regrrun, fspath.purebasename)
+            cmd = "%s -m test -v %s" % (execpath, fspath.purebasename)
             # add watchdog for timing out
             cmd = "%s %s %s %s" % (python, watchdog_script, TIMEOUT, cmd)
         else:
