@@ -49,10 +49,13 @@ class _CDataMeta(type):
         else:
             return self.from_param(as_parameter)
 
+    def _build_ffiargtype(self):
+        return _shape_to_ffi_type(self._ffiargshape_)
+
     def get_ffi_argtype(self):
         if self._ffiargtype:
             return self._ffiargtype
-        self._ffiargtype = _shape_to_ffi_type(self._ffiargshape)
+        self._ffiargtype = self._build_ffiargtype()
         return self._ffiargtype
 
     def _CData_output(self, resbuffer, base=None, index=-1):
@@ -64,8 +67,9 @@ class _CDataMeta(type):
         res = object.__new__(self)
         res.__class__ = self
         res.__dict__['_buffer'] = resbuffer
-        res.__dict__['_base'] = base
-        res.__dict__['_index'] = index
+        if base is not None:
+            res.__dict__['_base'] = base
+            res.__dict__['_index'] = index
         return res
 
     def _CData_retval(self, resbuffer):
@@ -81,38 +85,47 @@ class _CDataMeta(type):
         return False
 
     def in_dll(self, dll, name):
-        return self.from_address(dll._handle.getaddressindll(name))
+        return self.from_address(dll.__pypy_dll__.getaddressindll(name))
 
     def from_buffer(self, obj, offset=0):
         size = self._sizeofinstances()
-        buf = buffer(obj, offset, size)
-        if len(buf) < size:
+        buf = memoryview(obj)
+        if buf.nbytes < offset + size:
             raise ValueError(
                 "Buffer size too small (%d instead of at least %d bytes)"
-                % (len(buf) + offset, size + offset))
-        raw_addr = buf._pypy_raw_address()
+                % (buf.nbytes, offset + size))
+        raw_addr = buf._pypy_raw_address() + offset
         result = self.from_address(raw_addr)
-        result._ensure_objects()['ffffffff'] = obj
+        objects = result._ensure_objects()
+        if objects is not None:
+            objects['ffffffff'] = obj
+        else:   # case e.g. of a primitive type like c_int
+            result._objects = obj
         return result
 
     def from_buffer_copy(self, obj, offset=0):
         size = self._sizeofinstances()
-        buf = buffer(obj, offset, size)
-        if len(buf) < size:
+        buf = memoryview(obj)
+        if buf.nbytes < offset + size:
             raise ValueError(
                 "Buffer size too small (%d instead of at least %d bytes)"
-                % (len(buf) + offset, size + offset))
-        result = self()
+                % (buf.nbytes, offset + size))
+        result = self._newowninstance_()
         dest = result._buffer.buffer
         try:
-            raw_addr = buf._pypy_raw_address()
+            raw_addr = buf._pypy_raw_address() + offset
         except ValueError:
-            _rawffi.rawstring2charp(dest, buf)
+            _rawffi.rawstring2charp(dest, buf, offset, size)
         else:
             from ctypes import memmove
             memmove(dest, raw_addr, size)
         return result
 
+    def _newowninstance_(self):
+        result = self.__new__(self)
+        result._init_no_arg_()
+        return result
+
 
 class CArgObject(object):
     """ simple wrapper around buffer, just for the case of freeing
@@ -135,15 +148,15 @@ class CArgObject(object):
     def __ne__(self, other):
         return self._obj != other
 
-class _CData(object):
+class _CData(object, metaclass=_CDataMeta):
     """ The most basic object for all ctypes types
     """
-    __metaclass__ = _CDataMeta
     _objects = None
     _ffiargtype = None
 
     def __init__(self, *args, **kwds):
         raise TypeError("%s has no type" % (type(self),))
+    _init_no_arg_ = __init__
 
     def _ensure_objects(self):
         if '_objects' not in self.__dict__:
@@ -161,14 +174,18 @@ class _CData(object):
     def _get_buffer_value(self):
         return self._buffer[0]
 
+    def _copy_to(self, addr):
+        target = type(self).from_address(addr)._buffer
+        target[0] = self._get_buffer_value()
+
     def _to_ffi_param(self):
         if self.__class__._is_pointer_like():
             return self._get_buffer_value()
         else:
             return self.value
 
-    def __buffer__(self):
-        return buffer(self._buffer)
+    def __buffer__(self, flags):
+        return memoryview(self._buffer)
 
     def _get_b_base(self):
         try:
@@ -199,14 +216,17 @@ def alignment(tp):
     return tp._alignmentofinstances()
 
 @builtinify
-def byref(cdata):
+def byref(cdata, offset=0):
     # "pointer" is imported at the end of this module to avoid circular
     # imports
-    return pointer(cdata)
+    ptr = pointer(cdata)
+    if offset != 0:
+        ptr._buffer[0] += offset
+    return ptr
 
 def cdata_from_address(self, address):
     # fix the address: turn it into as unsigned, in case it's a negative number
-    address = address & (sys.maxint * 2 + 1)
+    address = address & (sys.maxsize * 2 + 1)
     instance = self.__new__(self)
     lgt = getattr(self, '_length_', 1)
     instance._buffer = self._ffiarray.fromaddress(address, lgt)
