@@ -36,9 +36,25 @@ def names_and_fields(self, _fields_, superclass, anonymous_fields=None):
     rawfields = []
     for f in all_fields:
         if len(f) > 2:
-            rawfields.append((f[0], f[1]._ffishape, f[2]))
+            rawfields.append((f[0], f[1]._ffishape_, f[2]))
         else:
-            rawfields.append((f[0], f[1]._ffishape))
+            rawfields.append((f[0], f[1]._ffishape_))
+
+    # hack for duplicate field names
+    already_seen = set()
+    names1 = names
+    names = []
+    for f in names1:
+        if f not in already_seen:
+            names.append(f)
+            already_seen.add(f)
+    already_seen = set()
+    for i in reversed(range(len(rawfields))):
+        if rawfields[i][0] in already_seen:
+            rawfields[i] = (('$DUP%d$%s' % (i, rawfields[i][0]),)
+                            + rawfields[i][1:])
+        already_seen.add(rawfields[i][0])
+    # /hack
 
     _set_shape(self, rawfields, self._is_union)
 
@@ -48,8 +64,8 @@ def names_and_fields(self, _fields_, superclass, anonymous_fields=None):
         value = field[1]
         is_bitfield = (len(field) == 3)
         fields[name] = Field(name,
-                             self._ffistruct.fieldoffset(name),
-                             self._ffistruct.fieldsize(name),
+                             self._ffistruct_.fieldoffset(name),
+                             self._ffistruct_.fieldsize(name),
                              value, i, is_bitfield)
 
     if anonymous_fields:
@@ -58,32 +74,35 @@ def names_and_fields(self, _fields_, superclass, anonymous_fields=None):
             name = field[0]
             value = field[1]
             is_bitfield = (len(field) == 3)
-            startpos = self._ffistruct.fieldoffset(name)
+            startpos = self._ffistruct_.fieldoffset(name)
             if name in anonymous_fields:
-                for subname in value._names:
+                for subname in value._names_:
                     resnames.append(subname)
                     subfield = getattr(value, subname)
                     relpos = startpos + subfield.offset
                     subvalue = subfield.ctype
                     fields[subname] = Field(subname,
                                             relpos, subvalue._sizeofinstances(),
-                                            subvalue, i, is_bitfield)
+                                            subvalue, i, is_bitfield,
+                                            inside_anon_field=fields[name])
             else:
                 resnames.append(name)
         names = resnames
-    self._names = names
+    self._names_ = names
     for name, field in fields.items():
         setattr(self, name, field)
 
 
 class Field(object):
-    def __init__(self, name, offset, size, ctype, num, is_bitfield):
+    def __init__(self, name, offset, size, ctype, num, is_bitfield,
+                 inside_anon_field=None):
         self.__dict__['name'] = name
         self.__dict__['offset'] = offset
         self.__dict__['size'] = size
         self.__dict__['ctype'] = ctype
         self.__dict__['num'] = num
         self.__dict__['is_bitfield'] = is_bitfield
+        self.__dict__['inside_anon_field'] = inside_anon_field
 
     def __setattr__(self, name, value):
         raise AttributeError(name)
@@ -95,6 +114,8 @@ class Field(object):
     def __get__(self, obj, cls=None):
         if obj is None:
             return self
+        if self.inside_anon_field is not None:
+            return getattr(self.inside_anon_field.__get__(obj), self.name)
         if self.is_bitfield:
             # bitfield member, use direct access
             return obj._buffer.__getattr__(self.name)
@@ -105,6 +126,9 @@ class Field(object):
             return fieldtype._CData_output(suba, obj, offset)
 
     def __set__(self, obj, value):
+        if self.inside_anon_field is not None:
+            setattr(self.inside_anon_field.__get__(obj), self.name, value)
+            return
         fieldtype = self.ctype
         cobj = fieldtype.from_param(value)
         key = keepalive_key(self.num)
@@ -114,19 +138,20 @@ class Field(object):
         elif ensure_objects(cobj) is not None:
             store_reference(obj, key, cobj._objects)
         arg = cobj._get_buffer_value()
-        if fieldtype._fficompositesize is not None:
+        if fieldtype._fficompositesize_ is not None:
             from ctypes import memmove
             dest = obj._buffer.fieldaddress(self.name)
-            memmove(dest, arg, fieldtype._fficompositesize)
+            memmove(dest, arg, fieldtype._fficompositesize_)
         else:
             obj._buffer.__setattr__(self.name, arg)
 
 
+
 def _set_shape(tp, rawfields, is_union=False):
-    tp._ffistruct = _rawffi.Structure(rawfields, is_union,
+    tp._ffistruct_ = _rawffi.Structure(rawfields, is_union,
                                       getattr(tp, '_pack_', 0))
-    tp._ffiargshape = tp._ffishape = (tp._ffistruct, 1)
-    tp._fficompositesize = tp._ffistruct.size
+    tp._ffiargshape_ = tp._ffishape_ = (tp._ffistruct_, 1)
+    tp._fficompositesize_ = tp._ffistruct_.size
 
 
 def struct_setattr(self, name, value):
@@ -135,6 +160,10 @@ def struct_setattr(self, name, value):
             raise AttributeError("_fields_ is final")
         if self in [f[1] for f in value]:
             raise AttributeError("Structure or union cannot contain itself")
+        if self._ffiargtype is not None:
+            raise NotImplementedError("Too late to set _fields_: we already "
+                        "said to libffi that the structure type %s is opaque"
+                        % (self,))
         names_and_fields(
             self,
             value, self.__bases__[0],
@@ -180,30 +209,30 @@ class StructOrUnionMeta(_CDataMeta):
         if isinstance(address, _rawffi.StructureInstance):
             address = address.buffer
         # fix the address: turn it into as unsigned, in case it is negative
-        address = address & (sys.maxint * 2 + 1)
-        instance.__dict__['_buffer'] = self._ffistruct.fromaddress(address)
+        address = address & (sys.maxsize * 2 + 1)
+        instance.__dict__['_buffer'] = self._ffistruct_.fromaddress(address)
         return instance
 
     def _sizeofinstances(self):
-        if not hasattr(self, '_ffistruct'):
+        if not hasattr(self, '_ffistruct_'):
             return 0
-        return self._ffistruct.size
+        return self._ffistruct_.size
 
     def _alignmentofinstances(self):
-        return self._ffistruct.alignment
+        return self._ffistruct_.alignment
 
     def from_param(self, value):
         if isinstance(value, tuple):
             try:
                 value = self(*value)
-            except Exception, e:
+            except Exception as e:
                 # XXX CPython does not even respect the exception type
                 raise RuntimeError("(%s) %s: %s" % (self.__name__, type(e), e))
         return _CDataMeta.from_param(self, value)
 
     def _CData_output(self, resarray, base=None, index=-1):
         res = StructOrUnion.__new__(self)
-        ffistruct = self._ffistruct.fromaddress(resarray.buffer)
+        ffistruct = self._ffistruct_.fromaddress(resarray.buffer)
         res.__dict__['_buffer'] = ffistruct
         res.__dict__['_base'] = base
         res.__dict__['_index'] = index
@@ -216,35 +245,48 @@ class StructOrUnionMeta(_CDataMeta):
         res.__dict__['_index'] = -1
         return res
 
-
-class StructOrUnion(_CData):
-    __metaclass__ = StructOrUnionMeta
+class StructOrUnion(_CData, metaclass=StructOrUnionMeta):
 
     def __new__(cls, *args, **kwds):
-        self = super(_CData, cls).__new__(cls, *args, **kwds)
-        if '_abstract_' in cls.__dict__:
+        from _ctypes import union
+        if ('_abstract_' in cls.__dict__ or cls is Structure
+                                         or cls is union.Union):
             raise TypeError("abstract class")
-        if hasattr(cls, '_ffistruct'):
-            self.__dict__['_buffer'] = self._ffistruct(autofree=True)
+        if hasattr(cls, '_swappedbytes_'):
+            fields = [None] * len(cls._fields_)
+            for i in range(len(cls._fields_)):
+                if cls._fields_[i][1] == cls._fields_[i][1].__dict__.get('__ctype_be__', None):
+                    swapped = cls._fields_[i][1].__dict__.get('__ctype_le__', cls._fields_[i][1])
+                else:
+                    swapped = cls._fields_[i][1].__dict__.get('__ctype_be__', cls._fields_[i][1])
+                if len(cls._fields_[i]) < 3:
+                    fields[i] = (cls._fields_[i][0], swapped)
+                else:
+                    fields[i] = (cls._fields_[i][0], swapped, cls._fields_[i][2])
+            names_and_fields(cls, fields, _CData, cls.__dict__.get('_anonymous_', None))
+        self = super(_CData, cls).__new__(cls)
+        if hasattr(cls, '_ffistruct_'):
+            self.__dict__['_buffer'] = self._ffistruct_(autofree=True)
         return self
 
     def __init__(self, *args, **kwds):
         type(self)._make_final()
-        if len(args) > len(self._names):
+        if len(args) > len(self._names_):
             raise TypeError("too many initializers")
-        for name, arg in zip(self._names, args):
+        for name, arg in zip(self._names_, args):
             if name in kwds:
                 raise TypeError("duplicate value for argument %r" % (
                     name,))
             self.__setattr__(name, arg)
         for name, arg in kwds.items():
             self.__setattr__(name, arg)
+    _init_no_arg_ = __init__
 
     def _subarray(self, fieldtype, name):
         """Return a _rawffi array of length 1 whose address is the same as
         the address of the field 'name' of self."""
         address = self._buffer.fieldaddress(name)
-        A = _rawffi.Array(fieldtype._ffishape)
+        A = _rawffi.Array(fieldtype._ffishape_)
         return A.fromaddress(address, 1)
 
     def _get_buffer_for_param(self):
@@ -253,6 +295,11 @@ class StructOrUnion(_CData):
     def _get_buffer_value(self):
         return self._buffer.buffer
 
+    def _copy_to(self, addr):
+        from ctypes import memmove
+        origin = self._get_buffer_value()
+        memmove(addr, origin, self._fficompositesize_)
+
     def _to_ffi_param(self):
         return self._buffer
 
@@ -261,5 +308,5 @@ class StructureMeta(StructOrUnionMeta):
     _is_union = False
 
 
-class Structure(StructOrUnion):
-    __metaclass__ = StructureMeta
+class Structure(StructOrUnion, metaclass=StructureMeta):
+    pass
