@@ -1,7 +1,8 @@
 import sys
+import __pypy__
 import _continuation
 
-__version__ = "0.4.5"
+__version__ = "0.4.13"
 
 # ____________________________________________________________
 # Exceptions
@@ -79,8 +80,8 @@ class greenlet(_continulet):
             # convert a "raise GreenletExit" into "return GreenletExit"
             if methodname == 'throw':
                 try:
-                    raise baseargs[0], baseargs[1]
-                except GreenletExit, e:
+                    raise __pypy__.normalize_exc(baseargs[0], baseargs[1])
+                except GreenletExit as e:
                     methodname = 'switch'
                     baseargs = (((e,), {}),)
                 except:
@@ -88,9 +89,19 @@ class greenlet(_continulet):
         #
         try:
             unbound_method = getattr(_continulet, methodname)
+            _tls.leaving = current
             args, kwds = unbound_method(current, *baseargs, to=target)
-        finally:
             _tls.current = current
+        except:
+            _tls.current = current
+            if hasattr(_tls, 'trace'):
+                _run_trace_callback('throw')
+            _tls.leaving = None
+            raise
+        else:
+            if hasattr(_tls, 'trace'):
+                _run_trace_callback('switch')
+            _tls.leaving = None
         #
         if kwds:
             if args:
@@ -101,7 +112,7 @@ class greenlet(_continulet):
         else:
             return args
 
-    def __nonzero__(self):
+    def __bool__(self):
         return self.__main or _continulet.is_pending(self)
 
     @property
@@ -116,11 +127,39 @@ class greenlet(_continulet):
             return None
         if self.__main:
             self = getcurrent()
-        f = _continulet.__reduce__(self)[2][0]
+        f = self._get_frame()
         if not f:
             return None
         return f.f_back.f_back.f_back   # go past start(), __switch(), switch()
 
+# ____________________________________________________________
+# Recent additions
+
+GREENLET_USE_GC = True
+GREENLET_USE_TRACING = True
+
+def gettrace():
+    return getattr(_tls, 'trace', None)
+
+def settrace(callback):
+    try:
+        prev = _tls.trace
+        del _tls.trace
+    except AttributeError:
+        prev = None
+    if callback is not None:
+        _tls.trace = callback
+    return prev
+
+def _run_trace_callback(event):
+    try:
+        _tls.trace(event, (_tls.leaving, _tls.current))
+    except:
+        # In case of exceptions trace function is removed
+        if hasattr(_tls, 'trace'):
+            del _tls.trace
+        raise
+
 # ____________________________________________________________
 # Internal stuff
 
@@ -143,22 +182,32 @@ def _green_create_main():
     _tls.current = gmain
 
 def _greenlet_start(greenlet, args):
-    args, kwds = args
-    _tls.current = greenlet
     try:
-        res = greenlet.run(*args, **kwds)
-    except GreenletExit, e:
-        res = e
+        args, kwds = args
+        _tls.current = greenlet
+        try:
+            if hasattr(_tls, 'trace'):
+                _run_trace_callback('switch')
+            res = greenlet.run(*args, **kwds)
+        except GreenletExit as e:
+            res = e
+        finally:
+            _continuation.permute(greenlet, greenlet.parent)
+        return ((res,), None)
     finally:
-        _continuation.permute(greenlet, greenlet.parent)
-    return ((res,), None)
+        _tls.leaving = greenlet
 
 def _greenlet_throw(greenlet, exc, value, tb):
-    _tls.current = greenlet
     try:
-        raise exc, value, tb
-    except GreenletExit, e:
-        res = e
+        _tls.current = greenlet
+        try:
+            if hasattr(_tls, 'trace'):
+                _run_trace_callback('throw')
+            raise __pypy__.normalize_exc(exc, value, tb)
+        except GreenletExit as e:
+            res = e
+        finally:
+            _continuation.permute(greenlet, greenlet.parent)
+        return ((res,), None)
     finally:
-        _continuation.permute(greenlet, greenlet.parent)
-    return ((res,), None)
+        _tls.leaving = greenlet
