@@ -60,8 +60,8 @@ def build_menu(cons, wordlist, start, use_brackets, sort_in_column):
         item = "%s  "
         padding = 2
     maxlen = min(max(map(real_len, wordlist)), cons.width - padding)
-    cols = cons.width / (maxlen + padding)
-    rows = (len(wordlist) - 1)/cols + 1
+    cols = cons.width // (maxlen + padding)
+    rows = (len(wordlist) - 1)//cols + 1
 
     if sort_in_column:
         # sort_in_column=False (default)     sort_in_column=True
@@ -161,7 +161,7 @@ class complete(commands.Command):
             r.insert(completions[0][len(stem):])
         else:
             p = prefix(completions, len(stem))
-            if p <> '':
+            if p:
                 r.insert(p)
             if r.last_command_is(self.__class__):
                 if not r.cmpltn_menu_vis:
@@ -252,7 +252,7 @@ class CompletingReader(Reader):
         p = self.pos - 1
         while p >= 0 and st.get(b[p], SW) == SW:
             p -= 1
-        return u''.join(b[p+1:self.pos])
+        return ''.join(b[p+1:self.pos])
 
     def get_completions(self, stem):
         return []
