@@ -38,7 +38,11 @@ def get_merged_branches(path, startrev, endrev, current_branch=None):
                                    current_branch, current_branch)
     cmd = r'hg log -R "%s" -r "%s" --template "{branches}\n"' % (path, revset)
     out = getoutput(cmd)
-    branches = set(map(str.strip, out.splitlines()))
+    branches = set()
+    for item in out.splitlines():
+        item = item.strip()
+        if not item.startswith('release-'):
+            branches.add(item)
     branches.discard("default")
     return branches, current_branch
 
@@ -81,10 +85,11 @@ def test_whatsnew():
     #whatsnew_list = doc.listdir('whatsnew-*.rst')
     #whatsnew_list.sort()
     #last_whatsnew = whatsnew_list[-1].read()
-    last_whatsnew = doc.join('whatsnew-head.rst').read()
+    last_whatsnew = doc.join('whatsnew-pypy3-head.rst').read()
     startrev, documented = parse_doc(last_whatsnew)
     merged, branch = get_merged_branches(ROOT, startrev, '')
     merged.discard('default')
+    merged.discard('py3.5')
     merged.discard('')
     not_documented = merged.difference(documented)
     not_merged = documented.difference(merged)
@@ -95,5 +100,17 @@ def test_whatsnew():
     print '\n'.join(not_merged)
     print
     assert not not_documented
-    if branch == 'default':
+    if branch == 'py3.5':
         assert not not_merged
+    else:
+        assert branch in documented, 'Please document this branch before merging: %s' % branch
+
+def test_startrev_on_default():
+    doc = ROOT.join('pypy', 'doc')
+    last_whatsnew = doc.join('whatsnew-head.rst').read()
+    startrev, documented = parse_doc(last_whatsnew)
+    errcode, wc_branch = getstatusoutput(
+        "hg log -r %s --template '{branch}'" % startrev)
+    if errcode != 0:
+        py.test.skip('no Mercurial repo')
+    assert wc_branch in ('default', "'default'") # sometimes the ' leaks (windows)
