@@ -6,8 +6,8 @@ Compiler instances are stored into 'space.getexecutioncontext().compiler'.
 from pypy.interpreter import pycode
 from pypy.interpreter.pyparser import future, pyparse, error as parseerror
 from pypy.interpreter.astcompiler import (astbuilder, codegen, consts, misc,
-                                          optimize, ast)
-from pypy.interpreter.error import OperationError
+                                          optimize, ast, validate)
+from pypy.interpreter.error import OperationError, oefmt
 
 
 class AbstractCompiler(object):
@@ -55,21 +55,21 @@ class AbstractCompiler(object):
         try:
             code = self.compile(source, filename, mode, flags)
             return code   # success
-        except OperationError, err:
+        except OperationError as err:
             if not err.match(space, space.w_SyntaxError):
                 raise
 
         try:
             self.compile(source + "\n", filename, mode, flags)
             return None   # expect more
-        except OperationError, err1:
+        except OperationError as err1:
             if not err1.match(space, space.w_SyntaxError):
                 raise
 
         try:
             self.compile(source + "\n\n", filename, mode, flags)
             raise     # uh? no error with \n\n.  re-raise the previous error
-        except OperationError, err2:
+        except OperationError as err2:
             if not err2.match(space, space.w_SyntaxError):
                 raise
 
@@ -101,12 +101,12 @@ class PythonAstCompiler(PyCodeCompiler):
     """
     def __init__(self, space, override_version=None):
         PyCodeCompiler.__init__(self, space)
-        self.future_flags = future.futureFlags_2_7
+        self.future_flags = future.futureFlags_3_5
         self.parser = pyparse.PythonParser(space, self.future_flags)
         self.additional_rules = {}
         self.compiler_flags = self.future_flags.allowed_flags
 
-    def compile_ast(self, node, filename, mode, flags):
+    def compile_ast(self, node, filename, mode, flags, optimize=-1):
         if mode == 'eval':
             check = isinstance(node, ast.Expression)
         elif mode == 'exec':
@@ -116,14 +116,14 @@ class PythonAstCompiler(PyCodeCompiler):
         else:
             check = True
         if not check:
-            raise OperationError(self.space.w_TypeError, self.space.wrap(
-                "invalid node type"))
+            raise oefmt(self.space.w_TypeError, "invalid node type")
 
         fut = misc.parse_future(node, self.future_flags.compiler_features)
         f_flags, f_lineno, f_col = fut
         future_pos = f_lineno, f_col
         flags |= f_flags
-        info = pyparse.CompileInfo(filename, mode, flags, future_pos)
+        info = pyparse.CompileInfo(filename, mode, flags, future_pos,
+                optimize=optimize)
         return self._compile_ast(node, info)
 
     def _compile_ast(self, node, info):
@@ -131,11 +131,17 @@ class PythonAstCompiler(PyCodeCompiler):
         try:
             mod = optimize.optimize_ast(space, node, info)
             code = codegen.compile_ast(space, mod, info)
-        except parseerror.SyntaxError, e:
-            raise OperationError(space.w_SyntaxError,
-                                 e.wrap_info(space))
+        except parseerror.SyntaxError as e:
+            raise OperationError(space.w_SyntaxError, e.wrap_info(space))
         return code
 
+    def validate_ast(self, node):
+        try:
+            validate.validate_ast(self.space, node)
+        except validate.ValidationError as e:
+            raise OperationError(self.space.w_ValueError,
+                                 self.space.newtext(e.message))
+
     def compile_to_ast(self, source, filename, mode, flags):
         info = pyparse.CompileInfo(filename, mode, flags)
         return self._compile_to_ast(source, info)
@@ -144,17 +150,20 @@ class PythonAstCompiler(PyCodeCompiler):
         space = self.space
         try:
             parse_tree = self.parser.parse_source(source, info)
-            mod = astbuilder.ast_from_node(space, parse_tree, info)
-        except parseerror.IndentationError, e:
-            raise OperationError(space.w_IndentationError,
-                                 e.wrap_info(space))
-        except parseerror.SyntaxError, e:
-            raise OperationError(space.w_SyntaxError,
+            mod = astbuilder.ast_from_node(space, parse_tree, info,
+                                           recursive_parser=self.parser)
+        except parseerror.TabError as e:
+            raise OperationError(space.w_TabError,
                                  e.wrap_info(space))
+        except parseerror.IndentationError as e:
+            raise OperationError(space.w_IndentationError, e.wrap_info(space))
+        except parseerror.SyntaxError as e:
+            raise OperationError(space.w_SyntaxError, e.wrap_info(space))
         return mod
 
-    def compile(self, source, filename, mode, flags, hidden_applevel=False):
+    def compile(self, source, filename, mode, flags, hidden_applevel=False,
+            optimize=-1):
         info = pyparse.CompileInfo(filename, mode, flags,
-                                   hidden_applevel=hidden_applevel)
+                hidden_applevel=hidden_applevel, optimize=optimize)
         mod = self._compile_to_ast(source, info)
         return self._compile_ast(mod, info)
