@@ -4,6 +4,7 @@ from __future__ import division, print_function  # for test_app2interp_future
 from pypy.interpreter import gateway, argument
 from pypy.interpreter.gateway import ObjSpace, W_Root, WrappedDefault
 from pypy.interpreter.signature import Signature
+from pypy.interpreter.error import OperationError
 import py
 import sys
 
@@ -13,10 +14,11 @@ class FakeFunc(object):
         self.space = space
         self.name = name
         self.defs_w = []
+        self.w_kw_defs = None
 
 
 class TestBuiltinCode:
-    def test_signature(self):
+    def test_signature(self, space):
         def c(space, w_x, w_y, hello_w):
             pass
         code = gateway.BuiltinCode(c, unwrap_spec=[gateway.ObjSpace,
@@ -46,6 +48,14 @@ class TestBuiltinCode:
         code = gateway.BuiltinCode(f, unwrap_spec=[gateway.ObjSpace, "index"])
         assert code.signature() == Signature(["index"], None, None)
 
+        def f(space, __kwonly__, w_x):
+            pass
+        code = gateway.BuiltinCode(f, unwrap_spec=[gateway.ObjSpace,
+                                                   "kwonly", W_Root])
+        assert code.signature() == Signature([], kwonlyargnames=['x'])
+        assert space.int_w(space.getattr(
+            code, space.newtext('co_kwonlyargcount'))) == 1
+
 
     def test_call(self):
         def c(space, w_x, w_y, hello_w):
@@ -248,6 +258,31 @@ class TestGateway:
         assert self.space.eq_w(space.call_function(w_app_g, space.wrap(True)),
                                space.wrap(True))
 
+    def test_interp2app_unwrap_spec_bytes(self):
+        # we can't use the "bytes" object for the unwrap_spec, because that's
+        # an alias for "str" on the underlying Python2
+        space = self.space
+        def g(space, b):
+            return space.newbytes(b)
+        app_g = gateway.interp2app(g, unwrap_spec=[gateway.ObjSpace, 'bytes'])
+        app_g2 = gateway.interp2app(g, unwrap_spec=[gateway.ObjSpace, 'bytes'])
+        assert app_g is app_g2
+        w_app_g = space.wrap(app_g)
+        assert self.space.eq_w(space.call_function(w_app_g, space.newbytes("abc")),
+                               space.newbytes("abc"))
+
+    def test_interp2app_unwrap_spec_text(self):
+        space = self.space
+        def g(space, b):
+            assert isinstance(b, str)
+            return space.newtext(b)
+        app_g = gateway.interp2app(g, unwrap_spec=[gateway.ObjSpace, 'text'])
+        app_g2 = gateway.interp2app(g, unwrap_spec=[gateway.ObjSpace, 'text'])
+        assert app_g is app_g2
+        w_app_g = space.wrap(app_g)
+        assert self.space.eq_w(space.call_function(w_app_g, space.newtext("abc")),
+                               space.newtext("abc"))
+
     def test_caching_methods(self):
         class Base(gateway.W_Root):
             def f(self):
@@ -351,7 +386,7 @@ class TestGateway:
             return space.wrap(s0+s1)
         app_g3_ss = gateway.interp2app_temp(g3_ss,
                                          unwrap_spec=[gateway.ObjSpace,
-                                                      str, 'str_or_None'])
+                                                      'text', 'text_or_none'])
         w_app_g3_ss = space.wrap(app_g3_ss)
         assert self.space.eq_w(
             space.call(w_app_g3_ss,
@@ -466,6 +501,17 @@ class TestGateway:
                        w_app_g3_idx,
                        space.mul(space.wrap(sys.maxint), space.wrap(-7)))
 
+    def test_interp2app_unwrap_spec_fsencode(self):
+        space = self.space
+        w = space.wrap
+        def f(filename):
+            return space.newbytes(filename)
+        app_f = gateway.interp2app_temp(f, unwrap_spec=['fsencode'])
+        w_app_f = space.wrap(app_f)
+        assert space.eq_w(
+            space.call_function(w_app_f, w(u'\udc80')),
+            space.newbytes('\x80'))
+
     def test_interp2app_unwrap_spec_typechecks(self):
         from rpython.rlib.rarithmetic import r_longlong
 
@@ -479,19 +525,19 @@ class TestGateway:
         w_app_g3_i = space.wrap(app_g3_i)
         assert space.eq_w(space.call_function(w_app_g3_i,w(1)),w(1))
         assert space.eq_w(space.call_function(w_app_g3_i,w(1L)),w(1))
-        raises(gateway.OperationError,space.call_function,w_app_g3_i,w(sys.maxint*2))
-        raises(gateway.OperationError,space.call_function,w_app_g3_i,w(None))
-        raises(gateway.OperationError,space.call_function,w_app_g3_i,w("foo"))
-        raises(gateway.OperationError,space.call_function,w_app_g3_i,w(1.0))
+        space.raises_w(space.w_OverflowError, space.call_function,w_app_g3_i,w(sys.maxint*2))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_i,w(None))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_i,w("foo"))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_i,w(1.0))
 
         app_g3_s = gateway.interp2app_temp(g3_id,
                                          unwrap_spec=[gateway.ObjSpace,
-                                                      str])
+                                                      'text'])
         w_app_g3_s = space.wrap(app_g3_s)
         assert space.eq_w(space.call_function(w_app_g3_s,w("foo")),w("foo"))
-        raises(gateway.OperationError,space.call_function,w_app_g3_s,w(None))
-        raises(gateway.OperationError,space.call_function,w_app_g3_s,w(1))
-        raises(gateway.OperationError,space.call_function,w_app_g3_s,w(1.0))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_s,w(None))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_s,w(1))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_s,w(1.0))
 
         app_g3_f = gateway.interp2app_temp(g3_id,
                                          unwrap_spec=[gateway.ObjSpace,
@@ -500,14 +546,14 @@ class TestGateway:
         assert space.eq_w(space.call_function(w_app_g3_f,w(1.0)),w(1.0))
         assert space.eq_w(space.call_function(w_app_g3_f,w(1)),w(1.0))
         assert space.eq_w(space.call_function(w_app_g3_f,w(1L)),w(1.0))
-        raises(gateway.OperationError,space.call_function,w_app_g3_f,w(None))
-        raises(gateway.OperationError,space.call_function,w_app_g3_f,w("foo"))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_f,w(None))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_f,w("foo"))
 
         app_g3_r = gateway.interp2app_temp(g3_id,
                                            unwrap_spec=[gateway.ObjSpace,
                                                         r_longlong])
         w_app_g3_r = space.wrap(app_g3_r)
-        raises(gateway.OperationError,space.call_function,w_app_g3_r,w(1.0))
+        space.raises_w(space.w_TypeError, space.call_function,w_app_g3_r,w(1.0))
 
     def test_interp2app_unwrap_spec_unicode(self):
         space = self.space
@@ -524,9 +570,9 @@ class TestGateway:
         assert self.space.eq_w(
             space.call_function(w_app_g3_u, w("baz")),
             w(3))
-        raises(gateway.OperationError, space.call_function, w_app_g3_u,
+        space.raises_w(space.w_TypeError, space.call_function, w_app_g3_u,
                w(None))
-        raises(gateway.OperationError, space.call_function, w_app_g3_u,
+        space.raises_w(space.w_TypeError, space.call_function, w_app_g3_u,
                w(42))
 
     def test_interp2app_unwrap_spec_unwrapper(self):
@@ -543,14 +589,14 @@ class TestGateway:
                                                       Unwrapper])
         assert self.space.eq_w(
             space.call_function(w(app_g3_u), w(42)), w(43))
-        raises(gateway.OperationError, space.call_function,
+        space.raises_w(space.w_TypeError, space.call_function,
                w(app_g3_u), w(None))
 
     def test_interp2app_classmethod(self):
         space = self.space
         w = space.wrap
         def g_run(space, w_type):
-            assert space.is_w(w_type, space.w_str)
+            assert space.is_w(w_type, space.w_text)
             return w(42)
 
         app_g_run = gateway.interp2app_temp(g_run,
@@ -558,7 +604,7 @@ class TestGateway:
                                                          gateway.W_Root],
                                             as_classmethod=True)
         w_app_g_run = space.wrap(app_g_run)
-        w_bound = space.get(w_app_g_run, w("hello"), space.w_str)
+        w_bound = space.get(w_app_g_run, w("hello"), space.w_text)
         assert space.eq_w(space.call_function(w_bound), w(42))
 
     def test_interp2app_fastcall(self):
@@ -741,7 +787,7 @@ class TestGateway:
         @gateway.unwrap_spec(w_x = WrappedDefault(42), y=int)
         def g(space, w_x, y):
             never_called
-        py.test.raises(AssertionError, space.wrap, gateway.interp2app_temp(g))
+        py.test.raises(KeyError, space.wrap, gateway.interp2app_temp(g))
 
     def test_unwrap_spec_default_applevel_bug2(self):
         space = self.space
@@ -771,6 +817,147 @@ class TestGateway:
         w_g = space.wrap(gateway.interp2app_temp(g, doc='bar'))
         assert space.unwrap(space.getattr(w_g, space.wrap('__doc__'))) == 'bar'
 
+    def test_system_error(self):
+        py.test.skip("we don't wrap a random exception inside SystemError "
+                     "when untranslated, because it makes testing harder")
+        class UnexpectedException(Exception):
+            pass
+        space = self.space
+        def g(space):
+            raise UnexpectedException
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        e = py.test.raises(OperationError, space.appexec, [w_g], """(my_g):
+            my_g()
+        """)
+        err = str(e.value)
+        assert 'SystemError' in err
+        assert ('unexpected internal exception (please '
+                'report a bug): UnexpectedException') in err
+
+    def test_system_error_2(self):
+        py.test.skip("we don't wrap a random exception inside SystemError "
+                     "when untranslated, because it makes testing harder")
+        class UnexpectedException(Exception):
+            pass
+        space = self.space
+        def g(space):
+            raise UnexpectedException
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        w_msg = space.appexec([w_g], """(my_g):
+            try:
+                my_g()
+            except SystemError as e:
+                return str(e)
+        """)
+        err = space.str_w(w_msg)
+        assert ('unexpected internal exception (please '
+                'report a bug): UnexpectedException') in err
+
+    def test_bare_raise_in_app_helper(self):
+        space = self.space
+        w = space.wrap
+        def app_g3(a):
+            try:
+                1 / a
+            except ZeroDivisionError:
+                raise
+        g3 = gateway.app2interp(app_g3)
+        space.raises_w(space.w_ZeroDivisionError, g3, space, w(0))
+
+    def test_unwrap_spec_default_bytes(self):
+        space = self.space
+        @gateway.unwrap_spec(s='bufferstr')
+        def g(space, s=''):
+            return space.wrap(type(s) is str)
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        args = argument.Arguments(space, [])
+        w_res = space.call_args(w_g, args)
+        assert space.eq_w(w_res, space.w_True)
+
+    def test_unwrap_spec_default_applevel_bytes(self):
+        space = self.space
+        @gateway.unwrap_spec(w_x=WrappedDefault('foo'))
+        def g(space, w_x):
+            return w_x
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        args = argument.Arguments(space, [])
+        w_res = space.call_args(w_g, args)
+        assert space.eq_w(w_res, space.newbytes('foo'))
+
+    def test_unwrap_spec_kwonly(self):
+        space = self.space
+        def g(space, w_x, __kwonly__, w_y):
+            return space.sub(w_x, w_y)
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        w = space.wrap
+        w1 = w(1)
+
+        for i in range(4):
+            a = argument.Arguments(space, [w1, w1, w1])
+            py.test.raises(gateway.OperationError, space.call_args, w_g, a)
+            py.test.raises(gateway.OperationError, space.call_function, w_g,
+                           *(i * (w1,)))
+
+        args = argument.Arguments(space, [w(1)],
+                                  w_starstararg = w({'y': 10}))
+        assert space.eq_w(space.call_args(w_g, args), w(-9))
+        args = argument.Arguments(space, [],
+                                  w_starstararg = w({'x': 2, 'y': 10}))
+        assert space.eq_w(space.call_args(w_g, args), w(-8))
+
+    def test_unwrap_spec_kwonly_default(self):
+        space = self.space
+        @gateway.unwrap_spec(w_x2=WrappedDefault(50), y2=int)
+        def g(space, w_x1, w_x2, __kwonly__, w_y1, y2=200):
+            return space.sub(space.sub(w_x1, w_x2),
+                             space.sub(w_y1, w(y2)))
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        w = space.wrap
+        w1 = w(1)
+
+        for i in range(6):
+            py.test.raises(gateway.OperationError, space.call_function, w_g,
+                           *(i * (w1,)))
+
+        def expected(x1, x2=50, y1="missing", y2=200):
+            return (x1 - x2) - (y1 - y2)
+
+        def check(*args, **kwds):
+            a = argument.Arguments(space, [], w_stararg = w(args),
+                                          w_starstararg = w(kwds))
+            w_res = space.call_args(w_g, a)
+            assert space.eq_w(w_res, w(expected(*args, **kwds)))
+
+            del kwds['y1']
+            a = argument.Arguments(space, [], w_stararg = w(args),
+                                          w_starstararg = w(kwds))
+            py.test.raises(gateway.OperationError, space.call_args, w_g, a)
+
+            args += (1234,)
+            a = argument.Arguments(space, [], w_stararg = w(args),
+                                          w_starstararg = w(kwds))
+            py.test.raises(gateway.OperationError, space.call_args, w_g, a)
+
+        check(5,       y1=1234)
+        check(5, 1,    y1=1234)
+        check(5, x2=1, y1=1234)
+        check(5,       y1=1234, y2=343)
+        check(5, 1,    y1=1234, y2=343)
+        check(5, x2=1, y1=1234, y2=343)
+        check(x1=5,       y1=1234,       )
+        check(x1=5, x2=1, y1=1234,       )
+        check(x1=5,       y1=1234, y2=343)
+        check(x1=5, x2=1, y1=1234, y2=343)
+
+    def test_unwrap_spec_kwonly_default_2(self):
+        space = self.space
+        @gateway.unwrap_spec(w_x2=WrappedDefault(50))
+        def g(space, w_x2=None):
+            return w_x2
+        w_g = space.wrap(gateway.interp2app_temp(g))
+        w_res = space.call_function(w_g)
+        assert space.eq_w(w_res, space.wrap(50))
+
 
 class AppTestPyTestMark:
     @py.test.mark.unlikely_to_exist
@@ -863,24 +1050,46 @@ class TestPassThroughArguments:
         class A(object):
            m = g # not a builtin function, so works as method
         d = {'A': A}
-        exec \"\"\"
+        exec(\"\"\"
 # own compiler
 a = A()
 y = a.m(33)
-\"\"\" in d
+\"\"\", d)
         return d['y'] == ('g', d['a'], 33)
         """)
         assert space.is_true(w_res)
         assert len(called) == 1
         assert isinstance(called[0], argument.Arguments)
 
+    def test_pass_trough_arguments_method(self):
+        space = self.space
+
+        called = []
+
+        class W_Something(W_Root):
+            def f(self, space, __args__):
+                called.append(__args__)
+                a_w, _ = __args__.unpack()
+                return space.newtuple([space.wrap('f')]+a_w)
+
+        w_f = space.wrap(gateway.interp2app_temp(W_Something.f))
+
+        w_self = space.wrap(W_Something())
+        args = argument.Arguments(space, [space.wrap(7)])
+
+        w_res = space.call_obj_args(w_f, w_self, args)
+        assert space.is_true(space.eq(w_res, space.wrap(('f', 7))))
+
+        # white-box check for opt
+        assert called[0] is args
+
 
 class AppTestKeywordsToBuiltinSanity(object):
     def test_type(self):
         class X(object):
-            def __init__(self, **kw):
+            def __init__(myself, **kw):
                 pass
-        clash = type.__call__.func_code.co_varnames[0]
+        clash = type.__call__.__code__.co_varnames[0]
 
         X(**{clash: 33})
         type.__call__(X, **{clash: 33})
@@ -889,28 +1098,27 @@ class AppTestKeywordsToBuiltinSanity(object):
         class X(object):
             def __init__(self, **kw):
                 pass
-        clash = object.__new__.func_code.co_varnames[0]
+        clash = object.__new__.__code__.co_varnames[0]
 
         X(**{clash: 33})
         object.__new__(X, **{clash: 33})
 
-
     def test_dict_new(self):
-        clash = dict.__new__.func_code.co_varnames[0]
+        clash = dict.__new__.__code__.co_varnames[0]
 
         dict(**{clash: 33})
         dict.__new__(dict, **{clash: 33})
 
     def test_dict_init(self):
         d = {}
-        clash = dict.__init__.func_code.co_varnames[0]
+        clash = dict.__init__.__code__.co_varnames[0]
 
         d.__init__(**{clash: 33})
         dict.__init__(d, **{clash: 33})
 
     def test_dict_update(self):
         d = {}
-        clash = dict.update.func_code.co_varnames[0]
+        clash = dict.update.__code__.co_varnames[0]
 
         d.update(**{clash: 33})
         dict.update(d, **{clash: 33})
