@@ -6,9 +6,9 @@ class TestTargetPyPy(object):
     def test_run(self):
         config = get_pypy_config(translating=False)
         entry_point = get_entry_point(config)[0]
-        entry_point(['pypy-c' , '-S', '-c', 'print 3'])
+        entry_point(['pypy3-c' , '-S', '-c', 'print 3'])
 
-def test_exeucte_source(space):
+def test_execute_source(space):
     _, d = create_entry_point(space, None)
     execute_source = d['pypy_execute_source']
     lls = rffi.str2charp("import sys; sys.modules['xyz'] = 3")
@@ -20,7 +20,7 @@ def test_exeucte_source(space):
                                                 space.wrap('modules')),
                                                 space.wrap('xyz')))
     assert x == 3
-    lls = rffi.str2charp("sys")
+    lls = rffi.str2charp("sys   # should give a NameError")
     execute_source(lls)
     lltype.free(lls, flavor='raw')
     # did not crash - the same globals
