@@ -13,7 +13,7 @@ class AppTestTraceBackAttributes:
         # XXX why is this called newstring?
         import sys
         def f():
-            raise TypeError, "hello"
+            raise TypeError("hello")
 
         def g():
             f()
@@ -23,7 +23,7 @@ class AppTestTraceBackAttributes:
         except:
             typ,val,tb = sys.exc_info()
         else:
-            raise AssertionError, "should have raised"
+            raise AssertionError("should have raised")
         assert hasattr(tb, 'tb_frame')
         assert hasattr(tb, 'tb_lasti')
         assert hasattr(tb, 'tb_lineno')
@@ -44,7 +44,9 @@ class AppTestTraceBackAttributes:
         # but not to objects without a dict, obviously
         dictdescr = type.__dict__['__dict__']
         raises(TypeError, dictdescr.__get__, 5)
-        raises(TypeError, dictdescr.__set__, 5, d)
+        # TypeError on CPython because descr applies only to its
+        # __objclass__
+        raises((AttributeError, TypeError), dictdescr.__set__, 5, d)
 
     def test_descr_member_descriptor(self):
         class X(object):
@@ -52,12 +54,13 @@ class AppTestTraceBackAttributes:
         member = X.x
         assert member.__name__ == 'x'
         assert member.__objclass__ is X
-        raises((TypeError, AttributeError), "member.__name__ = 'x'")
-        raises((TypeError, AttributeError), "member.__objclass__ = X")
+        raises(AttributeError, "member.__name__ = 'x'")
+        raises(AttributeError, "member.__objclass__ = X")
 
     def test_descr_getsetproperty(self):
         from types import FrameType
         assert FrameType.f_lineno.__name__ == 'f_lineno'
+        assert FrameType.f_lineno.__qualname__ == 'frame.f_lineno'
         assert FrameType.f_lineno.__objclass__ is FrameType
         class A(object):
             pass
@@ -119,7 +122,7 @@ class TestTypeDef:
                             x = X()
                             import __pypy__
                             irepr = __pypy__.internal_repr(x)
-                            print irepr
+                            print(irepr)
                             %s
                             %s
                             %s
@@ -127,10 +130,7 @@ class TestTypeDef:
                         """ % (slots, methodname, checks[0], checks[1],
                                checks[2], checks[3]))
         subclasses = {}
-        for key, subcls in typedef._subclass_cache.items():
-            if key[0] is not space.config:
-                continue
-            cls = key[1]
+        for cls, subcls in typedef._unique_subclass_cache.items():
             subclasses.setdefault(cls, {})
             prevsubcls = subclasses[cls].setdefault(subcls.__name__, subcls)
             assert subcls is prevsubcls
@@ -143,7 +143,11 @@ class TestTypeDef:
             pass
         def fget(self, space, w_self):
             assert self is prop
-        prop = typedef.GetSetProperty(fget, use_closure=True)
+        # NB. this GetSetProperty is not copied when creating the
+        # W_TypeObject because of 'cls'.  Without it, a duplicate of the
+        # GetSetProperty is taken and it is given the w_objclass that is
+        # the W_TypeObject
+        prop = typedef.GetSetProperty(fget, use_closure=True, cls=W_SomeType)
         W_SomeType.typedef = typedef.TypeDef(
             'some_type',
             x=prop)
@@ -178,7 +182,7 @@ class TestTypeDef:
         self.space.appexec([w_obj], """(obj):
             assert type(obj).__hash__ is None
             err = raises(TypeError, hash, obj)
-            assert err.value.message == "'some_type' objects are unhashable"
+            assert str(err.value) == "'some_type' objects are unhashable"
             """)
 
     def test_destructor(self):
@@ -186,35 +190,20 @@ class TestTypeDef:
         class W_Level1(W_Root):
             def __init__(self, space1):
                 assert space1 is space
-            def __del__(self):
+                self.register_finalizer(space)
+            def _finalize_(self):
                 space.call_method(w_seen, 'append', space.wrap(1))
-        class W_Level2(W_Root):
-            def __init__(self, space1):
-                assert space1 is space
-            def __del__(self):
-                self.enqueue_for_destruction(space, W_Level2.destructormeth,
-                                             'FOO ')
-            def destructormeth(self):
-                space.call_method(w_seen, 'append', space.wrap(2))
         W_Level1.typedef = typedef.TypeDef(
             'level1',
             __new__ = typedef.generic_new_descr(W_Level1))
-        W_Level2.typedef = typedef.TypeDef(
-            'level2',
-            __new__ = typedef.generic_new_descr(W_Level2))
         #
         w_seen = space.newlist([])
         W_Level1(space)
         gc.collect(); gc.collect()
-        assert space.unwrap(w_seen) == [1]
-        #
-        w_seen = space.newlist([])
-        W_Level2(space)
-        gc.collect(); gc.collect()
         assert space.str_w(space.repr(w_seen)) == "[]"  # not called yet
         ec = space.getexecutioncontext()
         self.space.user_del_action.perform(ec, None)
-        assert space.unwrap(w_seen) == [2]
+        assert space.unwrap(w_seen) == [1]   # called by user_del_action
         #
         w_seen = space.newlist([])
         self.space.appexec([self.space.gettypeobject(W_Level1.typedef)],
@@ -236,29 +225,17 @@ class TestTypeDef:
             A4()
         """)
         gc.collect(); gc.collect()
-        assert space.unwrap(w_seen) == [4, 1]
+        assert space.unwrap(w_seen) == [4, 1]    # user __del__, and _finalize_
         #
         w_seen = space.newlist([])
-        self.space.appexec([self.space.gettypeobject(W_Level2.typedef)],
+        self.space.appexec([self.space.gettypeobject(W_Level1.typedef)],
         """(level2):
             class A5(level2):
                 pass
             A5()
         """)
         gc.collect(); gc.collect()
-        assert space.unwrap(w_seen) == [2]
-        #
-        w_seen = space.newlist([])
-        self.space.appexec([self.space.gettypeobject(W_Level2.typedef),
-                            w_seen],
-        """(level2, seen):
-            class A6(level2):
-                def __del__(self):
-                    seen.append(6)
-            A6()
-        """)
-        gc.collect(); gc.collect()
-        assert space.unwrap(w_seen) == [6, 2]
+        assert space.unwrap(w_seen) == [1]     # _finalize_ only
 
     def test_multiple_inheritance(self):
         class W_A(W_Root):
@@ -267,8 +244,10 @@ class TestTypeDef:
         class W_C(W_A):
             b = 3
         W_A.typedef = typedef.TypeDef("A",
-            a = typedef.interp_attrproperty("a", cls=W_A),
-            b = typedef.interp_attrproperty("b", cls=W_A),
+            a = typedef.interp_attrproperty("a", cls=W_A,
+                wrapfn="newint"),
+            b = typedef.interp_attrproperty("b", cls=W_A,
+                wrapfn="newint"),
         )
         class W_B(W_Root):
             pass
@@ -341,28 +320,74 @@ class TestTypeDef:
         assert space.is_true(space.ne(w_a, w_b))
         assert not space.is_true(space.ne(w_b, w_c))
 
+    def test_class_attr(self):
+        class W_SomeType(W_Root):
+            pass
+
+        seen = []
+        def make_me(space):
+            seen.append(1)
+            return space.wrap("foobar")
+
+        W_SomeType.typedef = typedef.TypeDef(
+            'some_type',
+            abc = typedef.ClassAttr(make_me)
+            )
+        assert seen == []
+        self.space.appexec([W_SomeType()], """(x):
+            assert type(x).abc == "foobar"
+            assert x.abc == "foobar"
+            assert type(x).abc == "foobar"
+        """)
+        assert seen == [1]
+
+    def test_mapdict_number_of_slots(self):
+        space = self.space
+        a, b, c = space.unpackiterable(space.appexec([], """():
+            class A(object):
+                pass
+            a = A()
+            a.x = 1
+            class B:
+                pass
+            b = B()
+            b.x = 1
+            class C(int):
+                pass
+            c = C(1)
+            c.x = 1
+            return a, b, c
+        """), 3)
+        assert not hasattr(a, "storage")
+        assert not hasattr(b, "storage")
+        assert hasattr(c, "storage")
+
 class AppTestTypeDef:
 
+    spaceconfig = dict(usemodules=['array'])
+
     def setup_class(cls):
         path = udir.join('AppTestTypeDef.txt')
         path.write('hello world\n')
         cls.w_path = cls.space.wrap(str(path))
 
     def test_destructor(self):
-        import gc, os
+        import gc, array
         seen = []
-        class MyFile(file):
+        class MyArray(array.array):
             def __del__(self):
+                # here we check that we can still access the array, i.e. that
+                # the interp-level __del__ has not been called yet
                 seen.append(10)
-                seen.append(os.lseek(self.fileno(), 2, 0))
-        f = MyFile(self.path, 'r')
-        fd = f.fileno()
-        seen.append(os.lseek(fd, 5, 0))
-        del f
+                seen.append(self[0])
+        a = MyArray('i')
+        a.append(42)
+        seen.append(a[0])
+        del a
         gc.collect(); gc.collect(); gc.collect()
         lst = seen[:]
-        assert lst == [5, 10, 2]
-        raises(OSError, os.lseek, fd, 7, 0)
+        print(lst)
+        assert lst == [42, 10, 42]
 
     def test_method_attrs(self):
         import sys
@@ -373,10 +398,10 @@ class AppTestTypeDef:
         class B(A):
             pass
 
-        bm = B().m
-        assert bm.__func__ is bm.im_func
-        assert bm.__self__ is bm.im_self
-        assert bm.im_class is B
+        obj = B()
+        bm = obj.m
+        assert bm.__func__ is A.m
+        assert bm.__self__ is obj
         assert bm.__doc__ == "aaa"
         assert bm.x == 3
         assert type(bm).__doc__ == "instancemethod(function, instance, class)\n\nCreate an instance method object."
@@ -389,15 +414,19 @@ class AppTestTypeDef:
         # differs from .im_class in case the method is
         # defined in some parent class of l's actual class
 
-    def test_classmethod_im_class(self):
-        class Foo(object):
-            @classmethod
-            def bar(cls):
-                pass
-        assert Foo.bar.im_class is type
-
     def test_func_closure(self):
         x = 2
         def f():
             return x
         assert f.__closure__[0].cell_contents is x
+
+    def test_get_with_none_arg(self):
+        raises(TypeError, type.__dict__['__mro__'].__get__, None)
+        raises(TypeError, type.__dict__['__mro__'].__get__, None, None)
+
+    def test_builtin_readonly_property(self):
+        import sys
+        x = lambda: 5
+        e = raises(AttributeError, 'x.__globals__ = {}')
+        if '__pypy__' in sys.builtin_module_names:
+            assert str(e.value) == "readonly attribute '__globals__'"
