@@ -2,37 +2,35 @@
 Interp-level implementation of the basic space operations.
 """
 
+import math
+
 from pypy.interpreter import gateway
-from pypy.interpreter.error import OperationError
+from pypy.interpreter.error import OperationError, oefmt
 from pypy.interpreter.gateway import unwrap_spec, WrappedDefault
 from rpython.rlib.runicode import UNICHR
-from rpython.rlib.rfloat import isnan, isinf, round_double
-from rpython.rlib import rfloat
 import __builtin__
 
 def abs(space, w_val):
     "abs(number) -> number\n\nReturn the absolute value of the argument."
     return space.abs(w_val)
 
-def chr(space, w_ascii):
-    "Return a string of one character with the given ascii code."
-    try:
-        char = __builtin__.chr(space.int_w(w_ascii))
-    except ValueError:  # chr(out-of-range)
-        raise OperationError(space.w_ValueError,
-                             space.wrap("character code not in range(256)"))
-    return space.wrap(char)
+def ascii(space, w_obj):
+    """"ascii(object) -> string
+
+    As repr(), return a string containing a printable representation of an
+    object, but escape the non-ASCII characters in the string returned by
+    repr() using \\x, \\u or \\U escapes.  This generates a string similar
+    to that returned by repr() in Python 2."""
+    from pypy.objspace.std.unicodeobject import ascii_from_object
+    return ascii_from_object(space, w_obj)
 
 @unwrap_spec(code=int)
-def unichr(space, code):
+def chr(space, code):
     "Return a Unicode string of one character with the given ordinal."
-    # XXX range checking!
-    try:
-        c = UNICHR(code)
-    except ValueError:
-        raise OperationError(space.w_ValueError,
-                             space.wrap("unichr() arg out of range"))
-    return space.wrap(c)
+    if code < 0 or code > 0x10FFFF:
+        raise oefmt(space.w_ValueError, "chr() arg out of range")
+    c = UNICHR(code)
+    return space.newunicode(c)
 
 def len(space, w_obj):
     "len(object) -> integer\n\nReturn the number of items of a sequence or mapping."
@@ -46,9 +44,9 @@ def checkattrname(space, w_name):
     # space.{get,set,del}attr()...
     # Note that if w_name is already an exact string it must be returned
     # unmodified (and not e.g. unwrapped-rewrapped).
-    if not space.is_w(space.type(w_name), space.w_str):
-        name = space.str_w(w_name)    # typecheck
-        w_name = space.wrap(name)     # rewrap as a real string
+    if not space.is_w(space.type(w_name), space.w_text):
+        name = space.text_w(w_name)    # typecheck
+        w_name = space.newtext(name)     # rewrap as a real string
     return w_name
 
 def delattr(space, w_object, w_name):
@@ -64,7 +62,7 @@ getattr(x, 'y') is equivalent to ``x.y''."""
     w_name = checkattrname(space, w_name)
     try:
         return space.getattr(w_object, w_name)
-    except OperationError, e:
+    except OperationError as e:
         if w_defvalue is not None:
             if e.match(space, space.w_AttributeError):
                 return w_defvalue
@@ -74,10 +72,14 @@ def hasattr(space, w_object, w_name):
     """Return whether the object has an attribute with the given name.
     (This is done by calling getattr(object, name) and catching exceptions.)"""
     w_name = checkattrname(space, w_name)
-    if space.findattr(w_object, w_name) is not None:
-        return space.w_True
+    try:
+        space.getattr(w_object, w_name)
+    except OperationError as e:
+        if e.match(space, space.w_AttributeError):
+            return space.w_False
+        raise
     else:
-        return space.w_False
+        return space.w_True
 
 def hash(space, w_object):
     """Return a hash value for the object.  Two objects which compare as
@@ -85,75 +87,30 @@ equal have the same hash value.  It is possible, but unlikely, for
 two un-equal objects to have the same hash value."""
     return space.hash(w_object)
 
-def oct(space, w_val):
-    """Return the octal representation of an integer."""
-    # XXX does this need to be a space operation?
-    return space.oct(w_val)
-
-def hex(space, w_val):
-    """Return the hexadecimal representation of an integer."""
-    return space.hex(w_val)
-
 def id(space, w_object):
     "Return the identity of an object: id(x) == id(y) if and only if x is y."
     return space.id(w_object)
 
-def cmp(space, w_x, w_y):
-    """return 0 when x == y, -1 when x < y and 1 when x > y """
-    return space.cmp(w_x, w_y)
-
-def coerce(space, w_x, w_y):
-    """coerce(x, y) -> (x1, y1)
-
-Return a tuple consisting of the two numeric arguments converted to
-a common type, using the same rules as used by arithmetic operations.
-If coercion is not possible, raise TypeError."""
-    return space.coerce(w_x, w_y)
-
 def divmod(space, w_x, w_y):
     """Return the tuple ((x-x%y)/y, x%y).  Invariant: div*y + mod == x."""
     return space.divmod(w_x, w_y)
 
-# semi-private: works only for new-style classes.
-def _issubtype(space, w_cls1, w_cls2):
-    return space.issubtype(w_cls1, w_cls2)
-
 # ____________________________________________________________
 
-# Here 0.30103 is an upper bound for log10(2)
-NDIGITS_MAX = int((rfloat.DBL_MANT_DIG - rfloat.DBL_MIN_EXP) * 0.30103)
-NDIGITS_MIN = -int((rfloat.DBL_MAX_EXP + 1) * 0.30103)
-
-@unwrap_spec(number=float, w_ndigits = WrappedDefault(0))
-def round(space, number, w_ndigits):
-    """round(number[, ndigits]) -> floating point number
+def round(space, w_number, w_ndigits=None):
+    """round(number[, ndigits]) -> number
 
 Round a number to a given precision in decimal digits (default 0 digits).
-This always returns a floating point number.  Precision may be negative."""
-    # Algorithm copied directly from CPython
-
-    # interpret 2nd argument as a Py_ssize_t; clip on overflow
-    ndigits = space.getindex_w(w_ndigits, None)
-
-    # nans, infinities and zeros round to themselves
-    if number == 0 or isinf(number) or isnan(number):
-        return space.wrap(number)
-
-    # Deal with extreme values for ndigits. For ndigits > NDIGITS_MAX, x
-    # always rounds to itself.  For ndigits < NDIGITS_MIN, x always
-    # rounds to +-0.0.
-    if ndigits > NDIGITS_MAX:
-        return space.wrap(number)
-    elif ndigits < NDIGITS_MIN:
-        # return 0.0, but with sign of x
-        return space.wrap(0.0 * number)
-
-    # finite x, and ndigits is not unreasonably large
-    z = round_double(number, ndigits)
-    if isinf(z):
-        raise OperationError(space.w_OverflowError,
-                             space.wrap("rounded value too large to represent"))
-    return space.wrap(z)
+This returns an int when called with one argument or if ndigits=None,
+otherwise the same type as the number. ndigits may be negative."""
+    round = space.lookup(w_number, '__round__')
+    if round is None:
+        raise oefmt(space.w_TypeError,
+                    "type %T doesn't define __round__ method", w_number)
+    if space.is_none(w_ndigits):
+        return space.get_and_call_function(round, w_number)
+    else:
+        return space.get_and_call_function(round, w_number, w_ndigits)
 
 # ____________________________________________________________
 
@@ -170,7 +127,7 @@ iter_sentinel = gateway.applevel('''
 
     def iter_sentinel(callable_, sentinel):
         if not callable(callable_):
-            raise TypeError, 'iter(v, w): v must be callable'
+            raise TypeError('iter(v, w): v must be callable')
         return iter_generator(callable_, sentinel)
 
 ''', filename=__file__).interphook("iter_sentinel")
@@ -179,7 +136,7 @@ def iter(space, w_collection_or_callable, w_sentinel=None):
     """iter(collection) -> iterator over the elements of the collection.
 
 iter(callable, sentinel) -> iterator calling callable() until it returns
-                            the sentinal.
+                            the sentinel.
 """
     if w_sentinel is None:
         return space.iter(w_collection_or_callable)
@@ -192,7 +149,7 @@ Return the next item from the iterator. If default is given and the iterator
 is exhausted, it is returned instead of raising StopIteration."""
     try:
         return space.next(w_iterator)
-    except OperationError, e:
+    except OperationError as e:
         if w_default is not None and e.match(space, space.w_StopIteration):
             return w_default
         raise
@@ -220,21 +177,12 @@ setattr(x, 'y', z) is equivalent to ``x.y = z''."""
     space.setattr(w_object, w_name, w_val)
     return space.w_None
 
-def intern(space, w_str):
-    """``Intern'' the given string.  This enters the string in the (global)
-table of interned strings whose purpose is to speed up dictionary lookups.
-Return the string itself or the previously interned string object with the
-same value."""
-    if space.is_w(space.type(w_str), space.w_str):
-        return space.new_interned_w_str(w_str)
-    raise OperationError(space.w_TypeError, space.wrap("intern() argument must be string."))
-
 def callable(space, w_object):
     """Check whether the object appears to be callable (i.e., some kind of
 function).  Note that classes are callable."""
     return space.callable(w_object)
 
-@unwrap_spec(w_format_spec = WrappedDefault(""))
+@unwrap_spec(w_format_spec = WrappedDefault(u""))
 def format(space, w_obj, w_format_spec):
     """Format a obj according to format_spec"""
     return space.format(w_obj, w_format_spec)
