@@ -1,4 +1,5 @@
-from rpython.rlib.rstring import StringBuilder
+from rpython.rlib.rstring import UnicodeBuilder
+from rpython.rlib import objectmodel
 from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.error import OperationError
 from pypy.interpreter.gateway import unwrap_spec
@@ -10,7 +11,7 @@ from pypy.module._csv.interp_csv import (QUOTE_MINIMAL, QUOTE_ALL,
 
 (START_RECORD, START_FIELD, ESCAPED_CHAR, IN_FIELD,
  IN_QUOTED_FIELD, ESCAPE_IN_QUOTED_FIELD, QUOTE_IN_QUOTED_FIELD,
- EAT_CRNL) = range(8)
+ EAT_CRNL, AFTER_ESCAPED_CRNL) = range(9)
 
 
 class W_Reader(W_Root):
@@ -22,37 +23,30 @@ class W_Reader(W_Root):
         self.line_num = 0
 
     def iter_w(self):
-        return self.space.wrap(self)
+        return self
 
+    @objectmodel.dont_inline
     def error(self, msg):
         space = self.space
-        msg = 'line %d: %s' % (self.line_num, msg)
+        msg = u'line %d: %s' % (self.line_num, msg)
         w_module = space.getbuiltinmodule('_csv')
-        w_error = space.getattr(w_module, space.wrap('Error'))
-        raise OperationError(w_error, space.wrap(msg))
-    error._dont_inline_ = True
+        w_error = space.getattr(w_module, space.newtext('Error'))
+        raise OperationError(w_error, space.newunicode(msg))
 
     def add_char(self, field_builder, c):
         assert field_builder is not None
         if field_builder.getlength() >= field_limit.limit:
-            raise self.error("field larger than field limit")
+            raise self.error(u"field larger than field limit")
         field_builder.append(c)
 
     def save_field(self, field_builder):
         space = self.space
         field = field_builder.build()
         if self.numeric_field:
-            from rpython.rlib.rstring import ParseStringError
-            from rpython.rlib.rfloat import string_to_float
             self.numeric_field = False
-            try:
-                ff = string_to_float(field)
-            except ParseStringError as e:
-                from pypy.objspace.std.intobject import wrap_parsestringerror
-                raise wrap_parsestringerror(space, e, space.wrap(field))
-            w_obj = space.wrap(ff)
+            w_obj = space.call_function(space.w_float, space.newunicode(field))
         else:
-            w_obj = space.wrap(field)
+            w_obj = space.newunicode(field)
         self.fields_w.append(w_obj)
 
     def next_w(self):
@@ -66,26 +60,29 @@ class W_Reader(W_Root):
         while True:
             try:
                 w_line = space.next(self.w_iter)
-            except OperationError, e:
+            except OperationError as e:
                 if e.match(space, space.w_StopIteration):
                     if (field_builder is not None and
                             state != START_RECORD and state != EAT_CRNL and
                             (len(field_builder.build()) > 0 or
                              state == IN_QUOTED_FIELD)):
                         if dialect.strict:
-                            raise self.error("newline inside string")
+                            raise self.error(u"unexpected end of data")
                         else:
                             self.save_field(field_builder)
                             break
                 raise
             self.line_num += 1
-            line = space.str_w(w_line)
+            if space.isinstance_w(w_line, space.w_bytes):
+                raise self.error(u"iterator should return strings, not bytes "
+                                 u"(did you open the file in text mode?")
+            line = space.unicode_w(w_line)
             for c in line:
-                if c == '\0':
-                    raise self.error("line contains NULL byte")
+                if c == u'\0':
+                    raise self.error(u"line contains NULL byte")
 
                 if state == START_RECORD:
-                    if c == '\n' or c == '\r':
+                    if c == u'\n' or c == u'\r':
                         state = EAT_CRNL
                         continue
                     # normal character - handle as START_FIELD
@@ -93,9 +90,9 @@ class W_Reader(W_Root):
                     # fall-through to the next case
 
                 if state == START_FIELD:
-                    field_builder = StringBuilder(64)
+                    field_builder = UnicodeBuilder(64)
                     # expecting field
-                    if c == '\n' or c == '\r':
+                    if c == u'\n' or c == u'\r':
                         # save empty field
                         self.save_field(field_builder)
                         state = EAT_CRNL
@@ -106,7 +103,7 @@ class W_Reader(W_Root):
                     elif c == dialect.escapechar:
                         # possible escaped character
                         state = ESCAPED_CHAR
-                    elif c == ' ' and dialect.skipinitialspace:
+                    elif c == u' ' and dialect.skipinitialspace:
                         # ignore space at start of field
                         pass
                     elif c == dialect.delimiter:
@@ -120,12 +117,16 @@ class W_Reader(W_Root):
                         state = IN_FIELD
 
                 elif state == ESCAPED_CHAR:
-                    self.add_char(field_builder, c)
-                    state = IN_FIELD
+                    if c in '\n\r':
+                        self.add_char(field_builder, c)
+                        state = AFTER_ESCAPED_CRNL
+                    else:
+                        self.add_char(field_builder, c)
+                        state = IN_FIELD
 
-                elif state == IN_FIELD:
+                elif state == IN_FIELD or state == AFTER_ESCAPED_CRNL:
                     # in unquoted field
-                    if c == '\n' or c == '\r':
+                    if c == u'\n' or c == u'\r':
                         # end of line
                         self.save_field(field_builder)
                         state = EAT_CRNL
@@ -172,7 +173,7 @@ class W_Reader(W_Root):
                         # save field - wait for new field
                         self.save_field(field_builder)
                         state = START_FIELD
-                    elif c == '\n' or c == '\r':
+                    elif c == u'\n' or c == u'\r':
                         # end of line
                         self.save_field(field_builder)
                         state = EAT_CRNL
@@ -181,31 +182,33 @@ class W_Reader(W_Root):
                         state = IN_FIELD
                     else:
                         # illegal
-                        raise self.error("'%s' expected after '%s'" % (
+                        raise self.error(u"'%s' expected after '%s'" % (
                             dialect.delimiter, dialect.quotechar))
 
                 elif state == EAT_CRNL:
-                    if not (c == '\n' or c == '\r'):
-                        raise self.error("new-line character seen in unquoted "
-                                        "field - do you need to open the file "
-                                        "in universal-newline mode?")
+                    if not (c == u'\n' or c == u'\r'):
+                        raise self.error(u"new-line character seen in unquoted "
+                                         u"field - do you need to open the file "
+                                         u"in universal-newline mode?")
 
             if state == IN_FIELD or state == QUOTE_IN_QUOTED_FIELD:
                 self.save_field(field_builder)
                 break
             elif state == ESCAPED_CHAR:
-                self.add_char(field_builder, '\n')
+                self.add_char(field_builder, u'\n')
                 state = IN_FIELD
             elif state == IN_QUOTED_FIELD:
                 pass
             elif state == ESCAPE_IN_QUOTED_FIELD:
-                self.add_char(field_builder, '\n')
+                self.add_char(field_builder, u'\n')
                 state = IN_QUOTED_FIELD
             elif state == START_FIELD:
                 # save empty field
-                field_builder = StringBuilder(1)
+                field_builder = UnicodeBuilder(1)
                 self.save_field(field_builder)
                 break
+            elif state == AFTER_ESCAPED_CRNL:
+                continue
             else:
                 break
         #
@@ -247,9 +250,10 @@ def csv_reader(space, w_iterator, w_dialect=None,
 W_Reader.typedef = TypeDef(
         '_csv.reader',
         dialect = interp_attrproperty_w('dialect', W_Reader),
-        line_num = interp_attrproperty('line_num', W_Reader),
+        line_num = interp_attrproperty('line_num', W_Reader,
+            wrapfn="newint"),
         __iter__ = interp2app(W_Reader.iter_w),
-        next = interp2app(W_Reader.next_w),
+        __next__ = interp2app(W_Reader.next_w),
         __doc__ = """CSV reader
 
 Reader objects are responsible for reading and parsing tabular data
@@ -267,4 +271,4 @@ def csv_field_size_limit(space, new_limit=-1):
     old_limit = field_limit.limit
     if new_limit >= 0:
         field_limit.limit = new_limit
-    return space.wrap(old_limit)
+    return space.newint(old_limit)
