@@ -23,39 +23,34 @@ def trap_eintr(space, error):
         return False
     try:
         w_value = error.get_w_value(space)
-        w_errno = space.getattr(w_value, space.wrap("errno"))
-        return space.eq_w(w_errno, space.wrap(EINTR))
+        w_errno = space.getattr(w_value, space.newtext("errno"))
+        return space.eq_w(w_errno, space.newint(EINTR))
     except OperationError:
         return False
 
 def unsupported(space, message):
     w_exc = space.getattr(space.getbuiltinmodule('_io'),
-                          space.wrap('UnsupportedOperation'))
-    return OperationError(w_exc, space.wrap(message))
+                          space.newtext('UnsupportedOperation'))
+    return OperationError(w_exc, space.newtext(message))
 
 # May be called with any object
 def check_readable_w(space, w_obj):
     if not space.is_true(space.call_method(w_obj, 'readable')):
-        raise OperationError(
-            space.w_IOError,
-            space.wrap("file or stream is not readable"))
+        raise unsupported(space, "File or stream is not readable")
 
 # May be called with any object
 def check_writable_w(space, w_obj):
     if not space.is_true(space.call_method(w_obj, 'writable')):
-        raise OperationError(
-            space.w_IOError,
-            space.wrap("file or stream is not writable"))
+        raise unsupported(space, "File or stream is not writable")
 
 # May be called with any object
 def check_seekable_w(space, w_obj):
     if not space.is_true(space.call_method(w_obj, 'seekable')):
-        raise OperationError(
-            space.w_IOError,
-            space.wrap("file or stream is not seekable"))
-
+        raise unsupported(space, "File or stream is not seekable")
 
 class W_IOBase(W_Root):
+    cffi_fileobj = None    # pypy/module/_cffi_backend
+
     def __init__(self, space, add_to_autoflusher=True):
         # XXX: IOBase thinks it has to maintain its own internal state in
         # `__IOBase_closed` and call flush() by itself, but it is redundant
@@ -65,6 +60,8 @@ class W_IOBase(W_Root):
         self.__IOBase_closed = False
         if add_to_autoflusher:
             get_autoflusher(space).add(self)
+        if self.needs_finalizer():
+            self.register_finalizer(space)
 
     def getdict(self, space):
         return self.w_dict
@@ -72,25 +69,34 @@ class W_IOBase(W_Root):
     def _closed(self, space):
         # This gets the derived attribute, which is *not* __IOBase_closed
         # in most cases!
-        w_closed = space.findattr(self, space.wrap('closed'))
+        w_closed = space.findattr(self, space.newtext('closed'))
         if w_closed is not None and space.is_true(w_closed):
             return True
         return False
 
-    def __del__(self):
-        self.clear_all_weakrefs()
-        self.enqueue_for_destruction(self.space, W_IOBase.destructor,
-                                     'internal __del__ of ')
+    def _finalize_(self):
+        # Note: there is only this empty _finalize_() method here, but
+        # we still need register_finalizer() so that descr_del() is
+        # called.  IMPORTANT: this is not the recommended way to have a
+        # finalizer!  It makes the finalizer appear as __del__() from
+        # app-level, and the user can call __del__() explicitly, or
+        # override it, with or without calling the parent's __del__().
+        # This matches 'tp_finalize' in CPython >= 3.4.  So far (3.5),
+        # this is the only built-in class with a 'tp_finalize' slot that
+        # can be subclassed.
+        pass
 
-    def destructor(self):
-        assert isinstance(self, W_IOBase)
+    def descr_del(self):
         space = self.space
-        w_closed = space.findattr(self, space.wrap('closed'))
+        w_closed = space.findattr(self, space.newtext('closed'))
         try:
             # If `closed` doesn't exist or can't be evaluated as bool, then
             # the object is probably in an unusable state, so ignore.
             if w_closed is not None and not space.is_true(w_closed):
-                space.call_method(self, "close")
+                try:
+                    self._dealloc_warn_w(space, self)
+                finally:
+                    space.call_method(self, "close")
         except OperationError:
             # Silencing I/O errors is bad, but printing spurious tracebacks is
             # equally as bad, and potentially more frequent (because of
@@ -111,7 +117,7 @@ class W_IOBase(W_Root):
             message = "I/O operation on closed file"
         if self._closed(space):
             raise OperationError(
-                space.w_ValueError, space.wrap(message))
+                space.w_ValueError, space.newtext(message))
 
     def check_closed_w(self, space):
         self._check_closed(space)
@@ -122,22 +128,35 @@ class W_IOBase(W_Root):
     def close_w(self, space):
         if self._CLOSED():
             return
+
+        cffifo = self.cffi_fileobj
+        self.cffi_fileobj = None
+        if cffifo is not None:
+            cffifo.close()
+
         try:
             space.call_method(self, "flush")
         finally:
             self.__IOBase_closed = True
 
+    def needs_finalizer(self):
+        # can return False if we know that the precise close() method
+        # of this class will have no effect
+        return True
+
+    def _dealloc_warn_w(self, space, w_source):
+        """Called when the io is implicitly closed via the deconstructor"""
+        pass
+
     def flush_w(self, space):
         if self._CLOSED():
-            raise OperationError(
-                space.w_ValueError,
-                space.wrap("I/O operation on closed file"))
+            raise oefmt(space.w_ValueError, "I/O operation on closed file")
 
     def seek_w(self, space, w_offset, w_whence=None):
         self._unsupportedoperation(space, "seek")
 
     def tell_w(self, space):
-        return space.call_method(self, "seek", space.wrap(0), space.wrap(1))
+        return space.call_method(self, "seek", space.newint(0), space.newint(1))
 
     def truncate_w(self, space, w_size=None):
         self._unsupportedoperation(space, "truncate")
@@ -147,14 +166,14 @@ class W_IOBase(W_Root):
 
     def enter_w(self, space):
         self._check_closed(space)
-        return space.wrap(self)
+        return self
 
     def exit_w(self, space, __args__):
         space.call_method(self, "close")
 
     def iter_w(self, space):
         self._check_closed(space)
-        return space.wrap(self)
+        return self
 
     def next_w(self, space):
         w_line = space.call_method(self, "readline")
@@ -175,13 +194,16 @@ class W_IOBase(W_Root):
     def seekable_w(self, space):
         return space.w_False
 
+    def getstate_w(self, space):
+        raise oefmt(space.w_TypeError, "cannot serialize '%T' object", self)
+
     # ______________________________________________________________
 
     def readline_w(self, space, w_limit=None):
         # For backwards compatibility, a (slowish) readline().
         limit = convert_size(space, w_limit)
 
-        has_peek = space.findattr(self, space.wrap("peek"))
+        has_peek = space.findattr(self, space.newtext("peek"))
 
         builder = StringBuilder()
         size = 0
@@ -191,19 +213,19 @@ class W_IOBase(W_Root):
 
             if has_peek:
                 try:
-                    w_readahead = space.call_method(self, "peek", space.wrap(1))
-                except OperationError, e:
+                    w_readahead = space.call_method(self, "peek", space.newint(1))
+                except OperationError as e:
                     if trap_eintr(space, e):
                         continue
                     raise
-                if not space.isinstance_w(w_readahead, space.w_str):
+                if not space.isinstance_w(w_readahead, space.w_bytes):
                     raise oefmt(space.w_IOError,
                                 "peek() should have returned a bytes object, "
                                 "not '%T'", w_readahead)
                 length = space.len_w(w_readahead)
                 if length > 0:
                     n = 0
-                    buf = space.str_w(w_readahead)
+                    buf = space.bytes_w(w_readahead)
                     if limit >= 0:
                         while True:
                             if n >= length or n >= limit:
@@ -221,16 +243,16 @@ class W_IOBase(W_Root):
                     nreadahead = n
 
             try:
-                w_read = space.call_method(self, "read", space.wrap(nreadahead))
-            except OperationError, e:
+                w_read = space.call_method(self, "read", space.newint(nreadahead))
+            except OperationError as e:
                 if trap_eintr(space, e):
                     continue
                 raise
-            if not space.isinstance_w(w_read, space.w_str):
+            if not space.isinstance_w(w_read, space.w_bytes):
                 raise oefmt(space.w_IOError,
                             "peek() should have returned a bytes object, not "
                             "'%T'", w_read)
-            read = space.str_w(w_read)
+            read = space.bytes_w(w_read)
             if not read:
                 break
 
@@ -240,7 +262,7 @@ class W_IOBase(W_Root):
             if read[-1] == '\n':
                 break
 
-        return space.wrap(builder.build())
+        return space.newbytes(builder.build())
 
     def readlines_w(self, space, w_hint=None):
         hint = convert_size(space, w_hint)
@@ -272,14 +294,14 @@ class W_IOBase(W_Root):
         while True:
             try:
                 w_line = space.next(w_iterator)
-            except OperationError, e:
+            except OperationError as e:
                 if not e.match(space, space.w_StopIteration):
                     raise
                 break  # done
             while True:
                 try:
                     space.call_method(self, "write", w_line)
-                except OperationError, e:
+                except OperationError as e:
                     if trap_eintr(space, e):
                         continue
                     raise
@@ -292,7 +314,7 @@ W_IOBase.typedef = TypeDef(
     __enter__ = interp2app(W_IOBase.enter_w),
     __exit__ = interp2app(W_IOBase.exit_w),
     __iter__ = interp2app(W_IOBase.iter_w),
-    next = interp2app(W_IOBase.next_w),
+    __next__ = interp2app(W_IOBase.next_w),
     close = interp2app(W_IOBase.close_w),
     flush = interp2app(W_IOBase.flush_w),
     seek = interp2app(W_IOBase.seek_w),
@@ -308,9 +330,12 @@ W_IOBase.typedef = TypeDef(
     _checkWritable = interp2app(check_writable_w),
     _checkSeekable = interp2app(check_seekable_w),
     _checkClosed = interp2app(W_IOBase.check_closed_w),
-    closed = GetSetProperty(W_IOBase.closed_get_w),
+    closed = GetSetProperty(W_IOBase.closed_get_w,
+                            doc="True if the file is closed"),
     __dict__ = GetSetProperty(descr_get_dict, descr_set_dict, cls=W_IOBase),
     __weakref__ = make_weakref_descr(W_IOBase),
+    __del__ = interp2app(W_IOBase.descr_del),
+    __confirm_applevel_del__ = True,
 
     readline = interp2app(W_IOBase.readline_w),
     readlines = interp2app(W_IOBase.readlines_w),
@@ -331,15 +356,15 @@ class W_RawIOBase(W_IOBase):
         if space.is_w(w_length, space.w_None):
             return w_length
         space.delslice(w_buffer, w_length, space.len(w_buffer))
-        return space.str(w_buffer)
+        return space.call_function(space.w_bytes, w_buffer)
 
     def readall_w(self, space):
         builder = StringBuilder()
         while True:
             try:
                 w_data = space.call_method(self, "read",
-                                           space.wrap(DEFAULT_BUFFER_SIZE))
-            except OperationError, e:
+                                           space.newint(DEFAULT_BUFFER_SIZE))
+            except OperationError as e:
                 if trap_eintr(space, e):
                     continue
                 raise
@@ -348,14 +373,13 @@ class W_RawIOBase(W_IOBase):
                     return w_data
                 break
 
-            if not space.isinstance_w(w_data, space.w_str):
-                raise OperationError(space.w_TypeError, space.wrap(
-                    "read() should return bytes"))
-            data = space.str_w(w_data)
+            if not space.isinstance_w(w_data, space.w_bytes):
+                raise oefmt(space.w_TypeError, "read() should return bytes")
+            data = space.bytes_w(w_data)
             if not data:
                 break
             builder.append(data)
-        return space.wrap(builder.build())
+        return space.newbytes(builder.build())
 
 W_RawIOBase.typedef = TypeDef(
     '_io._RawIOBase', W_IOBase.typedef,
@@ -385,20 +409,22 @@ class AutoFlusher(rweaklist.RWeakListMixin):
     def flush_all(self, space):
         while True:
             handles = self.get_all_handles()
-            if len(handles) == 0:
-                break
             self.initialize()  # reset the state here
+            progress = False
             for wr in handles:
                 w_iobase = wr()
                 if w_iobase is None:
                     continue
+                progress = True
                 try:
                     space.call_method(w_iobase, 'flush')
                 except OperationError:
                     # Silencing all errors is bad, but getting randomly
                     # interrupted here is equally as bad, and potentially
                     # more frequent (because of shutdown issues).
-                    pass 
+                    pass
+            if not progress:
+                break
 
 def get_autoflusher(space):
     return space.fromcache(AutoFlusher)
