@@ -80,18 +80,18 @@ def test_decode_hz_error():
     #
     e = py.test.raises(EncodeDecodeError, decode, c, "~{xyz}").value
     assert e.start == 2
-    assert e.end == 4
+    assert e.end == 3
     assert e.reason == "illegal multibyte sequence"
 
 def test_decode_hz_ignore():
     c = getcodec("hz")
     u = decode(c, 'def~{}abc', 'ignore')
-    assert u == u'def\u5fcf'
+    assert u == u'def\u5f95'
 
 def test_decode_hz_replace():
     c = getcodec("hz")
     u = decode(c, 'def~{}abc', 'replace')
-    assert u == u'def\ufffd\u5fcf'
+    assert u == u'def\ufffd\u5f95\ufffd'
 
 def test_encode_hz():
     c = getcodec("hz")
