@@ -4,9 +4,9 @@ to app-level with apropriate interface
 """
 
 from pypy.interpreter.gateway import interp2app, unwrap_spec
-from pypy.interpreter.typedef import TypeDef, GetSetProperty, interp_attrproperty
+from pypy.interpreter.typedef import TypeDef, GetSetProperty, interp_attrproperty_w
 from rpython.rtyper.lltypesystem import lltype, rffi
-from pypy.interpreter.error import OperationError
+from pypy.interpreter.error import OperationError, oefmt
 from pypy.module._rawffi.interp_rawffi import segfault_exception
 from pypy.module._rawffi.interp_rawffi import W_DataShape, W_DataInstance
 from pypy.module._rawffi.interp_rawffi import unwrap_value, wrap_value
@@ -15,7 +15,7 @@ from pypy.module._rawffi.interp_rawffi import size_alignment
 from pypy.module._rawffi.interp_rawffi import unpack_shape_with_length
 from pypy.module._rawffi.interp_rawffi import read_ptr, write_ptr
 from rpython.rlib.rarithmetic import r_uint
-from rpython.rlib import rgc
+from rpython.rlib import rgc, clibffi
 
 
 class W_Array(W_DataShape):
@@ -36,30 +36,29 @@ class W_Array(W_DataShape):
     def get_basic_ffi_type(self):
         return self.basicffitype
 
-    @unwrap_spec(length=int, autofree=bool)
+    @unwrap_spec(length=int, autofree=int)
     def descr_call(self, space, length, w_items=None, autofree=False):
-        result = self.allocate(space, length, autofree)
+        result = self.allocate(space, length, bool(autofree))
         if not space.is_none(w_items):
             items_w = space.unpackiterable(w_items)
             iterlength = len(items_w)
             if iterlength > length:
-                raise OperationError(space.w_ValueError,
-                                     space.wrap("too many items for specified"
-                                                " array length"))
+                raise oefmt(space.w_ValueError,
+                            "too many items for specified array length")
             for num in range(iterlength):
                 w_item = items_w[num]
                 unwrap_value(space, write_ptr, result.ll_buffer, num,
                              self.itemcode, w_item)
-        return space.wrap(result)
+        return result
 
     def descr_repr(self, space):
-        return space.wrap("<_rawffi.Array '%s' (%d, %d)>" % (self.itemcode,
-                                                             self.size,
-                                                             self.alignment))
+        return space.newtext("<_rawffi.Array '%s' (%d, %d)>" % (self.itemcode,
+                                                                self.size,
+                                                                self.alignment))
 
     @unwrap_spec(address=r_uint, length=int)
     def fromaddress(self, space, address, length):
-        return space.wrap(W_ArrayInstance(space, self, length, address))
+        return W_ArrayInstance(space, self, length, address)
 
 PRIMITIVE_ARRAY_TYPES = {}
 for _code in TYPEMAP:
@@ -84,22 +83,21 @@ W_Array.typedef.acceptable_as_base_class = False
 
 class W_ArrayInstance(W_DataInstance):
     def __init__(self, space, shape, length, address=r_uint(0)):
-        # Workaround for a strange behavior of libffi: make sure that
-        # we always have at least 8 bytes.  For W_ArrayInstances that are
-        # used as the result value of a function call, ffi_call() writes
-        # 8 bytes into it even if the function's result type asks for less.
-        # This strange behavior is documented.
         memsize = shape.size * length
-        if memsize < 8:
-            memsize = 8
+        # For W_ArrayInstances that are used as the result value of a
+        # function call, ffi_call() writes 8 bytes into it even if the
+        # function's result type asks for less.
+        memsize = clibffi.adjust_return_size(memsize)
         W_DataInstance.__init__(self, space, memsize, address)
         self.length = length
         self.shape = shape
+        self.fmt = shape.itemcode
+        self.itemsize = shape.size
 
     def descr_repr(self, space):
         addr = rffi.cast(lltype.Unsigned, self.ll_buffer)
-        return space.wrap("<_rawffi array %x of length %d>" % (addr,
-                                                               self.length))
+        return space.newtext("<_rawffi array %x of length %d>" % (addr,
+                                                                  self.length))
 
     # This only allows non-negative indexes.  Arrays of shape 'c' also
     # support simple slices.
@@ -109,13 +107,12 @@ class W_ArrayInstance(W_DataInstance):
             raise segfault_exception(space, "setting element of freed array")
         if num >= self.length or num < 0:
             raise OperationError(space.w_IndexError, space.w_None)
-        unwrap_value(space, write_ptr, self.ll_buffer, num,
-                     self.shape.itemcode, w_value)
+        unwrap_value(space, write_ptr, self.ll_buffer, num, self.fmt, w_value)
 
     def descr_setitem(self, space, w_index, w_value):
         try:
             num = space.int_w(w_index)
-        except OperationError, e:
+        except OperationError as e:
             if not e.match(space, space.w_TypeError):
                 raise
             self.setslice(space, w_index, w_value)
@@ -127,13 +124,12 @@ class W_ArrayInstance(W_DataInstance):
             raise segfault_exception(space, "accessing elements of freed array")
         if num >= self.length or num < 0:
             raise OperationError(space.w_IndexError, space.w_None)
-        return wrap_value(space, read_ptr, self.ll_buffer, num,
-                          self.shape.itemcode)
+        return wrap_value(space, read_ptr, self.ll_buffer, num, self.fmt)
 
     def descr_getitem(self, space, w_index):
         try:
             num = space.int_w(w_index)
-        except OperationError, e:
+        except OperationError as e:
             if not e.match(space, space.w_TypeError):
                 raise
             return self.getslice(space, w_index)
@@ -141,29 +137,24 @@ class W_ArrayInstance(W_DataInstance):
             return self.getitem(space, num)
 
     def getlength(self, space):
-        return space.wrap(self.length)
+        return space.newint(self.length)
 
     @unwrap_spec(num=int)
     def descr_itemaddress(self, space, num):
-        itemsize = self.shape.size
-        ptr = rffi.ptradd(self.ll_buffer, itemsize * num)
-        return space.wrap(rffi.cast(lltype.Unsigned, ptr))
+        ptr = rffi.ptradd(self.ll_buffer, self.itemsize * num)
+        return space.newint(rffi.cast(lltype.Unsigned, ptr))
 
     def getrawsize(self):
-        itemsize = self.shape.size
-        return itemsize * self.length
+        return self.itemsize * self.length
 
     def decodeslice(self, space, w_slice):
         if not space.isinstance_w(w_slice, space.w_slice):
-            raise OperationError(space.w_TypeError,
-                                 space.wrap('index must be int or slice'))
-        letter = self.shape.itemcode
-        if letter != 'c':
-            raise OperationError(space.w_TypeError,
-                                 space.wrap("only 'c' arrays support slicing"))
-        w_start = space.getattr(w_slice, space.wrap('start'))
-        w_stop = space.getattr(w_slice, space.wrap('stop'))
-        w_step = space.getattr(w_slice, space.wrap('step'))
+            raise oefmt(space.w_TypeError, "index must be int or slice")
+        if self.fmt != 'c':
+            raise oefmt(space.w_TypeError, "only 'c' arrays support slicing")
+        w_start = space.getattr(w_slice, space.newtext('start'))
+        w_stop = space.getattr(w_slice, space.newtext('stop'))
+        w_step = space.getattr(w_slice, space.newtext('step'))
 
         if space.is_w(w_start, space.w_None):
             start = 0
@@ -176,11 +167,9 @@ class W_ArrayInstance(W_DataInstance):
         if not space.is_w(w_step, space.w_None):
             step = space.int_w(w_step)
             if step != 1:
-                raise OperationError(space.w_ValueError,
-                                     space.wrap("no step support"))
+                raise oefmt(space.w_ValueError, "no step support")
         if not (0 <= start <= stop <= self.length):
-            raise OperationError(space.w_ValueError,
-                                 space.wrap("slice out of bounds"))
+            raise oefmt(space.w_ValueError, "slice out of bounds")
         if not self.ll_buffer:
             raise segfault_exception(space, "accessing a freed array")
         return start, stop
@@ -189,26 +178,26 @@ class W_ArrayInstance(W_DataInstance):
         start, stop = self.decodeslice(space, w_slice)
         ll_buffer = self.ll_buffer
         result = [ll_buffer[i] for i in range(start, stop)]
-        return space.wrap(''.join(result))
+        return space.newbytes(''.join(result))
 
     def setslice(self, space, w_slice, w_value):
         start, stop = self.decodeslice(space, w_slice)
-        value = space.str_w(w_value)
+        value = space.bytes_w(w_value)
         if start + len(value) != stop:
-            raise OperationError(space.w_ValueError,
-                                 space.wrap("cannot resize array"))
+            raise oefmt(space.w_ValueError, "cannot resize array")
         ll_buffer = self.ll_buffer
         for i in range(len(value)):
             ll_buffer[start + i] = value[i]
 
+
 W_ArrayInstance.typedef = TypeDef(
-    'ArrayInstance',
+    'ArrayInstance', None, None, "read-write",
     __repr__    = interp2app(W_ArrayInstance.descr_repr),
     __setitem__ = interp2app(W_ArrayInstance.descr_setitem),
     __getitem__ = interp2app(W_ArrayInstance.descr_getitem),
     __len__     = interp2app(W_ArrayInstance.getlength),
     buffer      = GetSetProperty(W_ArrayInstance.getbuffer),
-    shape       = interp_attrproperty('shape', W_ArrayInstance),
+    shape       = interp_attrproperty_w('shape', W_ArrayInstance),
     free        = interp2app(W_ArrayInstance.free),
     byptr       = interp2app(W_ArrayInstance.byptr),
     itemaddress = interp2app(W_ArrayInstance.descr_itemaddress),
@@ -226,13 +215,13 @@ class W_ArrayInstanceAutoFree(W_ArrayInstance):
             self._free()
 
 W_ArrayInstanceAutoFree.typedef = TypeDef(
-    'ArrayInstanceAutoFree',
+    'ArrayInstanceAutoFree', None, None, "read-write",
     __repr__    = interp2app(W_ArrayInstance.descr_repr),
     __setitem__ = interp2app(W_ArrayInstance.descr_setitem),
     __getitem__ = interp2app(W_ArrayInstance.descr_getitem),
     __len__     = interp2app(W_ArrayInstance.getlength),
     buffer      = GetSetProperty(W_ArrayInstance.getbuffer),
-    shape       = interp_attrproperty('shape', W_ArrayInstance),
+    shape       = interp_attrproperty_w('shape', W_ArrayInstance),
     byptr       = interp2app(W_ArrayInstance.byptr),
     itemaddress = interp2app(W_ArrayInstance.descr_itemaddress),
 )
