@@ -2,6 +2,9 @@
 #define Py_PYTHON_H
 
 /* Compat stuff */
+#ifdef __GNUC__
+#define _GNU_SOURCE 1
+#endif
 #ifndef _WIN32
 # include <inttypes.h>
 # include <stdint.h>
@@ -11,8 +14,8 @@
 # include <errno.h>
 # include <unistd.h>
 # define Py_DEPRECATED(VERSION_UNUSED) __attribute__((__deprecated__))
-# define PyAPI_FUNC(RTYPE) RTYPE
-# define PyAPI_DATA(RTYPE) extern RTYPE
+# define PyAPI_FUNC(RTYPE) __attribute__((visibility("default"))) RTYPE
+# define PyAPI_DATA(RTYPE) extern PyAPI_FUNC(RTYPE)
 # define Py_LOCAL_INLINE(type) static inline type
 #else
 # define MS_WIN32 1
@@ -24,6 +27,7 @@
 #  include <limits.h>
 # endif
 # include <io.h>
+# include <sys/types.h>   /* for 'off_t' */
 # define Py_DEPRECATED(VERSION_UNUSED)
 # ifdef Py_BUILD_CORE
 #  define PyAPI_FUNC(RTYPE) __declspec(dllexport) RTYPE
@@ -46,21 +50,13 @@
 # endif
 #endif
 #ifndef DL_EXPORT
-#       define DL_EXPORT(RTYPE) RTYPE
+#       define DL_EXPORT(RTYPE) PyAPI_FUNC(RTYPE)
 #endif
 #ifndef DL_IMPORT
 #       define DL_IMPORT(RTYPE) RTYPE
 #endif
-
 #include <stdlib.h>
 
-#ifndef _WIN32
-typedef intptr_t Py_ssize_t;
-#else
-typedef long Py_ssize_t;
-#endif
-#define PY_SSIZE_T_MAX ((Py_ssize_t)(((size_t)-1)>>1))
-#define PY_SSIZE_T_MIN (-PY_SSIZE_T_MAX-1)
 #define Py_SAFE_DOWNCAST(VALUE, WIDE, NARROW) (NARROW)(VALUE)
 
 #define Py_USING_UNICODE
@@ -77,13 +73,22 @@ typedef long Py_ssize_t;
 
 #define Py_MEMCPY memcpy
 
-#include <pypy_macros.h>
+#include "pypy_macros.h"
+
+#define PyExc_EnvironmentError PyExc_OSError
+#define PyExc_IOError PyExc_OSError
+// TODO: fix windows support
+// #define PyExc_WindowsError PyExc_OSError
 
 #include "patchlevel.h"
 #include "pyconfig.h"
 
 #include "object.h"
+#include "typeslots.h"
+#include "abstract.h"
+#include "pymath.h"
 #include "pyport.h"
+#include "pymacro.h"
 #include "warnings.h"
 
 #include <stdarg.h>
@@ -98,41 +103,53 @@ typedef long Py_ssize_t;
 #include "complexobject.h"
 #include "methodobject.h"
 #include "funcobject.h"
+#include "code.h"
 
+#include "moduleobject.h"
 #include "modsupport.h"
 #include "pythonrun.h"
 #include "pyerrors.h"
 #include "sysmodule.h"
-#include "stringobject.h"
+#include "bytearrayobject.h"
 #include "descrobject.h"
 #include "tupleobject.h"
 #include "dictobject.h"
-#include "intobject.h"
+#include "longobject.h"
 #include "listobject.h"
+#include "longobject.h"
 #include "unicodeobject.h"
 #include "compile.h"
 #include "frameobject.h"
+#include "memoryobject.h"
 #include "eval.h"
-#include "pymath.h"
 #include "pymem.h"
-#include "pycobject.h"
 #include "pycapsule.h"
-#include "bufferobject.h"
 #include "bytesobject.h"
 #include "sliceobject.h"
+#include "genobject.h"
 #include "datetime.h"
 #include "pystate.h"
 #include "fileobject.h"
 #include "pysignals.h"
 #include "pythread.h"
+#include "traceback.h"
+#include "pylifecycle.h"
 
 /* Missing definitions */
 #include "missing.h"
 
-// XXX This shouldn't be included here
-#include "structmember.h"
-
-#include <pypy_decl.h>
+/* The declarations of most API functions are generated in a separate file */
+/* Don't include them while building PyPy, RPython also generated signatures
+ * which are similar but not identical. */
+#ifndef PYPY_STANDALONE
+#ifdef __cplusplus
+extern "C" {
+#endif
+  #include "pypy_decl.h"
+#ifdef __cplusplus
+}
+#endif
+#endif  /* PYPY_STANDALONE */
 
 /* Define macros for inline documentation. */
 #define PyDoc_VAR(name) static char name[]
