@@ -1,15 +1,28 @@
 import errno
-from pypy.interpreter.error import OperationError
-from pypy.module.cpyext.api import cpython_api
+from pypy.interpreter.error import oefmt
+from pypy.module.cpyext.api import cpython_api, CONST_STRING, INTP_real
 from pypy.module.cpyext.pyobject import PyObject
 from rpython.rlib import rdtoa
 from rpython.rlib import rfloat
 from rpython.rlib import rposix, jit
+from rpython.rlib.rarithmetic import intmask
 from rpython.rtyper.lltypesystem import lltype
 from rpython.rtyper.lltypesystem import rffi
 
 
-@cpython_api([rffi.CCHARP, rffi.CCHARPP, PyObject], rffi.DOUBLE, error=-1.0)
+# PyOS_double_to_string's "type", if non-NULL, will be set to one of:
+Py_DTST_FINITE = 0
+Py_DTST_INFINITE = 1
+Py_DTST_NAN = 2
+
+# Match the "type" back to values in CPython
+DOUBLE_TO_STRING_TYPES_MAP = {
+    rfloat.DIST_FINITE: Py_DTST_FINITE,
+    rfloat.DIST_INFINITY: Py_DTST_INFINITE,
+    rfloat.DIST_NAN: Py_DTST_NAN
+}
+
+@cpython_api([CONST_STRING, rffi.CCHARPP, PyObject], rffi.DOUBLE, error=-1.0)
 @jit.dont_look_inside       # direct use of _get_errno()
 def PyOS_string_to_double(space, s, endptr, w_overflow_exception):
     """Convert a string s to a double, raising a Python
@@ -50,9 +63,8 @@ def PyOS_string_to_double(space, s, endptr, w_overflow_exception):
         endpos = (rffi.cast(rffi.LONG, endptr[0]) -
                   rffi.cast(rffi.LONG, s))
         if endpos == 0 or (not user_endptr and not endptr[0][0] == '\0'):
-            raise OperationError(
-                space.w_ValueError,
-                space.wrap('invalid input at position %s' % endpos))
+            raise oefmt(space.w_ValueError,
+                        "invalid input at position %d", endpos)
         err = rffi.cast(lltype.Signed, rposix._get_errno())
         if err == errno.ERANGE:
             rposix._set_errno(rffi.cast(rffi.INT, 0))
@@ -62,9 +74,47 @@ def PyOS_string_to_double(space, s, endptr, w_overflow_exception):
                 else:
                     return -rfloat.INFINITY
             else:
-                raise OperationError(w_overflow_exception,
-                                     space.wrap('value too large'))
+                raise oefmt(w_overflow_exception, "value too large")
         return result
     finally:
         if not user_endptr:
             lltype.free(endptr, flavor='raw')
+
+@cpython_api([rffi.DOUBLE, lltype.Char, rffi.INT_real, rffi.INT_real, INTP_real], rffi.CCHARP)
+def PyOS_double_to_string(space, val, format_code, precision, flags, ptype):
+    """Convert a double val to a string using supplied
+    format_code, precision, and flags.
+
+    format_code must be one of 'e', 'E', 'f', 'F',
+    'g', 'G' or 'r'.  For 'r', the supplied precision
+    must be 0 and is ignored.  The 'r' format code specifies the
+    standard repr() format.
+
+    flags can be zero or more of the values Py_DTSF_SIGN,
+    Py_DTSF_ADD_DOT_0, or Py_DTSF_ALT, or-ed together:
+
+    Py_DTSF_SIGN means to always precede the returned string with a sign
+    character, even if val is non-negative.
+
+    Py_DTSF_ADD_DOT_0 means to ensure that the returned string will not look
+    like an integer.
+
+    Py_DTSF_ALT means to apply "alternate" formatting rules.  See the
+    documentation for the PyOS_snprintf() '#' specifier for
+    details.
+
+    If ptype is non-NULL, then the value it points to will be set to one of
+    Py_DTST_FINITE, Py_DTST_INFINITE, or Py_DTST_NAN, signifying that
+    val is a finite number, an infinite number, or not a number, respectively.
+
+    The return value is a pointer to buffer with the converted string or
+    NULL if the conversion failed. The caller is responsible for freeing the
+    returned string by calling PyMem_Free().
+    """
+    buffer, rtype = rfloat.double_to_string(val, format_code,
+                                            intmask(precision),
+                                            intmask(flags))
+    if ptype != lltype.nullptr(INTP_real.TO):
+        ptype[0] = rffi.cast(rffi.INT_real, DOUBLE_TO_STRING_TYPES_MAP[rtype])
+    bufp = rffi.str2charp(buffer)
+    return bufp
