@@ -1,9 +1,8 @@
-from pypy.module.cpyext.api import fopen, fclose, fwrite
+import pytest
 from pypy.module.cpyext.test.test_api import BaseApiTest
 from pypy.module.cpyext.object import Py_PRINT_RAW
-from rpython.rtyper.lltypesystem import rffi, lltype
+from rpython.rtyper.lltypesystem import rffi
 from rpython.tool.udir import udir
-import pytest
 
 class TestFile(BaseApiTest):
 
@@ -14,11 +13,7 @@ class TestFile(BaseApiTest):
         rffi.free_charp(filename)
         rffi.free_charp(mode)
 
-        assert api.PyFile_Check(w_file)
-        assert api.PyFile_CheckExact(w_file)
-        assert not api.PyFile_Check(space.wrap("text"))
-
-        space.call_method(w_file, "write", space.wrap("text"))
+        space.call_method(w_file, "write", space.newbytes("text"))
         space.call_method(w_file, "close")
         assert (udir / "_test_file").read() == "text"
 
@@ -53,27 +48,11 @@ class TestFile(BaseApiTest):
 
         space.call_method(w_file, "close")
 
-    def test_file_name(self, space, api):
-        name = str(udir / "_test_file")
-        with rffi.scoped_str2charp(name) as filename:
-            with rffi.scoped_str2charp("wb") as mode:
-                w_file = api.PyFile_FromString(filename, mode)
-        assert space.str_w(api.PyFile_Name(w_file)) == name
-
-    @pytest.mark.xfail
-    def test_file_fromfile(self, space, api):
-        api.PyFile_Fromfile()
-
-    @pytest.mark.xfail
-    def test_file_setbufsize(self, space, api):
-        api.PyFile_SetBufSize()
-
     def test_file_writestring(self, space, api, capfd):
-        s = rffi.str2charp("test\n")
-        try:
-            api.PyFile_WriteString(s, space.sys.get("stdout"))
-        finally:
-            rffi.free_charp(s)
+        w_stdout = space.sys.get("stdout")
+        with rffi.scoped_str2charp("test\n") as s:
+            api.PyFile_WriteString(s, w_stdout)
+        space.call_method(w_stdout, "flush")
         out, err = capfd.readouterr()
         out = out.replace('\r\n', '\n')
         assert out == "test\n"
@@ -87,19 +66,3 @@ class TestFile(BaseApiTest):
         out, err = capfd.readouterr()
         out = out.replace('\r\n', '\n')
         assert out == "test\n'test\\n'"
-
-    def test_file_softspace(self, space, api, capfd):
-        w_stdout = space.sys.get("stdout")
-        assert api.PyFile_SoftSpace(w_stdout, 1) == 0
-        assert api.PyFile_SoftSpace(w_stdout, 0) == 1
-        
-        api.PyFile_SoftSpace(w_stdout, 1)
-        w_ns = space.newdict()
-        space.exec_("print 1,", w_ns, w_ns)
-        space.exec_("print 2,", w_ns, w_ns)
-        api.PyFile_SoftSpace(w_stdout, 0)
-        space.exec_("print 3", w_ns, w_ns)
-        space.call_method(w_stdout, "flush")
-        out, err = capfd.readouterr()
-        out = out.replace('\r\n', '\n')
-        assert out == " 1 23\n"
