@@ -1,3 +1,4 @@
+import pytest
 from rpython.tool.udir import udir
 
 
@@ -9,18 +10,19 @@ class AppTestMarshal:
         cls.w_tmpfile = cls.space.wrap(str(tmpfile))
 
     def w_marshal_check(self, case):
-        import marshal, StringIO
+        import marshal
+        from io import BytesIO
         s = marshal.dumps(case)
         print(repr(s))
         x = marshal.loads(s)
         assert x == case and type(x) is type(case)
 
-        exc = raises(TypeError, marshal.loads, memoryview(s))
-        assert str(exc.value) == "must be string or read-only buffer, not memoryview"
+        y = marshal.loads(memoryview(s))
+        assert y == case and type(y) is type(case)
 
         import sys
         if '__pypy__' in sys.builtin_module_names:
-            f = StringIO.StringIO()
+            f = BytesIO()
             marshal.dump(case, f)
             f.seek(0)
             x = marshal.load(f)
@@ -55,9 +57,9 @@ class AppTestMarshal:
         case = -17
         self.marshal_check(case)
 
-    def test_sys_dot_maxint(self):
+    def test_sys_dot_maxsize(self):
         import sys
-        case = sys.maxint
+        case = sys.maxsize
         self.marshal_check(case)
 
     def test__minus_1_dot_25(self):
@@ -77,8 +79,7 @@ class AppTestMarshal:
         self.marshal_check(case)
 
     def test_long(self):
-        self.marshal_check(42L)
-        case = -1234567890123456789012345678901234567890L
+        case = -1234567890123456789012345678901234567890
         self.marshal_check(case)
 
     def test_hello_____not_interned(self):
@@ -118,18 +119,18 @@ class AppTestMarshal:
     def test_func_dot_func_code(self):
         def func(x):
             return lambda y: x+y
-        case = func.func_code
+        case = func.__code__
         self.marshal_check(case)
 
     def test_scopefunc_dot_func_code(self):
         def func(x):
             return lambda y: x+y
         scopefunc = func(42)
-        case = scopefunc.func_code
+        case = scopefunc.__code__
         self.marshal_check(case)
 
-    def test_u_quote_hello_quote_(self):
-        case = u'hello'
+    def test_b_quote_hello_quote_(self):
+        case = b'hello'
         self.marshal_check(case)
 
     def test_set_brace__ecarb_(self):
@@ -157,7 +158,7 @@ class AppTestMarshal:
         f = open(self.tmpfile, 'wb')
         marshal.dump(obj1, f)
         marshal.dump(obj2, f)
-        f.write('END')
+        f.write(b'END')
         f.close()
         f = open(self.tmpfile, 'rb')
         obj1b = marshal.load(f)
@@ -166,18 +167,20 @@ class AppTestMarshal:
         f.close()
         assert obj1b == obj1
         assert obj2b == obj2
-        assert tail == 'END'
+        assert tail == b'END'
 
     def test_unicode(self):
         import marshal, sys
-        self.marshal_check(u'\uFFFF')
-        self.marshal_check(u'\ud800')
+        self.marshal_check('\uFFFF')
+        self.marshal_check('\ud800')
+        c = u"\ud800"
+        self.marshal_check(c + u'\udc00')
 
-        self.marshal_check(unichr(sys.maxunicode))
+        self.marshal_check(chr(sys.maxunicode))
 
     def test_reject_subtypes(self):
         import marshal
-        types = (float, complex, int, long, tuple, list, dict, set, frozenset)
+        types = (float, complex, int, tuple, list, dict, set, frozenset)
         for cls in types:
             print(cls)
             class subtype(cls):
@@ -186,27 +189,61 @@ class AppTestMarshal:
             assert str(exc.value) == 'unmarshallable object'
             exc = raises(ValueError, marshal.dumps, subtype())
             assert str(exc.value) == 'unmarshallable object'
+            exc = raises(ValueError, marshal.dumps, (subtype(),))
+            assert str(exc.value) == 'unmarshallable object'
 
     def test_valid_subtypes(self):
         import marshal
         from array import array
         class subtype(array):
             pass
-        assert marshal.dumps(subtype('c', 'test')) == marshal.dumps(array('c', 'test'))
+        assert marshal.dumps(subtype('b', b'test')) == marshal.dumps(array('b', b'test'))
 
     def test_bad_typecode(self):
         import marshal
-        exc = raises(ValueError, marshal.loads, chr(1))
-        assert str(exc.value) == "bad marshal data (unknown type code)"
+        exc = raises(ValueError, marshal.loads, bytes([1]))
+        assert str(exc.value).startswith("bad marshal data (unknown type code")
+
+    def test_bad_data(self):
+        # If you have sufficiently little memory, the line at the end of the
+        # test will fail immediately.  If not, the test will consume high
+        # amounts of memory and make your system unstable.  CPython (I tried
+        # 3.3 and 3.5) shows the same behaviour on my computers (4 GB and 12 GB).
+        skip("takes too much memory")
+
+        import marshal
+        # Yes, there is code that depends on this :-(
+        raises(EOFError, marshal.loads, b'<test>')
+        raises((MemoryError, ValueError), marshal.loads, b'(test)')
+
+    def test_bad_reader(self):
+        import marshal, io
+        class BadReader(io.BytesIO):
+            def read(self, n=-1):
+                b = super().read(n)
+                if n is not None and n > 4:
+                    b += b' ' * 10**6
+                return b
+        for value in (1.0, 1j, b'0123456789', '0123456789'):
+            raises(ValueError, marshal.load,
+                   BadReader(marshal.dumps(value)))
 
 
+@pytest.mark.skipif('config.option.runappdirect or sys.maxint > 2 ** 32')
 class AppTestSmallLong(AppTestMarshal):
     spaceconfig = AppTestMarshal.spaceconfig.copy()
     spaceconfig["objspace.std.withsmalllong"] = True
 
+    def setup_class(cls):
+        from pypy.interpreter import gateway
+        from pypy.objspace.std.smalllongobject import W_SmallLongObject
+        def w__small(space, w_obj):
+            return W_SmallLongObject.frombigint(space.bigint_w(w_obj))
+        cls.w__small = cls.space.wrap(gateway.interp2app(w__small))
+
     def test_smalllong(self):
         import __pypy__
-        x = -123456789012345L
+        x = self._small(-123456789012345)
         assert 'SmallLong' in __pypy__.internal_repr(x)
         y = self.marshal_check(x)
         assert y == x
