@@ -1,9 +1,11 @@
-from pypy.interpreter.error import OperationError, wrap_oserror
+from pypy.interpreter.error import OperationError, oefmt, wrap_oserror
 from pypy.interpreter.baseobjspace import W_Root
-from pypy.interpreter.typedef import TypeDef
-from pypy.interpreter.gateway import interp2app, unwrap_spec
-from rpython.rlib import rmmap, rarithmetic
-from rpython.rlib.buffer import Buffer
+from pypy.interpreter.typedef import TypeDef, GetSetProperty, make_weakref_descr
+from pypy.interpreter.gateway import interp2app, unwrap_spec, WrappedDefault
+from pypy.interpreter.buffer import SimpleView
+
+from rpython.rlib import rmmap, rarithmetic, objectmodel
+from rpython.rlib.buffer import RawBuffer
 from rpython.rlib.rmmap import RValueError, RTypeError, RMMapError
 from rpython.rlib.rstring import StringBuilder
 
@@ -18,9 +20,13 @@ class W_MMap(W_Root):
         self.space = space
         self.mmap = mmap_obj
 
-    def readbuf_w(self, space):
+    def buffer_w(self, space, flags):
         self.check_valid()
-        return MMapBuffer(self.space, self.mmap, True)
+        readonly = (self.mmap.access == ACCESS_READ)
+        write_required = bool(flags & space.BUF_WRITABLE)
+        if write_required and readonly:
+            raise oefmt(space.w_BufferError, "Object is not writable.")
+        return SimpleView(MMapBuffer(self.space, self.mmap, readonly))
 
     def close(self):
         self.mmap.close()
@@ -28,23 +34,27 @@ class W_MMap(W_Root):
     def read_byte(self):
         self.check_valid()
         try:
-            return self.space.wrap(self.mmap.read_byte())
-        except RValueError, v:
+            return self.space.newint(ord(self.mmap.read_byte()))
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
     def readline(self):
         self.check_valid()
-        return self.space.wrap(self.mmap.readline())
+        return self.space.newbytes(self.mmap.readline())
 
-    @unwrap_spec(num=int)
-    def read(self, num=-1):
+    @unwrap_spec(w_num=WrappedDefault(None))
+    def read(self, w_num):
         self.check_valid()
-        return self.space.wrap(self.mmap.read(num))
+        if self.space.is_none(w_num):
+            num = -1
+        else:
+            num = self.space.int_w(w_num)
+        return self.space.newbytes(self.mmap.read(num))
 
     def find(self, w_tofind, w_start=None, w_end=None):
         self.check_valid()
         space = self.space
-        tofind = space.getarg_w('s#', w_tofind)
+        tofind = space.charbuf_w(w_tofind)
         if w_start is None:
             start = self.mmap.pos
         else:
@@ -53,12 +63,12 @@ class W_MMap(W_Root):
             end = self.mmap.size
         else:
             end = space.getindex_w(w_end, None)
-        return space.wrap(self.mmap.find(tofind, start, end))
+        return space.newint(self.mmap.find(tofind, start, end))
 
     def rfind(self, w_tofind, w_start=None, w_end=None):
         self.check_valid()
         space = self.space
-        tofind = space.getarg_w('s#', w_tofind)
+        tofind = space.charbuf_w(w_tofind)
         if w_start is None:
             start = self.mmap.pos
         else:
@@ -67,53 +77,53 @@ class W_MMap(W_Root):
             end = self.mmap.size
         else:
             end = space.getindex_w(w_end, None)
-        return space.wrap(self.mmap.find(tofind, start, end, True))
+        return space.newint(self.mmap.find(tofind, start, end, True))
 
     @unwrap_spec(pos=OFF_T, whence=int)
     def seek(self, pos, whence=0):
         self.check_valid()
         try:
             self.mmap.seek(pos, whence)
-        except RValueError, v:
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
     def tell(self):
         self.check_valid()
-        return self.space.wrap(self.mmap.tell())
+        return self.space.newint(self.mmap.tell())
 
     def descr_size(self):
         self.check_valid()
         try:
-            return self.space.wrap(self.mmap.file_size())
-        except OSError, e:
+            return self.space.newint(self.mmap.file_size())
+        except OSError as e:
             raise mmap_error(self.space, e)
 
     def write(self, w_data):
         self.check_valid()
-        data = self.space.getarg_w('s#', w_data)
+        data = self.space.charbuf_w(w_data)
         self.check_writeable()
         try:
             self.mmap.write(data)
-        except RValueError, v:
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
-    @unwrap_spec(byte=str)
+    @unwrap_spec(byte=int)
     def write_byte(self, byte):
         self.check_valid()
         self.check_writeable()
         try:
-            self.mmap.write_byte(byte)
-        except RMMapError, v:
+            self.mmap.write_byte(chr(byte))
+        except RMMapError as v:
             raise mmap_error(self.space, v)
 
     @unwrap_spec(offset=int, size=int)
     def flush(self, offset=0, size=0):
         self.check_valid()
         try:
-            return self.space.wrap(self.mmap.flush(offset, size))
-        except RValueError, v:
+            return self.space.newint(self.mmap.flush(offset, size))
+        except RValueError as v:
             raise mmap_error(self.space, v)
-        except OSError, e:
+        except OSError as e:
             raise mmap_error(self.space, e)
 
     @unwrap_spec(dest=int, src=int, count=int)
@@ -122,7 +132,7 @@ class W_MMap(W_Root):
         self.check_writeable()
         try:
             self.mmap.move(dest, src, count)
-        except RValueError, v:
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
     @unwrap_spec(newsize=int)
@@ -131,33 +141,40 @@ class W_MMap(W_Root):
         self.check_resizeable()
         try:
             self.mmap.resize(newsize)
-        except OSError, e:
+        except OSError as e:
             raise mmap_error(self.space, e)
-        except RValueError, e:
+        except RValueError as e:
             # obscure: in this case, RValueError translates to an app-level
             # SystemError.
             raise OperationError(self.space.w_SystemError,
-                                 self.space.wrap(e.message))
+                                 self.space.newtext(e.message))
 
     def __len__(self):
-        return self.space.wrap(self.mmap.size)
+        return self.space.newint(self.mmap.size)
+
+    def closed_get(self, space):
+        try:
+            self.mmap.check_valid()
+        except RValueError:
+            return space.w_True
+        return space.w_False
 
     def check_valid(self):
         try:
             self.mmap.check_valid()
-        except RValueError, v:
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
     def check_writeable(self):
         try:
             self.mmap.check_writeable()
-        except RMMapError, v:
+        except RMMapError as v:
             raise mmap_error(self.space, v)
 
     def check_resizeable(self):
         try:
             self.mmap.check_resizeable()
-        except RMMapError, v:
+        except RMMapError as v:
             raise mmap_error(self.space, v)
 
     def descr_getitem(self, w_index):
@@ -166,35 +183,34 @@ class W_MMap(W_Root):
         space = self.space
         start, stop, step, length = space.decode_index4(w_index, self.mmap.size)
         if step == 0:  # index only
-            return space.wrap(self.mmap.getitem(start))
+            return space.newint(ord(self.mmap.getitem(start)))
         elif step == 1:
             if stop - start < 0:
-                return space.wrap("")
-            return space.wrap(self.mmap.getslice(start, length))
+                return space.newbytes("")
+            return space.newbytes(self.mmap.getslice(start, length))
         else:
             b = StringBuilder(length)
             for i in range(start, stop, step):
                 b.append(self.mmap.getitem(i))
-            return space.wrap(b.build())
+            return space.newbytes(b.build())
 
     def descr_setitem(self, w_index, w_value):
         space = self.space
-        value = space.realstr_w(w_value)
         self.check_valid()
-
         self.check_writeable()
 
         start, stop, step, length = space.decode_index4(w_index, self.mmap.size)
         if step == 0:  # index only
-            if len(value) != 1:
-                raise OperationError(space.w_ValueError,
-                                     space.wrap("mmap assignment must be "
-                                                "single-character string"))
-            self.mmap.setitem(start, value)
+            value = space.int_w(w_value)
+            if not 0 <= value < 256:
+                raise oefmt(space.w_ValueError,
+                            "mmap item value must be in range(0, 256)")
+            self.mmap.setitem(start, chr(value))
         else:
+            value = space.bytes_w(w_value)
             if len(value) != length:
-                raise OperationError(space.w_ValueError,
-                          space.wrap("mmap slice assignment is wrong size"))
+                raise oefmt(space.w_ValueError,
+                            "mmap slice assignment is wrong size")
             if step == 1:
                 self.mmap.setslice(start, value)
             else:
@@ -202,45 +218,13 @@ class W_MMap(W_Root):
                     self.mmap.setitem(start, value[i])
                     start += step
 
-    def descr_getslice(self, space, w_ilow, w_ihigh):
+    def descr_enter(self, space):
         self.check_valid()
-        i = space.getindex_w(w_ilow, None)
-        j = space.getindex_w(w_ihigh, None)
-        if i < 0:
-            i = 0
-        elif i > self.mmap.size:
-            i = self.mmap.size
-        if j < 0:
-            j = 0
-        if j < i:
-            j = i
-        elif j > self.mmap.size:
-            j = self.mmap.size
-        return space.wrap(self.mmap.getslice(i, (j - i)))
+        return self
+
+    def descr_exit(self, space, __args__):
+        self.close()
 
-    def descr_setslice(self, space, w_ilow, w_ihigh, w_item):
-        self.check_valid()
-        i = space.getindex_w(w_ilow, None)
-        j = space.getindex_w(w_ihigh, None)
-        if i < 0:
-            i = 0
-        elif i > self.mmap.size:
-            i = self.mmap.size
-        if j < 0:
-            j = 0
-        if j < i:
-            j = i
-        elif j > self.mmap.size:
-            j = self.mmap.size
-        if not space.isinstance_w(w_item, space.w_str):
-            raise OperationError(space.w_IndexError, space.wrap(
-                "mmap slice assignment must be a string"))
-        value = space.realstr_w(w_item)
-        if len(value) != (j - i):
-            raise OperationError(space.w_IndexError, space.wrap(
-                "mmap slice assignment is wrong size"))
-        self.check_writeable()
-        self.mmap.setslice(i, value)
 
 if rmmap._POSIX:
 
@@ -254,15 +238,15 @@ if rmmap._POSIX:
             W_MMap.__init__(self, space,
                             rmmap.mmap(fileno, length, flags, prot, access,
                                        offset))
-        except OSError, e:
+        except OSError as e:
             raise mmap_error(space, e)
-        except RMMapError, e:
+        except RMMapError as e:
             raise mmap_error(space, e)
-        return space.wrap(self)
+        return self
 
 elif rmmap._MS_WINDOWS:
 
-    @unwrap_spec(fileno=int, length=int, tagname=str,
+    @unwrap_spec(fileno=int, length=int, tagname='text',
                  access=int, offset=OFF_T)
     def mmap(space, w_subtype, fileno, length, tagname="",
              access=rmmap._ACCESS_DEFAULT, offset=0):
@@ -271,13 +255,13 @@ elif rmmap._MS_WINDOWS:
             W_MMap.__init__(self, space,
                             rmmap.mmap(fileno, length, tagname, access,
                                        offset))
-        except OSError, e:
+        except OSError as e:
             raise mmap_error(space, e)
-        except RMMapError, e:
+        except RMMapError as e:
             raise mmap_error(space, e)
-        return space.wrap(self)
+        return self
 
-W_MMap.typedef = TypeDef("mmap.mmap",
+W_MMap.typedef = TypeDef("mmap.mmap", None, None, 'read-write',
     __new__ = interp2app(mmap),
     close = interp2app(W_MMap.close),
     read_byte = interp2app(W_MMap.read_byte),
@@ -297,8 +281,11 @@ W_MMap.typedef = TypeDef("mmap.mmap",
     __len__ = interp2app(W_MMap.__len__),
     __getitem__ = interp2app(W_MMap.descr_getitem),
     __setitem__ = interp2app(W_MMap.descr_setitem),
-    __getslice__ = interp2app(W_MMap.descr_getslice),
-    __setslice__ = interp2app(W_MMap.descr_setslice),
+    __enter__ = interp2app(W_MMap.descr_enter),
+    __exit__ = interp2app(W_MMap.descr_exit),
+    __weakref__ = make_weakref_descr(W_MMap),
+
+    closed = GetSetProperty(W_MMap.closed_get),
 )
 
 constants = rmmap.constants
@@ -308,28 +295,21 @@ ACCESS_READ  = rmmap.ACCESS_READ
 ACCESS_WRITE = rmmap.ACCESS_WRITE
 ACCESS_COPY  = rmmap.ACCESS_COPY
 
-class Cache:
-    def __init__(self, space):
-        self.w_error = space.new_exception_class("mmap.error",
-                                                 space.w_EnvironmentError)
 
+@objectmodel.dont_inline
 def mmap_error(space, e):
     if isinstance(e, RValueError):
-        return OperationError(space.w_ValueError,
-                              space.wrap(e.message))
+        return OperationError(space.w_ValueError, space.newtext(e.message))
     elif isinstance(e, RTypeError):
-        return OperationError(space.w_TypeError,
-                              space.wrap(e.message))
+        return OperationError(space.w_TypeError, space.newtext(e.message))
     elif isinstance(e, OSError):
-        w_error = space.fromcache(Cache).w_error
-        return wrap_oserror(space, e, w_exception_class=w_error)
+        return wrap_oserror(space, e)
     else:
         # bogus 'e'?
-        return OperationError(space.w_SystemError, space.wrap('%s' % e))
-mmap_error._dont_inline_ = True
+        return OperationError(space.w_SystemError, space.newtext('%s' % e))
 
 
-class MMapBuffer(Buffer):
+class MMapBuffer(RawBuffer):
     _immutable_ = True
 
     def __init__(self, space, mmap, readonly):
@@ -349,7 +329,7 @@ class MMapBuffer(Buffer):
         if step == 1:
             return self.mmap.getslice(start, size)
         else:
-            return Buffer.getslice(self, start, stop, step, size)
+            return RawBuffer.getslice(self, start, stop, step, size)
 
     def setitem(self, index, char):
         self.check_valid_writeable()
@@ -366,12 +346,12 @@ class MMapBuffer(Buffer):
     def check_valid(self):
         try:
             self.mmap.check_valid()
-        except RValueError, v:
+        except RValueError as v:
             raise mmap_error(self.space, v)
 
     def check_valid_writeable(self):
         try:
             self.mmap.check_valid()
             self.mmap.check_writeable()
-        except RMMapError, v:
+        except RMMapError as v:
             raise mmap_error(self.space, v)
