@@ -8,7 +8,8 @@ from rpython.rtyper.lltypesystem import lltype, rffi
 from rpython.translator import cdir
 from rpython.translator.tool.cbuild import ExternalCompilationInfo
 
-from pypy.interpreter.error import oefmt
+from pypy.interpreter.error import OperationError, oefmt
+from pypy.interpreter.unicodehelper import encode
 
 cwd = py.path.local(__file__).dirpath()
 eci = ExternalCompilationInfo(
@@ -27,10 +28,6 @@ pypy_tscmp = llexternal(
     'pypy_tscmp',
     [rffi.CCHARP, rffi.CCHARP, rffi.LONG, rffi.LONG],
     rffi.INT)
-pypy_tscmp_wide = llexternal(
-    'pypy_tscmp_wide',
-    [rffi.CWCHARP, rffi.CWCHARP, rffi.LONG, rffi.LONG],
-    rffi.INT)
 
 
 def compare_digest(space, w_a, w_b):
@@ -47,27 +44,22 @@ def compare_digest(space, w_a, w_b):
     """
     if (space.isinstance_w(w_a, space.w_unicode) and
         space.isinstance_w(w_b, space.w_unicode)):
-        a = space.unicode_w(w_a)
-        b = space.unicode_w(w_b)
-        with rffi.scoped_nonmoving_unicodebuffer(a) as a_buf:
-            with rffi.scoped_nonmoving_unicodebuffer(b) as b_buf:
-                result = pypy_tscmp_wide(a_buf, b_buf, len(a), len(b))
-        return space.wrap(rffi.cast(lltype.Bool, result))
+        try:
+            w_a = encode(space, w_a, 'ascii')
+            w_b = encode(space, w_b, 'ascii')
+        except OperationError as e:
+            if not e.match(space, space.w_UnicodeEncodeError):
+                raise
+            raise oefmt(space.w_TypeError,
+                        "comparing strings with non-ASCII characters is not "
+                        "supported")
     return compare_digest_buffer(space, w_a, w_b)
 
 
 def compare_digest_buffer(space, w_a, w_b):
-    try:
-        a_buf = w_a.buffer_w(space, space.BUF_SIMPLE)
-        b_buf = w_b.buffer_w(space, space.BUF_SIMPLE)
-    except TypeError:
-        raise oefmt(space.w_TypeError,
-                    "unsupported operand types(s) or combination of types: "
-                    "'%T' and '%T'", w_a, w_b)
-
-    a = a_buf.as_str()
-    b = b_buf.as_str()
+    a = space.charbuf_w(w_a)
+    b = space.charbuf_w(w_b)
     with rffi.scoped_nonmovingbuffer(a) as a_buf:
         with rffi.scoped_nonmovingbuffer(b) as b_buf:
             result = pypy_tscmp(a_buf, b_buf, len(a), len(b))
-    return space.wrap(rffi.cast(lltype.Bool, result))
+    return space.newbool(rffi.cast(lltype.Bool, result))
