@@ -2,7 +2,9 @@ from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.typedef import TypeDef, GetSetProperty
 from pypy.interpreter.gateway import interp2app, unwrap_spec, WrappedDefault
 from pypy.interpreter.error import OperationError, oefmt
+from pypy.interpreter.unicodehelper import encode_utf8
 from rpython.rlib import rgc, jit
+from rpython.rlib.objectmodel import specialize
 from rpython.rtyper.lltypesystem import rffi, lltype
 from rpython.rtyper.tool import rffi_platform
 from rpython.translator.tool.cbuild import ExternalCompilationInfo
@@ -118,18 +120,6 @@ class CConfigure:
         locals()[name] = rffi_platform.ConstantInteger(name)
     for name in xml_model_list:
         locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
-    for name in xml_model_list:
-        locals()[name] = rffi_platform.ConstantInteger(name)
     XML_Parser_SIZE = rffi_platform.SizeOf("XML_Parser")
 
 for k, v in rffi_platform.configure(CConfigure).items():
@@ -255,9 +245,10 @@ for index, (name, params) in enumerate(HANDLERS.items()):
                 'w_arg%d = parser.w_convert_model(space, arg%d)' % (i, i))
             converters.append(
                 'XML_FreeContentModel(parser.itself, arg%d)' % (i,))
+        elif ARG == rffi.INT:
+            converters.append("w_arg%d = space.newint(arg%d)" % (i, i))
         else:
-            converters.append(
-                'w_arg%d = space.wrap(arg%d)' % (i, i))
+            assert 0, "missing conversion case"
         real_params.append(ARG)
     converters = '; '.join(converters)
 
@@ -338,7 +329,7 @@ def UnknownEncodingHandlerData_callback(ll_userdata, name, info):
 
     try:
         parser.UnknownEncodingHandler(space, name, info)
-    except OperationError, e:
+    except OperationError as e:
         if not parser._exc_info:
             parser._exc_info = e
         XML_StopParser(parser.itself, XML_FALSE)
@@ -351,6 +342,8 @@ callback_type = lltype.Ptr(lltype.FuncType(
 XML_SetUnknownEncodingHandler = expat_external(
     'XML_SetUnknownEncodingHandler',
     [XML_Parser, callback_type, rffi.VOIDP], lltype.Void)
+XML_SetEncoding = expat_external(
+    'XML_SetEncoding', [XML_Parser, rffi.CCHARP], rffi.INT)
 
 # Declarations of external functions
 
@@ -407,13 +400,13 @@ XML_ExpatVersion = expat_external(
     'XML_ExpatVersion', [], rffi.CCHARP)
 
 def get_expat_version(space):
-    return space.wrap(rffi.charp2str(XML_ExpatVersion()))
+    return space.newtext(rffi.charp2str(XML_ExpatVersion()))
 
 def get_expat_version_info(space):
     return space.newtuple([
-        space.wrap(XML_MAJOR_VERSION),
-        space.wrap(XML_MINOR_VERSION),
-        space.wrap(XML_MICRO_VERSION)])
+        space.newint(XML_MAJOR_VERSION),
+        space.newint(XML_MINOR_VERSION),
+        space.newint(XML_MICRO_VERSION)])
 
 class Cache:
     def __init__(self, space):
@@ -421,14 +414,17 @@ class Cache:
 
 
 class W_XMLParserType(W_Root):
+    id = -1
+
     def __init__(self, space, parser, w_intern):
         self.itself = parser
+        self.register_finalizer(space)
 
         self.w_intern = w_intern
 
-        self.returns_unicode = True
         self.ordered_attributes = False
         self.specified_attributes = False
+        self.ns_prefixes = False
 
         self.handlers = [None] * NB_HANDLERS
 
@@ -444,14 +440,17 @@ class W_XMLParserType(W_Root):
             CallbackData(space, self))
         XML_SetUserData(self.itself, rffi.cast(rffi.VOIDP, self.id))
 
-    def __del__(self):
+    def _finalize_(self):
         if XML_ParserFree: # careful with CPython interpreter shutdown
-            XML_ParserFree(self.itself)
-        if global_storage:
+            if self.itself:
+                XML_ParserFree(self.itself)
+                self.itself = lltype.nullptr(XML_Parser.TO)
+        if global_storage and self.id >= 0:
             try:
                 global_storage.free_nonmoving_id(self.id)
             except KeyError:
                 pass    # maybe global_storage.clear() was already called
+            self.id = -1
 
     @unwrap_spec(flag=int)
     def SetParamEntityParsing(self, space, flag):
@@ -477,11 +476,8 @@ getting the advantage of providing document type information to the parser.
     # Handlers management
 
     def w_convert(self, space, s):
-        if self.returns_unicode:
-            from pypy.interpreter.unicodehelper import decode_utf8
-            return space.wrap(decode_utf8(space, s))
-        else:
-            return space.wrap(s)
+        from pypy.interpreter.unicodehelper import decode_utf8
+        return space.newtext(s)
 
     def w_convert_charp(self, space, data):
         if data:
@@ -498,7 +494,7 @@ getting the advantage of providing document type information to the parser.
 
         try:
             return space.getitem(self.w_intern, w_data)
-        except OperationError, e:
+        except OperationError as e:
             if not e.match(space, space.w_KeyError):
                 raise
         space.setitem(self.w_intern, w_data, w_data)
@@ -516,8 +512,8 @@ getting the advantage of providing document type information to the parser.
             maxindex = XML_GetSpecifiedAttributeCount(self.itself)
         else:
             maxindex = 0
-        while attrs[maxindex]:
-            maxindex += 2 # copied
+            while attrs[maxindex]:
+                maxindex += 2 # copied
 
         if self.ordered_attributes:
             w_attrs = space.newlist([
@@ -537,8 +533,8 @@ getting the advantage of providing document type information to the parser.
         children = [self.w_convert_model(space, model.c_children[i])
                     for i in range(model.c_numchildren)]
         return space.newtuple([
-            space.wrap(model.c_type),
-            space.wrap(model.c_quant),
+            space.newint(model.c_type),
+            space.newint(model.c_quant),
             self.w_convert_charp(space, model.c_name),
             space.newtuple(children)])
 
@@ -565,6 +561,7 @@ getting the advantage of providing document type information to the parser.
             return self.w_character_data_handler or space.w_None
         return self.handlers[index]
 
+    @specialize.arg(2)
     def sethandler(self, space, name, w_handler, index, setter, handler):
         if name == 'CharacterDataHandler':
             self.flush_character_buffer(space)
@@ -576,16 +573,14 @@ getting the advantage of providing document type information to the parser.
         self.handlers[index] = w_handler
         setter(self.itself, handler)
 
-    sethandler._annspecialcase_ = 'specialize:arg(2)'
-
     all_chars = ''.join(chr(i) for i in range(256))
 
     def UnknownEncodingHandler(self, space, name, info):
         # Yes, supports only 8bit encodings
         translationmap = space.unicode_w(
             space.call_method(
-                space.wrap(self.all_chars), "decode",
-                space.wrap(name), space.wrap("replace")))
+                space.newbytes(self.all_chars), "decode",
+                space.newtext(name), space.newtext("replace")))
 
         if len(translationmap) != 256:
             raise oefmt(space.w_ValueError,
@@ -620,20 +615,25 @@ getting the advantage of providing document type information to the parser.
 
 
     def get_namespace_prefixes(self, space):
-        raise OperationError(space.w_AttributeError,
-            space.wrap("not implemented: reading namespace_prefixes"))
+        return space.newbool(self.ns_prefixes)
 
-    @unwrap_spec(value=int)
-    def set_namespace_prefixes(self, space, value):
-        XML_SetReturnNSTriplet(self.itself, bool(value))
+    def set_namespace_prefixes(self, space, w_value):
+        self.ns_prefixes = space.bool_w(w_value)
+        XML_SetReturnNSTriplet(self.itself, self.ns_prefixes)
 
     # Parse methods
 
-    @unwrap_spec(data=str, isfinal=bool)
-    def Parse(self, space, data, isfinal=False):
+    @unwrap_spec(isfinal=int)
+    def Parse(self, space, w_data, isfinal=False):
         """Parse(data[, isfinal])
 Parse XML data.  `isfinal' should be true at end of input."""
-
+        if space.isinstance_w(w_data, space.w_unicode):
+            data = encode_utf8(space, w_data.unicode_w(space))
+            # Explicitly set UTF-8 encoding. Return code ignored.
+            XML_SetEncoding(self.itself, "utf-8")
+        else:
+            data = space.charbuf_w(w_data)
+        isfinal = bool(isfinal)
         res = XML_Parse(self.itself, data, len(data), isfinal)
         if self._exc_info:
             e = self._exc_info
@@ -643,20 +643,19 @@ Parse XML data.  `isfinal' should be true at end of input."""
             exc = self.set_error(space, XML_GetErrorCode(self.itself))
             raise exc
         self.flush_character_buffer(space)
-        return space.wrap(res)
+        return space.newint(res)
 
     def ParseFile(self, space, w_file):
         """ParseFile(file)
 Parse XML data from file-like object."""
         eof = False
         while not eof:
-            w_data = space.call_method(w_file, 'read', space.wrap(2048))
-            data = space.str_w(w_data)
-            eof = len(data) == 0
-            w_res = self.Parse(space, data, isfinal=eof)
+            w_data = space.call_method(w_file, 'read', space.newint(2048))
+            eof = space.len_w(w_data) == 0
+            w_res = self.Parse(space, w_data, isfinal=eof)
         return w_res
 
-    @unwrap_spec(base=str)
+    @unwrap_spec(base='text')
     def SetBase(self, space, base):
         XML_SetBase(self.itself, base)
 
@@ -667,12 +666,12 @@ information passed to the ExternalEntityRefHandler."""
         if space.is_w(w_context, space.w_None):
             context = None
         else:
-            context = space.str_w(w_context)
+            context = space.text_w(w_context)
 
         if space.is_none(w_encoding):
             encoding = None
         else:
-            encoding = space.str_w(w_encoding)
+            encoding = space.text_w(w_encoding)
 
         xmlparser = XML_ExternalEntityParserCreate(
             self.itself, context, encoding)
@@ -685,13 +684,13 @@ information passed to the ExternalEntityRefHandler."""
         for i in range(NB_HANDLERS):
             parser.handlers[i] = self.handlers[i]
 
-        return space.wrap(parser)
+        return parser
 
     def flush_character_buffer(self, space):
         if not self.buffer:
             return
         w_data = space.call_function(
-            space.getattr(space.wrap(''), space.wrap('join')),
+            space.getattr(space.newtext(''), space.newtext('join')),
             space.newlist(self.buffer))
         self.buffer = []
         self.buffer_used = 0
@@ -707,38 +706,38 @@ information passed to the ExternalEntityRefHandler."""
         colno = XML_GetCurrentColumnNumber(self.itself)
         msg = "%s: line %d, column %d" % (err, lineno, colno)
         w_errorcls = space.fromcache(Cache).w_error
-        w_error = space.call_function(w_errorcls, space.wrap(msg))
-        space.setattr(w_error, space.wrap("code"), space.wrap(code))
-        space.setattr(w_error, space.wrap("offset"), space.wrap(colno))
-        space.setattr(w_error, space.wrap("lineno"), space.wrap(lineno))
+        w_error = space.call_function(w_errorcls, space.newtext(msg))
+        space.setattr(w_error, space.newtext("code"), space.newint(code))
+        space.setattr(w_error, space.newtext("offset"), space.newint(colno))
+        space.setattr(w_error, space.newtext("lineno"), space.newint(lineno))
 
         self.w_error = w_error
         return OperationError(w_errorcls, w_error)
 
     def descr_ErrorCode(self, space):
-        return space.wrap(XML_GetErrorCode(self.itself))
+        return space.newint(XML_GetErrorCode(self.itself))
 
     def descr_ErrorLineNumber(self, space):
-        return space.wrap(XML_GetErrorLineNumber(self.itself))
+        return space.newint(XML_GetErrorLineNumber(self.itself))
 
     def descr_ErrorColumnNumber(self, space):
-        return space.wrap(XML_GetErrorColumnNumber(self.itself))
+        return space.newint(XML_GetErrorColumnNumber(self.itself))
 
     def descr_ErrorByteIndex(self, space):
-        return space.wrap(XML_GetErrorByteIndex(self.itself))
+        return space.newint(XML_GetErrorByteIndex(self.itself))
 
     def get_buffer_size(self, space):
-        return space.wrap(self.buffer_size)
+        return space.newint(self.buffer_size)
     def set_buffer_size(self, space, w_value):
-        value = space.getindex_w(w_value, space.w_TypeError)
+        value = space.getindex_w(w_value, space.w_OverflowError)
         if value <= 0:
-            raise OperationError(space.w_ValueError, space.wrap(
-                "buffer_size must be greater than zero"))
+            raise oefmt(space.w_ValueError,
+                        "buffer_size must be greater than zero")
         self.flush_character_buffer(space)
         self.buffer_size = value
 
     def get_buffer_text(self, space):
-        return space.wrap(self.buffer is not None)
+        return space.newbool(self.buffer is not None)
     def set_buffer_text(self, space, w_value):
         if space.is_true(w_value):
             self.buffer = []
@@ -756,7 +755,7 @@ information passed to the ExternalEntityRefHandler."""
 
 def bool_property(name, cls, doc=None):
     def fget(space, obj):
-        return space.wrap(getattr(obj, name))
+        return space.newbool(getattr(obj, name))
     def fset(space, obj, value):
         setattr(obj, name, space.bool_w(value))
     return GetSetProperty(fget, fset, cls=cls, doc=doc)
@@ -778,7 +777,6 @@ W_XMLParserType.typedef = TypeDef(
     namespace_prefixes = GetSetProperty(W_XMLParserType.get_namespace_prefixes,
                                         W_XMLParserType.set_namespace_prefixes,
                                         cls=W_XMLParserType),
-    returns_unicode = bool_property('returns_unicode', W_XMLParserType),
     ordered_attributes = bool_property('ordered_attributes', W_XMLParserType),
     specified_attributes = bool_property('specified_attributes', W_XMLParserType),
     intern = GetSetProperty(W_XMLParserType.get_intern, cls=W_XMLParserType),
@@ -805,29 +803,28 @@ def ParserCreate(space, w_encoding=None, w_namespace_separator=None,
 Return a new XML parser object."""
     if space.is_none(w_encoding):
         encoding = None
-    elif space.isinstance_w(w_encoding, space.w_str):
-        encoding = space.str_w(w_encoding)
+    elif space.isinstance_w(w_encoding, space.w_text):
+        encoding = space.text_w(w_encoding)
     else:
         raise oefmt(space.w_TypeError,
-                    "ParserCreate() argument 1 must be string or None, not %T",
+                    "ParserCreate() argument 1 must be str or None, not %T",
                     w_encoding)
 
     if space.is_none(w_namespace_separator):
-        namespace_separator = 0
-    elif space.isinstance_w(w_namespace_separator, space.w_str):
-        separator = space.str_w(w_namespace_separator)
+        namespace_separator = -1
+    elif space.isinstance_w(w_namespace_separator, space.w_text):
+        separator = space.text_w(w_namespace_separator)
         if len(separator) == 0:
             namespace_separator = 0
         elif len(separator) == 1:
             namespace_separator = ord(separator[0])
         else:
-            raise OperationError(
-                space.w_ValueError,
-                space.wrap('namespace_separator must be at most one character,'
-                           ' omitted, or None'))
+            raise oefmt(space.w_ValueError,
+                        "namespace_separator must be at most one character, "
+                        "omitted, or None")
     else:
         raise oefmt(space.w_TypeError,
-                    "ParserCreate() argument 2 must be string or None, not %T",
+                    "ParserCreate() argument 2 must be str or None, not %T",
                     w_namespace_separator)
 
     # Explicitly passing None means no interning is desired.
@@ -837,7 +834,7 @@ Return a new XML parser object."""
     elif space.is_w(w_intern, space.w_None):
         w_intern = None
 
-    if namespace_separator:
+    if namespace_separator >= 0:
         xmlparser = XML_ParserCreateNS(
             encoding,
             rffi.cast(rffi.CHAR, namespace_separator))
@@ -846,20 +843,19 @@ Return a new XML parser object."""
     # Currently this is just the size of the pointer and some estimated bytes.
     # The struct isn't actually defined in expat.h - it is in xmlparse.c
     # XXX: find a good estimate of the XML_ParserStruct
-    rgc.add_memory_pressure(XML_Parser_SIZE + 300)
     if not xmlparser:
-        raise OperationError(space.w_RuntimeError,
-                             space.wrap('XML_ParserCreate failed'))
+        raise oefmt(space.w_RuntimeError, "XML_ParserCreate failed")
 
     parser = W_XMLParserType(space, xmlparser, w_intern)
+    rgc.add_memory_pressure(XML_Parser_SIZE + 300, parser)
     XML_SetUnknownEncodingHandler(
         parser.itself, UnknownEncodingHandlerData_callback,
         rffi.cast(rffi.VOIDP, parser.id))
-    return space.wrap(parser)
+    return parser
 
 @unwrap_spec(code=int)
 def ErrorString(space, code):
     """ErrorString(errno) -> string
 Returns string error for given number."""
-    return space.wrap(rffi.charp2str(XML_ErrorString(code)))
+    return space.newtext(rffi.charp2str(XML_ErrorString(code)))
 
