@@ -11,37 +11,24 @@ from rpython.tool.jitlogparser.parser import (SimpleParser, Function,
 from rpython.tool.jitlogparser.storage import LoopStorage
 
 
-def find_ids_range(code):
+def find_id_linenos(code):
     """
     Parse the given function and return a dictionary mapping "ids" to
-    "line ranges".  Ids are identified by comments with a special syntax::
+    "line number".  Ids are identified by comments with a special syntax::
 
         # "myid" corresponds to the whole line
         print 'foo' # ID: myid
     """
     result = {}
-    start_lineno = code.co.co_firstlineno
+    start_lineno = code.startlineno
     for i, line in enumerate(py.code.Source(code.source)):
         m = re.search('# ID: (\w+)', line)
         if m:
             name = m.group(1)
             lineno = start_lineno+i
-            result[name] = xrange(lineno, lineno+1)
+            result.setdefault(name, lineno)
     return result
 
-def find_ids(code):
-    """
-    Parse the given function and return a dictionary mapping "ids" to
-    "opcodes".
-    """
-    ids = {}
-    ranges = find_ids_range(code)
-    for name, linerange in ranges.iteritems():
-        opcodes = [opcode for opcode in code.opcodes
-                   if opcode.lineno in linerange]
-        ids[name] = opcodes
-    return ids
-
 
 class Log(object):
     def __init__(self, rawtraces):
@@ -107,13 +94,18 @@ class TraceWithIds(Function):
     def compute_ids(self, ids):
         #
         # 1. compute the ids of self, i.e. the outer function
-        id2opcodes = find_ids(self.code)
-        all_my_opcodes = self.get_set_of_opcodes()
-        for id, opcodes in id2opcodes.iteritems():
-            if not opcodes:
-                continue
-            target_opcodes = set(opcodes)
-            if all_my_opcodes.intersection(target_opcodes):
+        id2lineno = find_id_linenos(self.code)
+        all_my_tracecodes = self.get_list_of_tracecodes()
+        for id, lineno in id2lineno.iteritems():
+            seen = set()
+            opcodes = []
+            for tracecode in all_my_tracecodes:
+                if tracecode.lineno == lineno:
+                    opcode = tracecode.getopcode()
+                    if opcode not in seen:
+                        seen.add(opcode)
+                        opcodes.append(opcode)
+            if opcodes:
                 ids[id] = opcodes
         #
         # 2. compute the ids of all the inlined functions
@@ -121,12 +113,11 @@ class TraceWithIds(Function):
             if isinstance(chunk, TraceWithIds) and chunk.code:
                 chunk.compute_ids(ids)
 
-    def get_set_of_opcodes(self):
-        result = set()
+    def get_list_of_tracecodes(self):
+        result = []
         for chunk in self.chunks:
             if isinstance(chunk, TraceForOpcode):
-                opcode = chunk.getopcode()
-                result.add(opcode)
+                result.append(chunk)
         return result
 
     def has_id(self, id):
@@ -134,7 +125,8 @@ class TraceWithIds(Function):
 
     def _ops_for_chunk(self, chunk, include_guard_not_invalidated):
         for op in chunk.operations:
-            if op.name != 'debug_merge_point' and \
+            if op.name not in ('debug_merge_point', 'enter_portal_frame',
+                               'leave_portal_frame') and \
                 (op.name != 'guard_not_invalidated' or include_guard_not_invalidated):
                 yield op
 
@@ -143,7 +135,7 @@ class TraceWithIds(Function):
         for chunk in self.flatten_chunks():
             opcode = chunk.getopcode()
             if opcode_name is None or \
-                   (opcode and opcode.__class__.__name__ == opcode_name):
+                   (opcode and opcode.match_name(opcode_name)):
                 for op in self._ops_for_chunk(chunk, include_guard_not_invalidated):
                     yield op
             else:
@@ -171,7 +163,7 @@ class TraceWithIds(Function):
         for chunk in self.flatten_chunks():
             opcode = chunk.getopcode()
             if opcode in target_opcodes and (opcode_name is None or
-                                             opcode.__class__.__name__ == opcode_name):
+                                             opcode.match_name(opcode_name)):
                 for op in self._ops_for_chunk(chunk, include_guard_not_invalidated):
                     if op in loop_ops:
                         yield op
@@ -261,6 +253,7 @@ class InvalidMatch(Exception):
 
 
 class OpMatcher(object):
+    DUMMY = 1000
 
     def __init__(self, ops, id=None):
         self.ops = ops
@@ -271,6 +264,7 @@ class OpMatcher(object):
     @classmethod
     def parse_ops(cls, src):
         ops = [cls.parse_op(line) for line in src.splitlines()]
+        ops.append(('--end--', None, [], '...', True))
         return [op for op in ops if op is not None]
 
     @classmethod
@@ -290,6 +284,13 @@ class OpMatcher(object):
         #
         if line.strip() == 'guard_not_invalidated?':
             return 'guard_not_invalidated', None, [], '...', False
+        if line.strip() == 'dummy_get_utf8?':
+            cls.DUMMY += 1
+            return ('getfield_gc_r',
+                    'p%d' % cls.DUMMY,
+                    ['ConstPtr(ptr%d)' % cls.DUMMY],
+                    '<FieldP pypy.objspace.std.unicodeobject.W_UnicodeObject.inst__utf8 .>',
+                    False)
         # find the resvar, if any
         if ' = ' in line:
             resvar, _, line = line.partition(' = ')
@@ -324,7 +325,7 @@ class OpMatcher(object):
         # to repeat it every time
         ticker_check = """
             guard_not_invalidated?
-            ticker0 = getfield_raw(#, descr=<FieldS pypysig_long_struct.c_value .*>)
+            ticker0 = getfield_raw_i(#, descr=<FieldS pypysig_long_struct.c_value .*>)
             ticker_cond0 = int_lt(ticker0, 0)
             guard_false(ticker_cond0, descr=...)
         """
@@ -333,7 +334,7 @@ class OpMatcher(object):
         # this is the ticker check generated if we have threads
         thread_ticker_check = """
             guard_not_invalidated?
-            ticker0 = getfield_raw(#, descr=<FieldS pypysig_long_struct.c_value .*>)
+            ticker0 = getfield_raw_i(#, descr=<FieldS pypysig_long_struct.c_value .*>)
             ticker1 = int_sub(ticker0, #)
             setfield_raw(#, ticker1, descr=<FieldS pypysig_long_struct.c_value .*>)
             ticker_cond0 = int_lt(ticker1, 0)
@@ -343,7 +344,7 @@ class OpMatcher(object):
         #
         # this is the ticker check generated in PyFrame.handle_operation_error
         exc_ticker_check = """
-            ticker2 = getfield_raw(#, descr=<FieldS pypysig_long_struct.c_value .*>)
+            ticker2 = getfield_raw_i(#, descr=<FieldS pypysig_long_struct.c_value .*>)
             ticker_cond1 = int_lt(ticker2, 0)
             guard_false(ticker_cond1, descr=...)
         """
@@ -403,6 +404,10 @@ class OpMatcher(object):
             raise InvalidMatch(message, frame=sys._getframe(1))
 
     def match_op(self, op, (exp_opname, exp_res, exp_args, exp_descr, _)):
+        if exp_opname == '--end--':
+            self._assert(op == '--end--', 'got more ops than expected')
+            return
+        self._assert(op != '--end--', 'got less ops than expected')
         self._assert(op.name == exp_opname, "operation mismatch")
         self.match_var(op.res, exp_res)
         if exp_args[-1:] == ['...']:      # exp_args ends with '...'
@@ -415,18 +420,15 @@ class OpMatcher(object):
         self.match_descr(op.descr, exp_descr)
 
 
-    def _next_op(self, iter_ops, assert_raises=False, ignore_ops=set()):
+    def _next_op(self, iter_ops, ignore_ops=set()):
         try:
             while True:
                 op = iter_ops.next()
                 if op.name not in ignore_ops:
                     break
         except StopIteration:
-            self._assert(assert_raises, "not enough operations")
-            return
-        else:
-            self._assert(not assert_raises, "operation list too long")
-            return op
+            return '--end--'
+        return op
 
     def try_match(self, op, exp_op):
         try:
@@ -447,6 +449,9 @@ class OpMatcher(object):
             if self.try_match(op, until_op):
                 # it matched! The '...' operator ends here
                 return op
+            self._assert(op != '--end--',
+                         'nothing in the end of the loop matches %r' %
+                          (until_op,))
 
     def match_any_order(self, iter_exp_ops, iter_ops, ignore_ops):
         exp_ops = []
@@ -493,16 +498,17 @@ class OpMatcher(object):
                     continue
                 else:
                     op = self._next_op(iter_ops, ignore_ops=ignore_ops)
-                self.match_op(op, exp_op)
-            except InvalidMatch, e:
-                if type(exp_op) is not str and exp_op[4] is False:    # optional operation
+                try:
+                    self.match_op(op, exp_op)
+                except InvalidMatch:
+                    if type(exp_op) is str or exp_op[4] is not False:
+                        raise
+                    #else: optional operation
                     iter_ops.revert_one()
                     continue       # try to match with the next exp_op
+            except InvalidMatch as e:
                 e.opindex = iter_ops.index - 1
                 raise
-        #
-        # make sure we exhausted iter_ops
-        self._next_op(iter_ops, assert_raises=True, ignore_ops=ignore_ops)
 
     def match(self, expected_src, ignore_ops=[]):
         def format(src, opindex=None):
@@ -518,7 +524,7 @@ class OpMatcher(object):
         expected_ops = self.parse_ops(expected_src)
         try:
             self.match_loop(expected_ops, ignore_ops)
-        except InvalidMatch, e:
+        except InvalidMatch as e:
             print '@' * 40
             print "Loops don't match"
             print "================="
@@ -545,9 +551,9 @@ class RevertableIterator(object):
         return self
     def next(self):
         index = self.index
-        if index == len(self.sequence):
-            raise StopIteration
         self.index = index + 1
+        if index >= len(self.sequence):
+            raise StopIteration
         return self.sequence[index]
     def revert_one(self):
         self.index -= 1
