@@ -105,8 +105,12 @@ class TestInstance(BaseTestPyPyC):
         # -------------------------------
         entry_bridge, = log.loops_by_filename(self.filepath, is_entry_bridge=True)
         ops = entry_bridge.ops_by_id('mutate', opcode='LOAD_ATTR')
-        assert log.opnames(ops) == ['guard_value', 'guard_not_invalidated',
-                                    'getfield_gc']
+        # in PyPy3 we get a dummy getfield_gc_r (*) for
+        # W_UnicodeObject._utf8, which is usually removed by the backend
+        assert log.opnames(ops) == ['guard_value',
+                                    'getfield_gc_r',  # <= (*)
+                                    'guard_not_invalidated',
+                                    'getfield_gc_i']
         # the STORE_ATTR is folded away
         assert list(entry_bridge.ops_by_id('meth1', opcode='STORE_ATTR')) == []
         #
@@ -120,11 +124,11 @@ class TestInstance(BaseTestPyPyC):
             i59 = int_add_ovf(i57, 1)
             guard_no_overflow(descr=...)
             p60 = force_token()
-            i61 = getfield_raw(..., descr=...)
+            i61 = getfield_raw_i(..., descr=...)
             setfield_gc(ConstPtr(ptr39), i59, descr=...)
             i62 = int_lt(i61, 0)
             guard_false(i62, descr=...)
-            jump(p0, p1, p3, p6, p7, p12, i59, p18, i31, i59, descr=...)
+            jump(..., descr=...)
         """)
 
     def test_mutate_class(self):
@@ -153,9 +157,13 @@ class TestInstance(BaseTestPyPyC):
         # -------------------------------
         entry_bridge, = log.loops_by_filename(self.filepath, is_entry_bridge=True)
         ops = entry_bridge.ops_by_id('mutate', opcode='LOAD_ATTR')
-        assert log.opnames(ops) == ['guard_value', 'guard_not_invalidated',
-                                    'getfield_gc', 'guard_nonnull_class',
-                                    'getfield_gc', 'guard_value', # type check on the attribute
+        # in PyPy3 we get a dummy getfield_gc_r (*) for
+        # W_UnicodeObject._utf8, which is usually removed by the backend
+        assert log.opnames(ops) == ['guard_value',
+                                    'getfield_gc_r',   # <= (*)
+                                    'guard_not_invalidated',
+                                    'getfield_gc_r', 'guard_nonnull_class',
+                                    'getfield_gc_r', 'guard_value', # type check on the attribute
                                     ]
         # the STORE_ATTR is folded away
         assert list(entry_bridge.ops_by_id('meth1', opcode='STORE_ATTR')) == []
@@ -167,15 +175,15 @@ class TestInstance(BaseTestPyPyC):
             i70 = int_lt(i58, i33)
             guard_true(i70, descr=...)
             guard_not_invalidated(descr=...)
-            p71 = getfield_gc(p64, descr=...)
+            p71 = getfield_gc_r(p64, descr=...)
             guard_value(p71, ConstPtr(ptr42), descr=...)
             p72 = force_token()
             p73 = force_token()
             i74 = int_add(i58, 1)
-            i75 = getfield_raw(..., descr=...)
+            i75 = getfield_raw_i(..., descr=...)
             i76 = int_lt(i75, 0)
             guard_false(i76, descr=...)
-            p77 = new_with_vtable(...)
+            p77 = new_with_vtable(descr=...)
             setfield_gc(p77, p64, descr=...)
             setfield_gc(p77, ConstPtr(null), descr=...)
             setfield_gc(p77, ConstPtr(null), descr=...)
@@ -183,43 +191,10 @@ class TestInstance(BaseTestPyPyC):
             setfield_gc(p77, ConstPtr(null), descr=...)
             setfield_gc(p77, ConstPtr(ptr42), descr=...)
             setfield_gc(ConstPtr(ptr69), p77, descr=...)
-            jump(p0, p1, p3, p6, p7, p12, i74, p20, p26, i33, p77, descr=...)
+            jump(..., descr=...)
 
         """)
 
-    def test_oldstyle_newstyle_mix(self):
-        def main():
-            class A:
-                pass
-
-            class B(object, A):
-                def __init__(self, x):
-                    self.x = x
-
-            i = 0
-            b = B(1)
-            while i < 100:
-                v = b.x # ID: loadattr1
-                v = b.x # ID: loadattr2
-                i += v
-            return i
-
-        log = self.run(main, [], threshold=80)
-        loop, = log.loops_by_filename(self.filepath)
-        assert loop.match_by_id('loadattr1',
-        '''
-        guard_not_invalidated(descr=...)
-        i19 = call(ConstClass(ll_call_lookup_function), _, _, _, 0, descr=...)
-        guard_no_exception(descr=...)
-        i22 = int_lt(i19, 0)
-        guard_true(i22, descr=...)
-        i26 = call(ConstClass(ll_call_lookup_function), _, _, _, 0, descr=...)
-        guard_no_exception(descr=...)
-        i29 = int_lt(i26, 0)
-        guard_true(i29, descr=...)
-        ''')
-        assert loop.match_by_id('loadattr2', "")   # completely folded away
-
     def test_python_contains(self):
         def main():
             class A(object):
@@ -283,3 +258,36 @@ class TestInstance(BaseTestPyPyC):
             --TICK--
             jump(..., descr=...)
         """)
+
+    def test_super_no_args(self):
+        def main():
+            class A(object):
+                def m(self, x):
+                    return x + 1
+            class B(A):
+                def m(self, x):
+                    return super().m(x)
+            i = 0
+            while i < 300:
+                i = B().m(i)
+            return i
+
+        log = self.run(main, [])
+        loop, = log.loops_by_filename(self.filepath)
+        assert loop.match("""
+            i78 = int_lt(i72, 300)
+            guard_true(i78, descr=...)
+            guard_not_invalidated(descr=...)
+            p1 = force_token()
+            p65 = force_token()
+            p3 = force_token()
+            i81 = int_add(i72, 1)
+
+            # can't use TICK here, because of the extra setfield_gc
+            ticker0 = getfield_raw_i(#, descr=<FieldS pypysig_long_struct.c_value .*>)
+            setfield_gc(p0, p65, descr=<FieldP pypy.interpreter.pyframe.PyFrame.vable_token .>)
+            ticker_cond0 = int_lt(ticker0, 0)
+            guard_false(ticker_cond0, descr=...)
+
+            jump(..., descr=...)
+        """)
