@@ -3,23 +3,12 @@ from pypy.interpreter.gateway import interp2app, unwrap_spec
 from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.typedef import TypeDef, interp_attrproperty
 from pypy.interpreter.error import OperationError, oefmt
-from rpython.rlib.rarithmetic import intmask, r_uint
+from rpython.rlib.rarithmetic import intmask, r_uint, r_uint32
 from rpython.rlib.objectmodel import keepalive_until_here
 
 from rpython.rlib import rzlib
 
 
-if intmask(2**31) == -2**31:
-    # 32-bit platforms
-    unsigned_to_signed_32bit = intmask
-else:
-    # 64-bit platforms
-    def unsigned_to_signed_32bit(x):
-        # assumes that 'x' is in range(0, 2**32) to start with
-        SIGN_EXTEND2 = 1 << 31
-        return intmask((x ^ SIGN_EXTEND2) - SIGN_EXTEND2)
-
-
 @unwrap_spec(string='bufferstr', start='truncatedint_w')
 def crc32(space, string, start = rzlib.CRC32_DEFAULT_START):
     """
@@ -28,17 +17,9 @@ def crc32(space, string, start = rzlib.CRC32_DEFAULT_START):
     An optional starting value can be specified.  The returned checksum is
     an integer.
     """
-    ustart = r_uint(start)
+    ustart = r_uint(r_uint32(start))
     checksum = rzlib.crc32(string, ustart)
-
-    # This is, perhaps, a little stupid.  zlib returns the checksum unsigned.
-    # CPython exposes it as a signed value, though. -exarkun
-    # Note that in CPython < 2.6 on 64-bit platforms the result is
-    # actually unsigned, but it was considered to be a bug so we stick to
-    # the 2.6 behavior and always return a number in range(-2**31, 2**31).
-    checksum = unsigned_to_signed_32bit(checksum)
-
-    return space.wrap(checksum)
+    return space.newint(checksum)
 
 
 @unwrap_spec(string='bufferstr', start='truncatedint_w')
@@ -49,12 +30,9 @@ def adler32(space, string, start=rzlib.ADLER32_DEFAULT_START):
     An optional starting value can be specified.  The returned checksum is
     an integer.
     """
-    ustart = r_uint(start)
+    ustart = r_uint(r_uint32(start))
     checksum = rzlib.adler32(string, ustart)
-    # See comments in crc32() for the following line
-    checksum = unsigned_to_signed_32bit(checksum)
-
-    return space.wrap(checksum)
+    return space.newint(checksum)
 
 
 class Cache:
@@ -63,13 +41,13 @@ class Cache:
 
 def zlib_error(space, msg):
     w_error = space.fromcache(Cache).w_error
-    return OperationError(w_error, space.wrap(msg))
+    return OperationError(w_error, space.newtext(msg))
 
 
-@unwrap_spec(string='bufferstr', level=int)
-def compress(space, string, level=rzlib.Z_DEFAULT_COMPRESSION):
+@unwrap_spec(data='bufferstr', level=int)
+def compress(space, data, level=rzlib.Z_DEFAULT_COMPRESSION):
     """
-    compress(string[, level]) -- Returned compressed string.
+    compress(data[, level]) -- Returned compressed string.
 
     Optional arg level is the compression level, in 1-9.
     """
@@ -79,12 +57,12 @@ def compress(space, string, level=rzlib.Z_DEFAULT_COMPRESSION):
         except ValueError:
             raise zlib_error(space, "Bad compression level")
         try:
-            result = rzlib.compress(stream, string, rzlib.Z_FINISH)
+            result = rzlib.compress(stream, data, rzlib.Z_FINISH)
         finally:
             rzlib.deflateEnd(stream)
-    except rzlib.RZlibError, e:
+    except rzlib.RZlibError as e:
         raise zlib_error(space, e.msg)
-    return space.wrap(result)
+    return space.newbytes(result)
 
 
 @unwrap_spec(string='bufferstr', wbits="c_int", bufsize=int)
@@ -104,9 +82,9 @@ def decompress(space, string, wbits=rzlib.MAX_WBITS, bufsize=0):
             result, _, _ = rzlib.decompress(stream, string, rzlib.Z_FINISH)
         finally:
             rzlib.inflateEnd(stream)
-    except rzlib.RZlibError, e:
+    except rzlib.RZlibError as e:
         raise zlib_error(space, e.msg)
-    return space.wrap(result)
+    return space.newbytes(result)
 
 
 class ZLibObject(W_Root):
@@ -139,18 +117,19 @@ class Compress(ZLibObject):
                  method=rzlib.Z_DEFLATED,             # \
                  wbits=rzlib.MAX_WBITS,               #  \   undocumented
                  memLevel=rzlib.DEF_MEM_LEVEL,        #  /    parameters
-                 strategy=rzlib.Z_DEFAULT_STRATEGY):  # /
+                 strategy=rzlib.Z_DEFAULT_STRATEGY,   # /
+                 zdict=None):
         ZLibObject.__init__(self, space)
         try:
             self.stream = rzlib.deflateInit(level, method, wbits,
-                                            memLevel, strategy)
-        except rzlib.RZlibError, e:
+                                            memLevel, strategy, zdict=zdict)
+        except rzlib.RZlibError as e:
             raise zlib_error(space, e.msg)
         except ValueError:
-            raise OperationError(space.w_ValueError,
-                                 space.wrap("Invalid initialization option"))
+            raise oefmt(space.w_ValueError, "Invalid initialization option")
+        self.register_finalizer(space)
 
-    def __del__(self):
+    def _finalize_(self):
         """Automatically free the resources used by the stream."""
         if self.stream:
             rzlib.deflateEnd(self.stream)
@@ -175,9 +154,9 @@ class Compress(ZLibObject):
                 result = rzlib.compress(self.stream, data)
             finally:
                 self.unlock()
-        except rzlib.RZlibError, e:
+        except rzlib.RZlibError as e:
             raise zlib_error(space, e.msg)
-        return space.wrap(result)
+        return space.newbytes(result)
 
     @unwrap_spec(mode="c_int")
     def flush(self, space, mode=rzlib.Z_FINISH):
@@ -202,11 +181,12 @@ class Compress(ZLibObject):
                 if mode == rzlib.Z_FINISH:    # release the data structures now
                     rzlib.deflateEnd(self.stream)
                     self.stream = rzlib.null_stream
+                    self.may_unregister_rpython_finalizer(space)
             finally:
                 self.unlock()
-        except rzlib.RZlibError, e:
+        except rzlib.RZlibError as e:
             raise zlib_error(space, e.msg)
-        return space.wrap(result)
+        return space.newbytes(result)
 
 
 @unwrap_spec(level=int, method=int, wbits=int, memLevel=int, strategy=int)
@@ -214,15 +194,20 @@ def Compress___new__(space, w_subtype, level=rzlib.Z_DEFAULT_COMPRESSION,
                      method=rzlib.Z_DEFLATED,             # \
                      wbits=rzlib.MAX_WBITS,               #  \   undocumented
                      memLevel=rzlib.DEF_MEM_LEVEL,        #  /    parameters
-                     strategy=rzlib.Z_DEFAULT_STRATEGY):  # /
+                     strategy=rzlib.Z_DEFAULT_STRATEGY,   # /
+                     w_zdict=None):
     """
     Create a new z_stream and call its initializer.
     """
+    if space.is_none(w_zdict):
+        zdict = None
+    else:
+        zdict = space.charbuf_w(w_zdict)
     stream = space.allocate_instance(Compress, w_subtype)
     stream = space.interp_w(Compress, stream)
     Compress.__init__(stream, space, level,
-                      method, wbits, memLevel, strategy)
-    return space.wrap(stream)
+                      method, wbits, memLevel, strategy, zdict)
+    return stream
 
 
 Compress.typedef = TypeDef(
@@ -241,7 +226,7 @@ class Decompress(ZLibObject):
     Wrapper around zlib's z_stream structure which provides convenient
     decompression functionality.
     """
-    def __init__(self, space, wbits=rzlib.MAX_WBITS):
+    def __init__(self, space, wbits=rzlib.MAX_WBITS, zdict=None):
         """
         Initialize a new decompression object.
 
@@ -253,15 +238,17 @@ class Decompress(ZLibObject):
         ZLibObject.__init__(self, space)
         self.unused_data = ''
         self.unconsumed_tail = ''
+        self.eof = False
         try:
-            self.stream = rzlib.inflateInit(wbits)
-        except rzlib.RZlibError, e:
+            self.stream = rzlib.inflateInit(wbits, zdict=zdict)
+        except rzlib.RZlibError as e:
             raise zlib_error(space, e.msg)
         except ValueError:
-            raise OperationError(space.w_ValueError,
-                                 space.wrap("Invalid initialization option"))
+            raise oefmt(space.w_ValueError, "Invalid initialization option")
+        self.zdict = zdict
+        self.register_finalizer(space)
 
-    def __del__(self):
+    def _finalize_(self):
         """Automatically free the resources used by the stream."""
         if self.stream:
             rzlib.inflateEnd(self.stream)
@@ -277,7 +264,7 @@ class Decompress(ZLibObject):
         else:
             self.unconsumed_tail = tail
 
-    @unwrap_spec(data='bufferstr', max_length="c_int")
+    @unwrap_spec(data='bufferstr', max_length=int)
     def decompress(self, space, data, max_length=0):
         """
         decompress(data[, max_length]) -- Return a string containing the
@@ -295,15 +282,18 @@ class Decompress(ZLibObject):
         try:
             self.lock()
             try:
-                result = rzlib.decompress(self.stream, data, max_length=max_length)
+                result = rzlib.decompress(self.stream, data,
+                                          max_length=max_length,
+                                          zdict=self.zdict)
             finally:
                 self.unlock()
-        except rzlib.RZlibError, e:
+        except rzlib.RZlibError as e:
             raise zlib_error(space, e.msg)
 
         string, finished, unused_len = result
+        self.eof = finished
         self._save_unconsumed_input(data, finished, unused_len)
-        return space.wrap(string)
+        return space.newbytes(string)
 
     def flush(self, space, w_length=None):
         """
@@ -312,7 +302,7 @@ class Decompress(ZLibObject):
         data as possible.
         """
         if w_length is not None:
-            length = space.c_int_w(w_length)
+            length = space.int_w(w_length)
             if length <= 0:
                 raise oefmt(space.w_ValueError,
                             "length must be greater than zero")
@@ -320,7 +310,8 @@ class Decompress(ZLibObject):
         try:
             self.lock()
             try:
-                result = rzlib.decompress(self.stream, data, rzlib.Z_FINISH)
+                result = rzlib.decompress(self.stream, data, rzlib.Z_FINISH,
+                                          zdict=self.zdict)
             finally:
                 self.unlock()
         except rzlib.RZlibError:
@@ -328,27 +319,34 @@ class Decompress(ZLibObject):
         else:
             string, finished, unused_len = result
             self._save_unconsumed_input(data, finished, unused_len)
-        return space.wrap(string)
+        return space.newbytes(string)
 
 
 @unwrap_spec(wbits=int)
-def Decompress___new__(space, w_subtype, wbits=rzlib.MAX_WBITS):
+def Decompress___new__(space, w_subtype, wbits=rzlib.MAX_WBITS, w_zdict=None):
     """
     Create a new Decompress and call its initializer.
     """
+    if space.is_none(w_zdict):
+        zdict = None
+    else:
+        zdict = space.charbuf_w(w_zdict)
     stream = space.allocate_instance(Decompress, w_subtype)
     stream = space.interp_w(Decompress, stream)
-    Decompress.__init__(stream, space, wbits)
-    return space.wrap(stream)
+    Decompress.__init__(stream, space, wbits, zdict)
+    return stream
 
+def default_buffer_size(space):
+    return space.newint(rzlib.OUTPUT_BUFFER_SIZE)
 
 Decompress.typedef = TypeDef(
     'Decompress',
     __new__ = interp2app(Decompress___new__),
     decompress = interp2app(Decompress.decompress),
     flush = interp2app(Decompress.flush),
-    unused_data = interp_attrproperty('unused_data', Decompress),
-    unconsumed_tail = interp_attrproperty('unconsumed_tail', Decompress),
+    unused_data = interp_attrproperty('unused_data', Decompress, wrapfn="newbytes"),
+    unconsumed_tail = interp_attrproperty('unconsumed_tail', Decompress, wrapfn="newbytes"),
+    eof = interp_attrproperty('eof', Decompress, wrapfn="newbool"),
     __doc__ = """decompressobj([wbits]) -- Return a decompressor object.
 
 Optional arg wbits is the window buffer size.
