@@ -1,18 +1,20 @@
 from rpython.annotator.model import SomeInstance, s_None
-from pypy.interpreter import argument, gateway
-from pypy.interpreter.baseobjspace import W_Root, ObjSpace, SpaceCache
-from pypy.interpreter.typedef import TypeDef, GetSetProperty
-from pypy.objspace.std.sliceobject import W_SliceObject
-from rpython.rlib.buffer import StringBuffer
-from rpython.rlib.objectmodel import instantiate, we_are_translated, specialize
+from rpython.rlib.objectmodel import (instantiate, we_are_translated, specialize,
+    not_rpython)
 from rpython.rlib.nonconst import NonConstant
 from rpython.rlib.rarithmetic import r_uint, r_singlefloat
 from rpython.rtyper.extregistry import ExtRegistryEntry
 from rpython.rtyper.lltypesystem import lltype
-from pypy.tool.option import make_config
 from rpython.tool.sourcetools import compile2, func_with_new_name
 from rpython.translator.translator import TranslationContext
 
+from pypy.tool.option import make_config
+from pypy.interpreter import argument, gateway
+from pypy.interpreter.baseobjspace import W_Root, ObjSpace, SpaceCache
+from pypy.interpreter.buffer import StringBuffer, SimpleView
+from pypy.interpreter.typedef import TypeDef, GetSetProperty
+from pypy.objspace.std.sliceobject import W_SliceObject
+
 
 class W_MyObject(W_Root):
     typedef = None
@@ -40,10 +42,11 @@ class W_MyObject(W_Root):
         is_root(w_subtype)
 
     def buffer_w(self, space, flags):
-        return StringBuffer("foobar")
+        return SimpleView(StringBuffer("foobar"))
 
-    def str_w(self, space):
+    def text_w(self, space):
         return NonConstant("foobar")
+    bytes_w = text_w
 
     def unicode_w(self, space):
         return NonConstant(u"foobar")
@@ -56,7 +59,10 @@ class W_MyObject(W_Root):
 
     def bigint_w(self, space, allow_conversion=True):
         from rpython.rlib.rbigint import rbigint
-        return rbigint.fromint(NonConstant(42))
+        x = 42
+        if we_are_translated():
+            x = NonConstant(x)
+        return rbigint.fromint(x)
 
 class W_MyListObj(W_MyObject):
     def append(self, w_other):
@@ -64,14 +70,20 @@ class W_MyListObj(W_MyObject):
 
 class W_MyType(W_MyObject):
     name = "foobar"
+    flag_map_or_seq = '?'
+    hasuserdel = False
 
     def __init__(self):
         self.mro_w = [w_some_obj(), w_some_obj()]
         self.dict_w = {'__str__': w_some_obj()}
+        self.hasuserdel = True
 
     def get_module(self):
         return w_some_obj()
 
+    def getname(self, space):
+        return self.name.decode('utf-8')
+
 def w_some_obj():
     if NonConstant(False):
         return W_Root()
@@ -111,16 +123,27 @@ class Entry(ExtRegistryEntry):
 # ____________________________________________________________
 
 
-BUILTIN_TYPES = ['int', 'str', 'float', 'long', 'tuple', 'list', 'dict',
-                 'unicode', 'complex', 'slice', 'bool', 'basestring', 'object',
-                 'bytearray', 'buffer']
+BUILTIN_TYPES = ['int', 'float', 'tuple', 'list', 'dict', 'bytes',
+                 'unicode', 'complex', 'slice', 'bool', 'text', 'object',
+                 'set', 'frozenset', 'bytearray', 'memoryview']
+
+INTERP_TYPES = ['function', 'builtin_function', 'module', 'getset_descriptor',
+                'instance', 'classobj']
 
 class FakeObjSpace(ObjSpace):
+    is_fake_objspace = True
+
     def __init__(self, config=None):
         self._seen_extras = []
         ObjSpace.__init__(self, config=config)
         self.setup()
 
+        # Be sure to annotate W_SliceObject constructor.
+        # In Python2, this is triggered by W_InstanceObject.__getslice__.
+        def build_slice():
+            self.newslice(self.w_None, self.w_None, self.w_None)
+        self._seen_extras.append(build_slice)
+
     def _freeze_(self):
         return True
 
@@ -132,8 +155,8 @@ class FakeObjSpace(ObjSpace):
         is_root(w_obj)
         return NonConstant(False)
 
+    @not_rpython
     def unwrap(self, w_obj):
-        "NOT_RPYTHON"
         raise NotImplementedError
 
     def newdict(self, module=False, instance=False, kwargs=False,
@@ -145,6 +168,12 @@ class FakeObjSpace(ObjSpace):
             is_root(w_x)
         return w_some_obj()
 
+    def newset(self, list_w=None):
+        if list_w is not None:
+            for w_x in list_w:
+                is_root(w_x)
+        return w_some_obj()
+
     def newlist(self, list_w):
         for w_x in list_w:
             is_root(w_x)
@@ -157,6 +186,7 @@ class FakeObjSpace(ObjSpace):
         W_SliceObject(w_start, w_end, w_step)
         return w_some_obj()
 
+    @specialize.argtype(1)
     def newint(self, x):
         return w_some_obj()
 
@@ -175,34 +205,40 @@ class FakeObjSpace(ObjSpace):
     def newseqiter(self, x):
         return w_some_obj()
 
-    def newbuffer(self, x):
+    def newmemoryview(self, x):
         return w_some_obj()
 
+    @not_rpython
     def marshal_w(self, w_obj):
-        "NOT_RPYTHON"
         raise NotImplementedError
 
+    def newbytes(self, x):
+        return w_some_obj()
+
+    def newunicode(self, x):
+        return w_some_obj()
+
+    def newtext(self, x):
+        return w_some_obj()
+    newtext_or_none = newtext
+    newfilename = newtext
+
+    @not_rpython
     def wrap(self, x):
         if not we_are_translated():
-            if isinstance(x, gateway.interp2app):
-                self._see_interp2app(x)
-            if isinstance(x, GetSetProperty):
-                self._see_getsetproperty(x)
+            if isinstance(x, W_Root):
+                x.spacebind(self)
         if isinstance(x, r_singlefloat):
             self._wrap_not_rpython(x)
         if isinstance(x, list):
             if x == []: # special case: it is used e.g. in sys/__init__.py
                 return w_some_obj()
-            self._wrap_not_rpython(x)
+            raise NotImplementedError
         return w_some_obj()
-    wrap._annspecialcase_ = "specialize:argtype(1)"
-
-    def _wrap_not_rpython(self, x):
-        "NOT_RPYTHON"
-        raise NotImplementedError
 
+    @not_rpython
     def _see_interp2app(self, interp2app):
-        "NOT_RPYTHON"
+        """Called by GatewayCache.build()"""
         activation = interp2app._code.activation
         def check():
             scope_w = [w_some_obj()] * NonConstant(42)
@@ -211,8 +247,9 @@ class FakeObjSpace(ObjSpace):
         check = func_with_new_name(check, 'check__' + interp2app.name)
         self._seen_extras.append(check)
 
+    @not_rpython
     def _see_getsetproperty(self, getsetproperty):
-        "NOT_RPYTHON"
+        """Called by GetSetProperty.spacebind()"""
         space = self
         def checkprop():
             getsetproperty.fget(getsetproperty, space, w_some_obj())
@@ -265,6 +302,14 @@ class FakeObjSpace(ObjSpace):
     def type(self, w_obj):
         return w_some_type()
 
+    def lookup_in_type_where(self, w_type, key):
+        return w_some_obj(), w_some_obj()
+
+    def issubtype_w(self, w_sub, w_type):
+        is_root(w_sub)
+        is_root(w_type)
+        return NonConstant(True)
+
     def isinstance_w(self, w_inst, w_type):
         is_root(w_inst)
         is_root(w_type)
@@ -280,10 +325,10 @@ class FakeObjSpace(ObjSpace):
         is_root(w_complex)
         return 1.1, 2.2
 
+    @specialize.arg(1)
     def allocate_instance(self, cls, w_subtype):
         is_root(w_subtype)
         return instantiate(cls)
-    allocate_instance._annspecialcase_ = "specialize:arg(1)"
 
     def decode_index(self, w_index_or_slice, seqlength):
         is_root(w_index_or_slice)
@@ -305,6 +350,27 @@ class FakeObjSpace(ObjSpace):
     def unicode_from_object(self, w_obj):
         return w_some_obj()
 
+    def encode_unicode_object(self, w_unicode, encoding, errors):
+        return w_some_obj()
+
+    def _try_fetch_pycode(self, w_func):
+        return None
+
+    def is_generator(self, w_obj):
+        return NonConstant(False)
+
+    def is_iterable(self, w_obj):
+        return NonConstant(False)
+
+    def lookup_in_type(self, w_type, name):
+        return w_some_obj()
+
+    def warn(self, w_msg, w_warningcls, stacklevel=2):
+        pass
+
+    def _try_buffer_w(self, w_obj, flags):
+        return w_obj.buffer_w(self, flags)
+
     # ----------
 
     def translates(self, func=None, argtypes=None, seeobj_w=[], **kwds):
@@ -319,6 +385,7 @@ class FakeObjSpace(ObjSpace):
         ann = t.buildannotator()
         def _do_startup():
             self.threadlocals.enter_thread(self)
+            W_SliceObject(w_some_obj(), w_some_obj(), w_some_obj())
         ann.build_types(_do_startup, [], complete_now=False)
         if func is not None:
             ann.build_types(func, argtypes, complete_now=False)
@@ -346,7 +413,10 @@ class FakeObjSpace(ObjSpace):
         for name in (ObjSpace.ConstantTable +
                      ObjSpace.ExceptionTable +
                      BUILTIN_TYPES):
-            setattr(space, 'w_' + name, w_some_obj())
+            if name != "str":
+                setattr(space, 'w_' + name, w_some_obj())
+        space.w_bytes = w_some_obj()
+        space.w_text = w_some_obj()
         space.w_type = w_some_type()
         #
         for (name, _, arity, _) in ObjSpace.MethodTable:
@@ -376,23 +446,32 @@ class FakeObjSpace(ObjSpace):
 @specialize.memo()
 def see_typedef(space, typedef):
     assert isinstance(typedef, TypeDef)
-    if typedef.name not in BUILTIN_TYPES:
+    if typedef.name not in BUILTIN_TYPES and typedef.name not in INTERP_TYPES:
+        print
+        print '------ seeing typedef %r ------' % (typedef.name,)
         for name, value in typedef.rawdict.items():
             space.wrap(value)
 
 class FakeCompiler(object):
-    pass
+    def compile(self, code, name, mode, flags):
+        return FakePyCode()
 FakeObjSpace.default_compiler = FakeCompiler()
 
+class FakePyCode(W_Root):
+    def exec_code(self, space, w_globals, w_locals):
+        return W_Root()
+
 
 class FakeModule(W_Root):
     def __init__(self):
         self.w_dict = w_some_obj()
-
     def get(self, name):
         name + "xx"   # check that it's a string
         return w_some_obj()
+    def setmodule(self, w_mod):
+        is_root(w_mod)
 FakeObjSpace.sys = FakeModule()
 FakeObjSpace.sys.filesystemencoding = 'foobar'
 FakeObjSpace.sys.defaultencoding = 'ascii'
+FakeObjSpace.sys.dlopenflags = 123
 FakeObjSpace.builtin = FakeModule()
