@@ -1,49 +1,67 @@
 """The builtin bytearray implementation"""
 
+import sys
 from rpython.rlib.objectmodel import (
     import_from_mixin, newlist_hint, resizelist_hint, specialize)
-from rpython.rlib.buffer import Buffer
+from rpython.rlib.rarithmetic import intmask
 from rpython.rlib.rstring import StringBuilder, ByteListBuilder
-from rpython.rlib.debug import check_list_of_chars
-
+from rpython.rlib.debug import check_list_of_chars, check_nonneg
+from rpython.rtyper.lltypesystem import rffi
+from rpython.rlib.rgc import (resizable_list_supporting_raw_ptr,
+                              nonmoving_raw_ptr_for_resizable_list)
+from rpython.rlib import jit
+from rpython.rlib.buffer import (GCBuffer,
+                                 get_gc_data_for_list_of_chars,
+                                 get_gc_data_offset_for_list_of_chars)
 from pypy.interpreter.baseobjspace import W_Root
 from pypy.interpreter.error import OperationError, oefmt
+from pypy.objspace.std.bytesobject import makebytesdata_w, newbytesdata_w
 from pypy.interpreter.gateway import WrappedDefault, interp2app, unwrap_spec
-from pypy.interpreter.signature import Signature
 from pypy.interpreter.typedef import TypeDef
-from pypy.objspace.std.sliceobject import W_SliceObject
-from pypy.objspace.std.stringmethods import StringMethods, _get_buffer
+from pypy.interpreter.buffer import SimpleView
+from pypy.objspace.std.sliceobject import W_SliceObject, unwrap_start_stop
+from pypy.objspace.std.stringmethods import StringMethods
+from pypy.objspace.std.stringmethods import _descr_getslice_slowpath
 from pypy.objspace.std.bytesobject import W_BytesObject
 from pypy.objspace.std.util import get_positive_index
-
-NON_HEX_MSG = "non-hexadecimal number found in fromhex() arg at position %d"
+from pypy.objspace.std.formatting import mod_format, FORMAT_BYTEARRAY
 
 
 class W_BytearrayObject(W_Root):
     import_from_mixin(StringMethods)
+    _KIND1 = "bytearray"
+    _KIND2 = "bytearray"
 
     def __init__(self, data):
         check_list_of_chars(data)
-        self.data = data
+        self._data = resizable_list_supporting_raw_ptr(data)
+        self._offset = 0
+        # NOTE: the bytearray data is in 'self._data[self._offset:]'
+        check_nonneg(self._offset)
+        _tweak_for_tests(self)
+
+    def getdata(self):
+        if self._offset > 0:
+            self._data = self._data[self._offset:]
+            self._offset = 0
+        return self._data
 
     def __repr__(self):
         """representation for debugging purposes"""
-        return "%s(%s)" % (self.__class__.__name__, ''.join(self.data))
+        return "%s(%s)" % (self.__class__.__name__,
+                           ''.join(self._data[self._offset:]))
 
     def buffer_w(self, space, flags):
-        return BytearrayBuffer(self.data, False)
+        return SimpleView(BytearrayBuffer(self))
 
-    def readbuf_w(self, space):
-        return BytearrayBuffer(self.data, True)
+    def bytearray_list_of_chars_w(self, space):
+        return self.getdata()
 
-    def writebuf_w(self, space):
-        return BytearrayBuffer(self.data, False)
-
-    def charbuf_w(self, space):
-        return ''.join(self.data)
+    def nonmovable_carray(self, space):
+        return BytearrayBuffer(self).get_raw_address()
 
     def _new(self, value):
-        if value is self.data:
+        if value is self._data:
             value = value[:]
         return W_BytearrayObject(value)
 
@@ -57,24 +75,43 @@ class W_BytearrayObject(W_Root):
         return W_BytearrayObject([])
 
     def _len(self):
-        return len(self.data)
+        return len(self._data) - self._offset
+
+    def _fixindex(self, space, index, errmsg="bytearray index out of range"):
+        # for getitem/setitem/delitem of a single char
+        if index >= 0:
+            index += self._offset
+            if index >= len(self._data):
+                raise OperationError(space.w_IndexError, space.newtext(errmsg))
+        else:
+            index += len(self._data)    # count from the end
+            if index < self._offset:
+                raise OperationError(space.w_IndexError, space.newtext(errmsg))
+        check_nonneg(index)
+        return index
 
     def _getitem_result(self, space, index):
-        try:
-            character = self.data[index]
-        except IndexError:
-            raise oefmt(space.w_IndexError, "bytearray index out of range")
-        return space.wrap(ord(character))
+        character = self._data[self._fixindex(space, index)]
+        return space.newint(ord(character))
 
     def _val(self, space):
-        return self.data
+        return self.getdata()
 
     @staticmethod
     def _use_rstr_ops(space, w_other):
         return False
 
     @staticmethod
-    def _op_val(space, w_other):
+    def _op_val(space, w_other, allow_char=False):
+        # Some functions (contains, find) allow a number to specify a
+        # single char.
+        if allow_char and space.isinstance_w(w_other, space.w_int):
+            return StringMethods._single_char(space, w_other)
+        try:
+            return space.bytes_w(w_other)
+        except OperationError as e:
+            if not e.match(space, space.w_TypeError):
+                raise
         return space.buffer_w(w_other, space.BUF_SIMPLE).as_str()
 
     def _chr(self, char):
@@ -136,18 +173,13 @@ class W_BytearrayObject(W_Root):
     def _join_return_one(self, space, w_obj):
         return False
 
-    def _join_check_item(self, space, w_obj):
-        if (space.isinstance_w(w_obj, space.w_str) or
-            space.isinstance_w(w_obj, space.w_bytearray)):
-            return 0
-        return 1
-
     def ord(self, space):
-        if len(self.data) != 1:
+        length = self._len()
+        if length != 1:
             raise oefmt(space.w_TypeError,
                         "ord() expected a character, but string of length %d "
-                        "found", len(self.data))
-        return space.wrap(ord(self.data[0]))
+                        "found", length)
+        return space.newint(ord(self._data[self._offset]))
 
     @staticmethod
     def descr_new(space, w_bytearraytype, __args__):
@@ -160,88 +192,38 @@ class W_BytearrayObject(W_Root):
             w_dict = space.w_None
         return space.newtuple([
             space.type(self), space.newtuple([
-                space.wrap(''.join(self.data).decode('latin-1')),
-                space.wrap('latin-1')]),
+                space.newunicode(''.join(self.getdata()).decode('latin-1')),
+                space.newtext('latin-1')]),
             w_dict])
 
     @staticmethod
     def descr_fromhex(space, w_bytearraytype, w_hexstring):
-        hexstring = space.str_w(w_hexstring)
-        hexstring = hexstring.lower()
-        data = []
-        length = len(hexstring)
-        i = -2
-        while True:
-            i += 2
-            while i < length and hexstring[i] == ' ':
-                i += 1
-            if i >= length:
-                break
-            if i + 1 == length:
-                raise oefmt(space.w_ValueError, NON_HEX_MSG, i)
-
-            top = _hex_digit_to_int(hexstring[i])
-            if top == -1:
-                raise oefmt(space.w_ValueError, NON_HEX_MSG, i)
-            bot = _hex_digit_to_int(hexstring[i+1])
-            if bot == -1:
-                raise oefmt(space.w_ValueError, NON_HEX_MSG, i + 1)
-            data.append(chr(top*16 + bot))
-
+        if not space.is_w(space.type(w_hexstring), space.w_unicode):
+            raise oefmt(space.w_TypeError, "must be str, not %T", w_hexstring)
+        hexstring = space.unicode_w(w_hexstring)
+        data = _hexstring_to_array(space, hexstring)
         # in CPython bytearray.fromhex is a staticmethod, so
         # we ignore w_type and always return a bytearray
         return new_bytearray(space, space.w_bytearray, data)
 
-    def descr_init(self, space, __args__):
-        # this is on the silly side
-        w_source, w_encoding, w_errors = __args__.parse_obj(
-                None, 'bytearray', init_signature, init_defaults)
-
-        if w_source is None:
-            w_source = space.wrap('')
-        if w_encoding is None:
-            w_encoding = space.w_None
-        if w_errors is None:
-            w_errors = space.w_None
-
-        # Unicode argument
-        if not space.is_w(w_encoding, space.w_None):
-            from pypy.objspace.std.unicodeobject import (
-                _get_encoding_and_errors, encode_object
-            )
-            encoding, errors = _get_encoding_and_errors(space, w_encoding,
-                                                        w_errors)
-
-            # if w_source is an integer this correctly raises a
-            # TypeError the CPython error message is: "encoding or
-            # errors without a string argument" ours is: "expected
-            # unicode, got int object"
-            w_source = encode_object(space, w_source, encoding, errors)
-
-        # Is it an int?
-        try:
-            count = space.int_w(w_source)
-        except OperationError as e:
-            if not e.match(space, space.w_TypeError):
-                raise
-        else:
-            if count < 0:
-                raise oefmt(space.w_ValueError, "bytearray negative count")
-            self.data = ['\0'] * count
-            return
-
-        data = makebytearraydata_w(space, w_source)
-        self.data = data
+    @unwrap_spec(encoding='text_or_none', errors='text_or_none')
+    def descr_init(self, space, w_source=None, encoding=None, errors=None):
+        assert isinstance(self, W_BytearrayObject)
+        data = [c for c in newbytesdata_w(space, w_source, encoding, errors)]
+        self._data = resizable_list_supporting_raw_ptr(data)
+        self._offset = 0
+        _tweak_for_tests(self)
 
     def descr_repr(self, space):
-        s = self.data
+        s, start, end, _ = self._convert_idx_params(space, None, None)
 
         # Good default if there are no replacements.
-        buf = StringBuilder(len("bytearray(b'')") + len(s))
+        buf = StringBuilder(len("bytearray(b'')") + (end - start))
 
         buf.append("bytearray(b")
         quote = "'"
-        for c in s:
+        for i in range(start, end):
+            c = s[i]
             if c == '"':
                 quote = "'"
                 break
@@ -249,7 +231,7 @@ class W_BytearrayObject(W_Root):
                 quote = '"'
         buf.append(quote)
 
-        for i in range(len(s)):
+        for i in range(start, end):
             c = s[i]
 
             if c == '\\' or c == "'":
@@ -272,17 +254,20 @@ class W_BytearrayObject(W_Root):
         buf.append(quote)
         buf.append(")")
 
-        return space.wrap(buf.build())
+        return space.newtext(buf.build())
 
     def descr_str(self, space):
-        return space.wrap(''.join(self.data))
+        if space.sys.get_flag('bytes_warning'):
+            space.warn(space.newtext("str() on a bytearray instance"),
+                       space.w_BytesWarning)
+        return self.descr_repr(space)
 
     def descr_eq(self, space, w_other):
         if isinstance(w_other, W_BytearrayObject):
-            return space.newbool(self.data == w_other.data)
+            return space.newbool(self.getdata() == w_other.getdata())
 
         try:
-            buffer = _get_buffer(space, w_other)
+            buffer = space.readbuf_w(w_other)
         except OperationError as e:
             if e.match(space, space.w_TypeError):
                 return space.w_NotImplemented
@@ -299,10 +284,10 @@ class W_BytearrayObject(W_Root):
 
     def descr_ne(self, space, w_other):
         if isinstance(w_other, W_BytearrayObject):
-            return space.newbool(self.data != w_other.data)
+            return space.newbool(self.getdata() != w_other.getdata())
 
         try:
-            buffer = _get_buffer(space, w_other)
+            buffer = space.readbuf_w(w_other)
         except OperationError as e:
             if e.match(space, space.w_TypeError):
                 return space.w_NotImplemented
@@ -321,7 +306,7 @@ class W_BytearrayObject(W_Root):
         value = self._val(space)
 
         if isinstance(w_other, W_BytearrayObject):
-            other = w_other.data
+            other = w_other.getdata()
             other_len = len(other)
             cmp = _memcmp(value, other, min(len(value), len(other)))
         elif isinstance(w_other, W_BytesObject):
@@ -330,7 +315,7 @@ class W_BytearrayObject(W_Root):
             cmp = _memcmp(value, other, min(len(value), len(other)))
         else:
             try:
-                buffer = _get_buffer(space, w_other)
+                buffer = space.readbuf_w(w_other)
             except OperationError as e:
                 if e.match(space, space.w_TypeError):
                     return False, 0, 0
@@ -364,24 +349,22 @@ class W_BytearrayObject(W_Root):
             return space.w_NotImplemented
         return space.newbool(cmp > 0 or (cmp == 0 and self._len() >= other_len))
 
-    def descr_iter(self, space):
-        return space.newseqiter(self)
-
     def descr_inplace_add(self, space, w_other):
         if isinstance(w_other, W_BytearrayObject):
-            self.data += w_other.data
+            self._data += w_other.getdata()
             return self
 
         if isinstance(w_other, W_BytesObject):
             self._inplace_add(self._op_val(space, w_other))
         else:
-            self._inplace_add(_get_buffer(space, w_other))
+            self._inplace_add(space.readbuf_w(w_other))
         return self
 
     @specialize.argtype(1)
     def _inplace_add(self, other):
+        resizelist_hint(self._data, len(self._data) + len(other))
         for i in range(len(other)):
-            self.data.append(other[i])
+            self._data.append(other[i])
 
     def descr_inplace_mul(self, space, w_times):
         try:
@@ -390,91 +373,94 @@ class W_BytearrayObject(W_Root):
             if e.match(space, space.w_TypeError):
                 return space.w_NotImplemented
             raise
-        self.data *= times
+        data = self.getdata()
+        data *= times
         return self
 
     def descr_setitem(self, space, w_index, w_other):
         if isinstance(w_index, W_SliceObject):
-            oldsize = len(self.data)
+            sequence2 = makebytesdata_w(space, w_other)
+            oldsize = self._len()
             start, stop, step, slicelength = w_index.indices4(space, oldsize)
-            sequence2 = makebytearraydata_w(space, w_other)
-            _setitem_slice_helper(space, self.data, start, step,
+            if start == 0 and step == 1 and len(sequence2) <= slicelength:
+                self._delete_from_start(slicelength - len(sequence2))
+                slicelength = len(sequence2)
+                if slicelength == 0:
+                    return
+            data = self._data
+            start += self._offset
+            _setitem_slice_helper(space, data, start, step,
                                   slicelength, sequence2, empty_elem='\x00')
         else:
-            idx = space.getindex_w(w_index, space.w_IndexError,
-                                   "bytearray index")
-            try:
-                self.data[idx] = getbytevalue(space, w_other)
-            except IndexError:
-                raise oefmt(space.w_IndexError, "bytearray index out of range")
+            idx = space.getindex_w(w_index, space.w_IndexError, "bytearray")
+            newvalue = space.byte_w(w_other)
+            self._data[self._fixindex(space, idx)] = newvalue
 
     def descr_delitem(self, space, w_idx):
         if isinstance(w_idx, W_SliceObject):
-            start, stop, step, slicelength = w_idx.indices4(space,
-                                                            len(self.data))
-            _delitem_slice_helper(space, self.data, start, step, slicelength)
+            start, stop, step, slicelength = w_idx.indices4(space, self._len())
+            if start == 0 and step == 1:
+                self._delete_from_start(slicelength)
+            else:
+                _delitem_slice_helper(space, self._data,
+                                      start + self._offset, step, slicelength)
         else:
-            idx = space.getindex_w(w_idx, space.w_IndexError,
-                                   "bytearray index")
-            try:
-                del self.data[idx]
-            except IndexError:
-                raise oefmt(space.w_IndexError,
-                            "bytearray deletion index out of range")
+            idx = space.getindex_w(w_idx, space.w_IndexError, "bytearray")
+            idx = self._fixindex(space, idx)
+            if idx == self._offset:    # fast path for del x[0] or del[-len]
+                self._delete_from_start(1)
+            else:
+                del self._data[idx]
+
+    def _delete_from_start(self, n):
+        assert n >= 0
+        self._offset += n
+        jit.conditional_call(self._offset > len(self._data) / 2,
+                             _shrink_after_delete_from_start, self)
 
     def descr_append(self, space, w_item):
-        self.data.append(getbytevalue(space, w_item))
+        self._data.append(space.byte_w(w_item))
 
     def descr_extend(self, space, w_other):
         if isinstance(w_other, W_BytearrayObject):
-            self.data += w_other.data
+            self._data += w_other.getdata()
         else:
-            self.data += makebytearraydata_w(space, w_other)
-        return self
+            self._inplace_add(makebytesdata_w(space, w_other))
 
     def descr_insert(self, space, w_idx, w_other):
         where = space.int_w(w_idx)
-        length = len(self.data)
-        index = get_positive_index(where, length)
-        val = getbytevalue(space, w_other)
-        self.data.insert(index, val)
-        return space.w_None
+        data = self.getdata()
+        index = get_positive_index(where, len(data))
+        val = space.byte_w(w_other)
+        data.insert(index, val)
 
     @unwrap_spec(w_idx=WrappedDefault(-1))
     def descr_pop(self, space, w_idx):
         index = space.int_w(w_idx)
-        try:
-            result = self.data.pop(index)
-        except IndexError:
-            if not self.data:
-                raise oefmt(space.w_IndexError, "pop from empty bytearray")
-            raise oefmt(space.w_IndexError, "pop index out of range")
-        return space.wrap(ord(result))
+        if self._len() == 0:
+            raise oefmt(space.w_IndexError, "pop from empty bytearray")
+        index = self._fixindex(space, index, "pop index out of range")
+        result = self._data.pop(index)
+        return space.newint(ord(result))
 
     def descr_remove(self, space, w_char):
         char = space.int_w(space.index(w_char))
-        try:
-            self.data.remove(chr(char))
-        except ValueError:
-            raise oefmt(space.w_ValueError, "value not found in bytearray")
-
-    _StringMethods_descr_contains = descr_contains
-    def descr_contains(self, space, w_sub):
-        if space.isinstance_w(w_sub, space.w_int):
-            char = space.int_w(w_sub)
-            return _descr_contains_bytearray(self.data, space, char)
-
-        return self._StringMethods_descr_contains(space, w_sub)
+        _data = self._data
+        for index in range(self._offset, len(_data)):
+            if ord(_data[index]) == char:
+                del _data[index]
+                return
+        raise oefmt(space.w_ValueError, "value not found in bytearray")
 
     def descr_add(self, space, w_other):
         if isinstance(w_other, W_BytearrayObject):
-            return self._new(self.data + w_other.data)
+            return self._new(self.getdata() + w_other.getdata())
 
         if isinstance(w_other, W_BytesObject):
             return self._add(self._op_val(space, w_other))
 
         try:
-            buffer = _get_buffer(space, w_other)
+            buffer = space.readbuf_w(w_other)
         except OperationError as e:
             if e.match(space, space.w_TypeError):
                 return space.w_NotImplemented
@@ -483,11 +469,61 @@ class W_BytearrayObject(W_Root):
 
     @specialize.argtype(1)
     def _add(self, other):
-        return self._new(self.data + [other[i] for i in range(len(other))])
+        return self._new(self.getdata() + [other[i] for i in range(len(other))])
 
     def descr_reverse(self, space):
-        self.data.reverse()
+        self.getdata().reverse()
 
+    def descr_clear(self, space):
+        self._data = []
+        self._offset = 0
+
+    def descr_copy(self, space):
+        return self._new(self._data[self._offset:])
+
+    def descr_hex(self, space):
+        data = self.getdata()
+        return _array_to_hexstring(space, data, 0, 1, len(data), True)
+
+    def descr_mod(self, space, w_values):
+        return mod_format(space, self, w_values, fmt_type=FORMAT_BYTEARRAY)
+
+    def descr_rmod(self, space, w_value):
+        if not isinstance(w_value, W_BytearrayObject):
+            return space.w_NotImplemented
+        return mod_format(space, w_value, self, fmt_type=FORMAT_BYTEARRAY)
+
+    @staticmethod
+    def _iter_getitem_result(self, space, index):
+        assert isinstance(self, W_BytearrayObject)
+        return self._getitem_result(space, index)
+
+    def descr_alloc(self, space):
+        return space.newint(len(self._data) + 1)   # includes the _offset part
+
+    def _convert_idx_params(self, space, w_start, w_end):
+        # optimization: this version doesn't force getdata()
+        start, end = unwrap_start_stop(space, self._len(), w_start, w_end)
+        ofs = self._offset
+        return (self._data, start + ofs, end + ofs, ofs)
+
+    def descr_getitem(self, space, w_index):
+        # optimization: this version doesn't force getdata()
+        if isinstance(w_index, W_SliceObject):
+            start, stop, step, sl = w_index.indices4(space, self._len())
+            if sl == 0:
+                return self._empty()
+            elif step == 1:
+                assert start >= 0 and stop >= 0
+                ofs = self._offset
+                return self._new(self._data[start + ofs : stop + ofs])
+            else:
+                start += self._offset
+                ret = _descr_getslice_slowpath(self._data, start, step, sl)
+                return self._new_from_list(ret)
+
+        index = space.getindex_w(w_index, space.w_IndexError, self._KIND1)
+        return self._getitem_result(space, index)
 
 
 # ____________________________________________________________
@@ -496,76 +532,74 @@ class W_BytearrayObject(W_Root):
 def _make_data(s):
     return [s[i] for i in range(len(s))]
 
-
-def _descr_contains_bytearray(data, space, char):
-    if not 0 <= char < 256:
-        raise oefmt(space.w_ValueError, "byte must be in range(0, 256)")
-    for c in data:
-        if ord(c) == char:
-            return space.w_True
-    return space.w_False
-
 # ____________________________________________________________
 
 
-def getbytevalue(space, w_value):
-    if space.isinstance_w(w_value, space.w_str):
-        string = space.str_w(w_value)
-        if len(string) != 1:
-            raise oefmt(space.w_ValueError, "string must be of size 1")
-        return string[0]
-
-    value = space.getindex_w(w_value, None)
-    if not 0 <= value < 256:
-        # this includes the OverflowError in case the long is too large
-        raise oefmt(space.w_ValueError, "byte must be in range(0, 256)")
-    return chr(value)
-
-
 def new_bytearray(space, w_bytearraytype, data):
     w_obj = space.allocate_instance(W_BytearrayObject, w_bytearraytype)
     W_BytearrayObject.__init__(w_obj, data)
     return w_obj
 
 
-def makebytearraydata_w(space, w_source):
-    # String-like argument
-    try:
-        buf = space.buffer_w(w_source, space.BUF_FULL_RO)
-    except OperationError as e:
-        if not e.match(space, space.w_TypeError):
-            raise
-    else:
-        return list(buf.as_str())
-
-    # sequence of bytes
-    w_iter = space.iter(w_source)
-    length_hint = space.length_hint(w_source, 0)
-    data = newlist_hint(length_hint)
-    extended = 0
-    while True:
-        try:
-            w_item = space.next(w_iter)
-        except OperationError as e:
-            if not e.match(space, space.w_StopIteration):
-                raise
-            break
-        value = getbytevalue(space, w_item)
-        data.append(value)
-        extended += 1
-    if extended < length_hint:
-        resizelist_hint(data, extended)
-    return data
-
-
 def _hex_digit_to_int(d):
     val = ord(d)
     if 47 < val < 58:
         return val - 48
+    if 64 < val < 71:
+        return val - 55
     if 96 < val < 103:
         return val - 87
     return -1
 
+NON_HEX_MSG = "non-hexadecimal number found in fromhex() arg at position %d"
+
+def _hexstring_to_array(space, s):
+    data = []
+    length = len(s)
+    i = 0
+    while True:
+        while i < length and s[i] == ' ':
+            i += 1
+        if i >= length:
+            break
+        if i + 1 == length:
+            raise oefmt(space.w_ValueError, NON_HEX_MSG, i)
+
+        top = _hex_digit_to_int(s[i])
+        if top == -1:
+            raise oefmt(space.w_ValueError, NON_HEX_MSG, i)
+        bot = _hex_digit_to_int(s[i + 1])
+        if bot == -1:
+            raise oefmt(space.w_ValueError, NON_HEX_MSG, i + 1)
+        data.append(chr(top * 16 + bot))
+        i += 2
+    return data
+
+HEXDIGITS = "0123456789abcdef"
+PY_SIZE_T_MAX = intmask(2**(rffi.sizeof(rffi.SIZE_T)*8-1)-1)
+
+@specialize.arg(5) # raw access
+def _array_to_hexstring(space, buf, start, step, length, rawaccess=False):
+    hexstring = StringBuilder(length*2)
+
+    if length > PY_SIZE_T_MAX/2:
+        raise OperationError(space.w_MemoryError, space.w_None)
+
+    stepped = 0
+    i = start
+    while stepped < length:
+        if rawaccess:
+            byte = ord(buf[i])
+        else:
+            byte = ord(buf.getitem(i))
+        c = (byte >> 4 & 0xf)
+        hexstring.append(HEXDIGITS[c])
+        c = (byte & 0xf)
+        hexstring.append(HEXDIGITS[c])
+        i += step
+        stepped += 1
+
+    return space.newtext(hexstring.build())
 
 class BytearrayDocstrings:
     """bytearray(iterable_of_ints) -> bytearray
@@ -591,7 +625,9 @@ class BytearrayDocstrings:
     def __alloc__():
         """B.__alloc__() -> int
 
-        Return the number of bytes actually allocated.
+        CPython compatibility: return len(B) + 1.
+        (The allocated size might be bigger, but getting it is
+        involved and may create pointless compatibility troubles.)
         """
 
     def __contains__():
@@ -622,7 +658,7 @@ class BytearrayDocstrings:
         """x.__imul__(y) <==> x*=y"""
 
     def __init__():
-        """x.__init__(...) initializes x; see help(type(x)) for signature"""
+        """Initialize self.  See help(type(self)) for accurate signature."""
 
     def __iter__():
         """x.__iter__() <==> iter(x)"""
@@ -639,6 +675,12 @@ class BytearrayDocstrings:
     def __mul__():
         """x.__mul__(n) <==> x*n"""
 
+    def __mod__():
+        """Return self%value."""
+
+    def __rmod__():
+        """Return value%self."""
+
     def __ne__():
         """x.__ne__(y) <==> x!=y"""
 
@@ -683,6 +725,18 @@ class BytearrayDocstrings:
         done using the specified fill character (default is a space).
         """
 
+    def clear():
+        """B.clear() -> None
+
+        Remove all items from B.
+        """
+
+    def copy():
+        """B.copy() -> bytearray
+
+        Return a copy of B.
+        """
+
     def count():
         """B.count(sub[, start[, end]]) -> int
 
@@ -989,9 +1043,16 @@ class BytearrayDocstrings:
         of the specified width.  B is never truncated.
         """
 
+    def hex():
+        """B.hex() -> unicode
+
+        Return a string object containing two hexadecimal digits
+        for each byte in the instance B.
+        """
+
 
 W_BytearrayObject.typedef = TypeDef(
-    "bytearray",
+    "bytearray", None, None, "read-write",
     __doc__ = BytearrayDocstrings.__doc__,
     __new__ = interp2app(W_BytearrayObject.descr_new),
     __hash__ = None,
@@ -999,6 +1060,8 @@ W_BytearrayObject.typedef = TypeDef(
                             doc=BytearrayDocstrings.__reduce__.__doc__),
     fromhex = interp2app(W_BytearrayObject.descr_fromhex, as_classmethod=True,
                          doc=BytearrayDocstrings.fromhex.__doc__),
+    maketrans = interp2app(W_BytearrayObject.descr_maketrans,
+                           as_classmethod=True),
 
     __repr__ = interp2app(W_BytearrayObject.descr_repr,
                           doc=BytearrayDocstrings.__repr__.__doc__),
@@ -1119,6 +1182,10 @@ W_BytearrayObject.typedef = TypeDef(
                              doc=BytearrayDocstrings.__setitem__.__doc__),
     __delitem__ = interp2app(W_BytearrayObject.descr_delitem,
                              doc=BytearrayDocstrings.__delitem__.__doc__),
+    __mod__ = interp2app(W_BytearrayObject.descr_mod,
+                           doc=BytearrayDocstrings.__mod__.__doc__),
+    __rmod__ = interp2app(W_BytearrayObject.descr_rmod,
+                           doc=BytearrayDocstrings.__rmod__.__doc__),
 
     append = interp2app(W_BytearrayObject.descr_append,
                         doc=BytearrayDocstrings.append.__doc__),
@@ -1132,12 +1199,17 @@ W_BytearrayObject.typedef = TypeDef(
                         doc=BytearrayDocstrings.remove.__doc__),
     reverse = interp2app(W_BytearrayObject.descr_reverse,
                          doc=BytearrayDocstrings.reverse.__doc__),
+    clear = interp2app(W_BytearrayObject.descr_clear,
+                       doc=BytearrayDocstrings.clear.__doc__),
+    copy = interp2app(W_BytearrayObject.descr_copy,
+                         doc=BytearrayDocstrings.copy.__doc__),
+    hex = interp2app(W_BytearrayObject.descr_hex,
+                           doc=BytearrayDocstrings.hex.__doc__),
+    __alloc__ = interp2app(W_BytearrayObject.descr_alloc,
+                           doc=BytearrayDocstrings.__alloc__.__doc__),
 )
 W_BytearrayObject.typedef.flag_sequence_bug_compat = True
 
-init_signature = Signature(['source', 'encoding', 'errors'], None, None)
-init_defaults = [None, None, None]
-
 
 # XXX share the code again with the stuff in listobject.py
 def _delitem_slice_helper(space, items, start, step, slicelength):
@@ -1199,48 +1271,80 @@ def _setitem_slice_helper(space, items, start, step, slicelength, sequence2,
                     "attempt to assign sequence of size %d to extended slice "
                     "of size %d", len2, slicelength)
 
-    if sequence2 is items:
-        if step > 0:
-            # Always copy starting from the right to avoid
-            # having to make a shallow copy in the case where
-            # the source and destination lists are the same list.
-            i = len2 - 1
-            start += i*step
-            while i >= 0:
-                items[start] = sequence2[i]
-                start -= step
-                i -= 1
-            return
-        else:
-            # Make a shallow copy to more easily handle the reversal case
-            sequence2 = list(sequence2)
     for i in range(len2):
         items[start] = sequence2[i]
         start += step
 
 
-class BytearrayBuffer(Buffer):
+@GCBuffer.decorate
+class BytearrayBuffer(GCBuffer):
     _immutable_ = True
 
-    def __init__(self, data, readonly):
-        self.data = data
+    def __init__(self, ba, readonly=False):
+        self.ba = ba     # the W_BytearrayObject
         self.readonly = readonly
 
     def getlength(self):
-        return len(self.data)
+        return self.ba._len()
 
     def getitem(self, index):
-        return self.data[index]
+        ba = self.ba
+        return ba._data[ba._offset + index]
 
     def setitem(self, index, char):
-        self.data[index] = char
+        ba = self.ba
+        ba._data[ba._offset + index] = char
+
+    def getslice(self, start, stop, step, size):
+        if size == 0:
+            return ""
+        if step == 1:
+            assert 0 <= start <= stop
+            ba = self.ba
+            start += ba._offset
+            stop += ba._offset
+            data = ba._data
+            if start != 0 or stop != len(data):
+                data = data[start:stop]
+            return "".join(data)
+        return GCBuffer.getslice(self, start, stop, step, size)
+
+    def setslice(self, start, string):
+        # No bounds checks.
+        ba = self.ba
+        start += ba._offset
+        for i in range(len(string)):
+            ba._data[start + i] = string[i]
+
+    def get_raw_address(self):
+        ba = self.ba
+        p = nonmoving_raw_ptr_for_resizable_list(ba._data)
+        p = rffi.ptradd(p, ba._offset)
+        return p
+
+    @staticmethod
+    def _get_gc_data_offset():
+        return get_gc_data_offset_for_list_of_chars()
+
+    def _get_gc_data_extra_offset(self):
+        return self.ba._offset
+
+    def _get_gc_data(self):
+        return get_gc_data_for_list_of_chars(self.ba._data)
 
 
 @specialize.argtype(1)
 def _memcmp(selfvalue, buffer, length):
+    # XXX that's very slow if selfvalue or buffer are Buffer objects
     for i in range(length):
         if selfvalue[i] < buffer[i]:
             return -1
         if selfvalue[i] > buffer[i]:
             return 1
     return 0
+
+def _tweak_for_tests(w_bytearray):
+    "Patched in test_bytearray.py"
+
+def _shrink_after_delete_from_start(w_bytearray):
+    w_bytearray.getdata()
