@@ -4,6 +4,10 @@ class TestW_BoolObject:
         self.false = self.space.w_False
         self.wrap = self.space.wrap
 
+    def test_init(self):
+        assert (self.false.intval, type(self.false.intval)) == (0, int)
+        assert (self.true.intval, type(self.true.intval)) == (1, int)
+
     def test_repr(self):
         assert self.space.eq_w(self.space.repr(self.true), self.wrap("True"))
         assert self.space.eq_w(self.space.repr(self.false), self.wrap("False"))
@@ -43,11 +47,6 @@ class AppTestAppBoolTest:
         assert int(False) is 0
         assert True.__int__() is 1
 
-    def test_bool_long(self):
-        assert long(True) == 1L
-        assert long(False) == 0L
-        assert True.__long__() == 1L
-
     def test_bool_ops(self):
         assert True + True == 2
         assert False | False is False
@@ -78,3 +77,14 @@ class AppTestAppBoolTest:
 
     def test_cant_subclass_bool(self):
         raises(TypeError, "class b(bool): pass")
+
+    def test_convert_to_bool(self):
+        check = lambda o: raises(TypeError, bool, o)
+        class Spam(int):
+            def __bool__(self):
+                return 1
+        raises(TypeError, bool, Spam())
+
+    def test_from_bytes(self):
+        assert bool.from_bytes(b"", 'little') is False
+        assert bool.from_bytes(b"dasijldjs" * 157, 'little') is True
